#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include "String.H"


String::String() {
#ifndef NO_OBJECTSTORE
    p = new(os_segment::of(this), srep::get_os_typespec()) srep;
#else
    p = new srep;
#endif
    //MJW p->s = 0;
    p->s = "";	//MJW Ensure that Strings always have a non-null char*
    p->size = 0;  //MJW
    p->len = 0; //MJW
}

String::String(const String& x) {
    int len = (x.p)->len; //MJW
#ifndef NO_OBJECTSTORE
    if (objectstore::is_persistent(this)) {
	if (objectstore::is_persistent(&x)) { // presumably we are in a update transaction
	    x.p->n++;
	    p = x.p;
	} else {
	    // a copy constructor, but make it persistent since <this> is
	    p = new(os_segment::of(this), srep::get_os_typespec()) srep;
	    p->s = new(os_segment::of(this), os_typespec::get_char(), len+1)
	      char[len+1]; //MJW
	    strcpy(p->s, x.p->s);
	    p->size = len;  //MJW
	    p->len = len;  //MJW
	} 
    } else {				// <this> is not persistent
	if (objectstore::is_persistent(&x)) {
	    // a copy constructor, but make it transient since <this> is
	    p = new srep;
	    p->s = new char[len+1];  //MJW
	    p->size = len;  //MJW
	    p->len = len;  //MJW
	    strcpy(p->s, x.p->s);
	} else {			// both are transient
	    x.p->n++;
	    p = x.p;
	}
    }
#else
    x.p->n++;
    p = x.p;
#endif
}

String::String(const char* s) {
    if (!s) s = ""; //MJW
    int len = strlen(s);  //MJW
#ifndef NO_OBJECTSTORE
    p = new(os_segment::of(this), srep::get_os_typespec()) srep;
    p->s = new(os_segment::of(this), os_typespec::get_char(), len+1)
      char[ len+1 ];  //MJW
#else
    p = new srep;
    p->s = new char[ len+1 ];  //MJW
#endif
    p->size = len;  //MJW
    p->len = len;  //MJW
    strcpy(p->s, s);
}

String::~String() {
    if (--p->n == 0) {
	delete[] p->s;
	delete p;
    }
}

String& 
String::operator=(const char* s) {
    if (!s) s = ""; //MJW
    int len = strlen(s);  //MJW
    if (p->n > 1) {			// disconnect self
	p->n--;
#ifndef NO_OBJECTSTORE
	p = new(os_segment::of(this), srep::get_os_typespec()) srep;
#else
	p = new srep;
#endif
    } else				// free old String
      delete[] p->s;

#ifndef NO_OBJECTSTORE
    p->s = new(os_segment::of(this), os_typespec::get_char(), len+1)
	char[len+1];  //MJW
#else
    p->s = new char[len+1];  //MJW
#endif
    p->size = len;  //MJW
    p->len = len;  //MJW
    strcpy(p->s, s);
    return *this;
}

String& 
String::operator=(const String& x) {
    x.p->n++;	// protect against "st == st"
    if (--p->n == 0) {
	delete[] p->s;
	delete p;
    }
    p = x.p;
    return *this;
}

// from The C++ Answer Book, Tony L. Hansen
//MJW  The $%^$ book above recommends O(n**2) algorithms??
String& 
String::operator+=(const String& st2) {
    char *s1 = p->s;
    char *s2 = st2.p->s;
    int s1len = p->len; //MJW
    int s2len = (st2.p)->len; //MJW
    int totlen = s1len + s2len + 1;
    if (p->size<totlen) { //MJW
#ifndef NO_OBJECTSTORE
        char *ns = new(os_segment::of(this), os_typespec::get_char(), totlen)
	char[totlen];
#else
        totlen += totlen>>2; //MJW increase physical size by 25%
        char *ns = new char[totlen];
#endif
        strcpy(ns, s1);
        strcpy(ns+s1len, s2);
        delete[] s1; //MJW
        p->s = ns;
        p->size = totlen-1;  //MJW
    } else {
        strcpy( p->s + s1len, s2 );
    }
    p->len = s1len + s2len; //MJW
    return *this;
}

String 
operator+(String& st1, String& st2) {
    char *s1 = st1.p->s;
    char *s2 = st2.p->s;
    int s1len = (st1.p)->len; //MJW
    int s2len = (st2.p)->len; //MJW
    int totlen = s1len + s2len + 1;
#ifndef NO_OBJECTSTORE
    char *ns = new(os_segment::of(&st1), os_typespec::get_char(), totlen)
	char[totlen];
#else
    char *ns = new char[totlen];
#endif

    strcpy(ns, s1);
    strcpy(ns+s1len, s2);
    
    String st(ns);
    delete[] ns; //MJW
    return st;
}

String 
operator+(String& st1, const char* s2) {
    if (!s2) s2 = ""; //MJW
    char *s1 = st1.p->s;
    int s1len = (st1.p)->len;  //MJW
    int s2len = strlen(s2);
    int totlen = s1len + s2len + 1;
#ifndef NO_OBJECTSTORE
    char *ns = new(os_segment::of(&st1), os_typespec::get_char(), totlen) char[totlen];
#else
    char *ns = new char[totlen];
#endif

    strcpy(ns, s1);
    strcpy(ns+s1len, s2);
    
    String st(ns);
    delete[] ns;  //MJW
    return st;
}

String 
operator+(const char* s1, String& st1) {
    if (!s1) s1 = ""; //MJW
    char *s2 = st1.p->s;
    int s1len = strlen(s1);
    int s2len = (st1.p)->len; //MJW
    int totlen = s1len + s2len + 1;
#ifndef NO_OBJECTSTORE
    char *ns = new(os_segment::of(&st1), os_typespec::get_char(), totlen) char[totlen];
#else
    char *ns = new char[totlen];
#endif

    strcpy(ns, s1);
    strcpy(ns+s1len, s2);
    
    String st(ns);
    delete[] ns;  //MJW
    return st;
}

ostream& 
operator<<(ostream& s, const String& x) {
    return s << x.p->s;
}

istream& 
operator>>(istream& s, String& x) {
    char buf[256];
    s.get(buf, 256, '\n');
    
    char c;
    if (s.get(c) && c != '\n')
      cerr << "Input string too long. First 256 characters:\n" << buf << endl;
    else
      x = buf;
    return s;
}

char& 
String::operator[](int i) {
    if (i<0 || strlen(p->s)<i) {
	cerr << "index out of range\n";
	exit(1);
    }
    return p->s[i];
}

const char& 
String::operator[](int i) const {
    if (i<0 || strlen(p->s)<i) {
	cerr << "index out of range\n";
	exit(1);
    }
    return p->s[i];
}
