// $Author: jmartin $
// $Date: 1997-12-05 01:57:18 $
// $Source: /root/cvs/Rigi/rigiedit/UT/src/avlnew.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

//	avlnew.c: AVL treesets of AVL treesets

#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>

#include	"avlnew.h"

const int INVALID	= (~0);
const int PROBEONLY	= ((~0)^1);

static const int EVEN	= 0;
static const int LEFT	= 1;
static const int RIGHT	= 2;

// This stuff is disgusting!! I'm trying to cram all
// info into 3 words. Don't read.

#define ul(x)		((unsigned long)(x))
#define ar(x)		((AvlRecord*)(x))
#define arr(x)		((AvlRecord**)(x))

#define Star(x)		(ar(ul(*(x))&~0x3))
#define SetStar(x,n)	(*(x)=ar((ul(*(x))&0x3)|ul(n)))

#define GetLeft(x)	(ar(ul((x)->left)&~0x3))
#define	GetLean(x)	(ul((x)->left)&0x3)
#define SetLeft(x,n)	(((x)->left)=ar((ul((x)->left)&0x3)|ul(n)))
#define	SetLean(x,l)	do{((x)->left)=ar(ul((x)->left)&~0x3);\
			((x)->left)=ar(ul((x)->left)|(l));}while(0)
/*
#define Star(x)		(*(x))
#define GetLeft(x)	(((x)->left))
#define	GetLean(x)	((x)->lean)
#define SetLeft(x,n)	((x)->left=(n))
#define	SetLean(x,l)	(x)->lean = (l)
*/

static AvlRecord*	NewRecord();

static AvlInfo info;
static AvlInfo *returnvalue;

static AvlSet *currset;
static AvlSet dictionary = NULL;

static AvlName string, lastkey;

static int debug=0;
static int mem=0, charmem=0, memrecs=0;
static int probeonly;

// either store a new record, or just find it

static int Store( AvlRecord **tree )
{
    // static vars: string, info, returnvalue, probeonly
    // currset, dictionary

    AvlRecord *n = Star(tree);
    int j;

    if (n==NULL)
    {
	returnvalue = NULL;
	if (probeonly) return 0;

	n = NewRecord();
	SetStar(tree,n);
	returnvalue = (info==INVALID) ? NULL : &(n->info);

	n->info = info;

	// If storing in the dictionary, we need a brand new name.
	// Otherwise, we need the last name stored in the dictionary

	if (currset==&dictionary)
	{
		n->name = strdup( string );
		charmem += strlen(string)+1;
	}
	else n->name = lastkey;

	string = n->name;

	if(debug)printf("new record %s, info is %d, returnvalue is %d\n",
	    string,info,returnvalue);

	// this tree just got deeper
	return 1;
    }

    // comparison

    if (!string) fprintf(stderr,"error: null string in AvlStore()\n");
    if (!n->name) fprintf(stderr,"error: null name in AvlStore()\n");
    j = strcmp( string, n->name );

    if (j==0) // found the appropriate record
    {
	if ( info && info!=INVALID && info!=PROBEONLY ) n->info = info;
	returnvalue = &(n->info);
	string = n->name;

	if(debug)printf("found record %s, info is %d, returnvalue is %d\n",
	string,info,returnvalue);

	return 0;
    }
    else if (j<0) // recur down the tree
    {
	j = Store( &(n->left) ) ? LEFT : EVEN;
    }
    else
    {
	j = Store( &(n->right) ) ? RIGHT : EVEN;
    }
    if (j==0) return 0;


    if (j!=GetLean(n))
    {
	// not a pivot - just set the weight factor
	if (GetLean(n)!=EVEN)
		SetLean(n,EVEN); else SetLean(n,j);

	// no change to nodes above if this subtree became EVEN
	return GetLean(n) ? 1 : 0;
    }


    // single rotations

    if (j==RIGHT && GetLean(n->right)==RIGHT)
    {
	SetLean(n,EVEN); SetLean(n->right,EVEN);

	SetStar(tree,n->right);
	n->right = GetLeft(n->right);
	SetLeft(Star(tree),n);
    }
    else if (j==LEFT && GetLean(GetLeft(n))==LEFT)
    {
	SetLean(n,EVEN); SetLean(GetLeft(n),EVEN);

	SetStar(tree,GetLeft(n));
	SetLeft(n,GetLeft(n)->right);
	Star(tree)->right = n;
    }

    // double rotations

    else if (j==RIGHT && GetLean(n->right)==LEFT)
    {
	SetStar(tree,GetLeft(n->right));
	SetLeft(n->right,Star(tree)->right);
	Star(tree)->right = n->right;
	n->right = GetLeft(Star(tree));
	SetLeft(Star(tree),n);

	SetLean(n,GetLean(Star(tree))!=RIGHT ? EVEN : LEFT );
	SetLean(Star(tree)->right, GetLean(Star(tree))!=LEFT ? EVEN : RIGHT );
	SetLean(Star(tree),EVEN);
    }
    else if (j==LEFT && GetLean(GetLeft(n))==RIGHT)
    {
	SetStar(tree,GetLeft(n)->right);
	GetLeft(n)->right = GetLeft(Star(tree));
	SetLeft(Star(tree),GetLeft(n));
	SetLeft(n,Star(tree)->right);

	Star(tree)->right = n;

	SetLean(n,GetLean(Star(tree))!=LEFT ? EVEN : RIGHT);
	SetLean(GetLeft(Star(tree)),GetLean(Star(tree))!=RIGHT ? EVEN : LEFT);
	SetLean(Star(tree),EVEN);
    }
    else fprintf(stderr,"error: inconsistent state in AVL tree-set!\n");

    return 0;
}

AvlInfo *AvlSetInfo( AvlSet *set, AvlName s, AvlInfo i )
{
    if (!s) fprintf( stderr,"error: null string in AvlSetInfo()\n" );

    // Before we do anything else, we must ensure that s is in the local
    // dictionary of AvlNames (we are not allowed to let traversals interleave)

    string = s; info = NULL;
    currset = &dictionary;
    Store( &dictionary );
    *returnvalue = (AvlInfo)(lastkey = string);

    // Now, the normal store

    string = s; info = i;
    currset = set;
    if ( i==PROBEONLY ) probeonly = 1;
    Store( set );
    lastkey = string;
    probeonly = 0;

    return returnvalue;
}

static AvlInfo Preorder( AvlRecord *R, int (*f)(AvlName,AvlInfo) )
{
    AvlInfo I;
    if ( R==NULL ) return NULL;

    if ( I=Preorder( GetLeft(R),f ) ) return I;
    if ( (*f)( R->name,R->info ) ) return R->info;
    if ( I=Preorder( R->right,f ) ) return I;
    return NULL;
}

static AvlInfo Preorder( AvlRecord *R, int (*f)(AvlName) )
{
    AvlInfo I;
    if ( R==NULL ) return NULL;

    if ( I=Preorder( GetLeft(R),f ) ) return I;
    if ( (*f)( R->name ) ) return R->info;
    if ( I=Preorder( R->right,f ) ) return I;
    return NULL;
}

AvlInfo AvlIterate( AvlSet set, int (*f)(AvlName,AvlInfo) )
{ return Preorder( set,f ); }

AvlInfo AvlIterate( AvlSet set, int(*f)(AvlName) )
{ return Preorder( set,f ); }


static int maxlevel;
static int numset;
static int unbalanced;

static int Print( AvlRecord *tree, int level, int flag )
{
    AvlRecord *n = tree; int l,r;

    if (n==NULL) return 0;

    ++ numset;
    if (level>maxlevel) maxlevel = level;

    l = 1 + Print( GetLeft(n), level+1, flag );
    if (flag) printf("%2d%c %4d ", level,
	GetLean(n)==EVEN?'E':(GetLean(n)==LEFT?'L':'R'),
	n->info );
    printf("%s\n", n->name );
    r = 1 + Print( (n->right), level+1, flag );

    if ((r-l)>1 || (r-l)<-1) unbalanced = 1;
    return l>r ? l : r;
}

int PrintSet( AvlSet set )
{
    int d;

    numset = maxlevel = unbalanced = 0;

    d = Print( set, 0, 1 );
    if (unbalanced)
	fprintf( stderr, "error: AVL sets not working properly\n" );
    if (maxlevel!=d-1)
	fprintf( stderr, "error: PrintSet() not working (d=%d)!\n",d );
    printf( "(total=%d maxdepth=%d)\n\n",numset,maxlevel );
    return numset;
}

int PrintDictionary()
{
    numset = maxlevel = unbalanced = 0;
    return Print( dictionary, 0, 0 );
}

// MEMORY ALLOCATION

static AvlRecord *rlist = NULL;

static AvlRecord *NewRecord()
{
    static int maxindex = 64;
    static int index = maxindex;
    static const int maxblock = 512;
    static AvlRecord *array = NULL;

    AvlRecord *r;
    memrecs++;

    if (rlist) {
	r = rlist; rlist = r->right;
    } else if (index<maxindex) {
	r = array + index++;
    } else {
	if ( maxindex < maxblock ) maxindex<<=1;
	array = new AvlRecord[maxindex];
	mem += maxindex*sizeof(AvlRecord);
	index = 0;
	r = array + index++;
    }
    r->name = NULL; r->left = r->right = NULL;
    r->info = 0;
    //r->lean = 0;
    return r;
}

static int FreeRecords( AvlRecord *n, void(*f)(AvlName,AvlInfo*) )
{
    if (n==NULL) return 0;
    int r = FreeRecords( GetLeft(n),f );
    r += FreeRecords( n->right,f );

    if (f) (*f)(n->name,&(n->info));

    n->right = rlist; rlist = n;
    memrecs--;
    return 1+r;
}

int FreeSet( AvlSet *set ) { return FreeSet( set, NULL ); }

int FreeSet( AvlSet *set, void(*f)(AvlName,AvlInfo*) )
{
    int r = FreeRecords( *set, f );
    *set = NULL;
    return r;
}

void AvlMem( int *a, int *b, int *c ) { *a = mem; *b = charmem; *c = memrecs; }

const AvlName AvlLastName()
{
    return lastkey;
}

// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
//   Revision 1.5  1995/10/25  20:52:11  mwhitney
//   put level info in left pointer (!)
//
//   Revision 1.4  1995/10/04  17:30:35  mwhitney
//   macros for lean values, dictionary of local names
//
//   Revision 1.3  1994/10/07  18:27:21  mwhitney
//   FreeRecords takes a func, used to free contents of indiv records
//
//   Revision 1.2  1994/08/29  22:31:40  mwhitney
//   comment out the last/newrecord shrtcut; apparently buggy
//
//   Revision 1.1  1994/08/29  18:05:14  mwhitney
//   Initial revision
//
//
