// $Author: jmartin $
// $Date: 2000-02-05 02:18:35 $
// $Source: /root/cvs/Rigi/rigiedit/UT/src/sexps.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

// Stuff for making and parsing s-expressions
// The parsing method is simple recursive descent

#include <stdio.h>
#include <ctype.h>

// External interface: IN - pointer to sexp. OUT - a callback function,
// called for each literal, with the literal's depth and index.

void parse_sexp( const char*, void (*)(char*,int,int,int) );
int parse_sexps( FILE*, void (*)(char*,int,int,int) );


static int  parse_list( int d );
static int  parse_elements( int d, int i );
static int  parse_element( int d, int i );

static int  scan_literal( int d, int i );
static int  scan_space();
static int  match( char );

static const char *s = NULL, *b = NULL;
static FILE *fp = NULL;
static void (*user_callback)(char*,int,int,int);

static const int BUFFSIZE = 2048;
static char buff[BUFFSIZE];

static int topindex = 0;


void parse_sexp( const char *p, void (*g)(char*,int,int,int) )
{
    b = s = p;
    fp = NULL;
    user_callback = g;
    parse_list(0);
}

int parse_sexps( FILE *file, void (*g)(char*,int,int,int) )
{
    fp = file;
    user_callback = g;
    int i;
    for (i=0; parse_list(0); i++ );
    fp = NULL;
    return i;
}

////////////////////////////////////////////////////////////////////////////

// Generic recursive-descent s-expression (LISP) parser

static inline int GETCH()		{ return fp ? getc(fp) : (int)*s++; }
static inline void UNGETCH( int c )	{ if (fp) ungetc(c,fp); else s--; }
static inline int EOE()			{ return fp ? EOF : '\0'; }

static int PEEK()
{
    if (fp==NULL) return (int)*s;
    int c = getc(fp); ungetc(c,fp); return c; 
}

static inline int scan_space()
{
    register int c;
    for (;;)
    {
	do c = GETCH(); while (isspace(c));

	// Get rid of C++ style comments
	if (c!='/') break;
	if (PEEK()=='/')
	{
	    c = GETCH();
	    do c = GETCH(); while (c!='\n' && c!=EOE());
	}
	else break;
    }
    UNGETCH(c);
    return c;
}

static inline int match( char x )
{
    scan_space();
    register int c = GETCH();
    return c==x;
}

static int parse_list( int d )
{
    if ( EOE() == scan_space() ) return 0;

    return match('(') && parse_elements( d,0 ) && match(')');
}

static int parse_elements( int d, int i )
{
    register int c = scan_space();

    if (c==EOE())
    {
	fprintf( stderr,"parse error: early EOS!\n");
	fprintf( stderr,"last literal scanned: %s\n",buff);
	return 0;
    }

    // c is a lookahead character here. It has not been consumed.
    if (c==')') return 1;

    return parse_element( d,i ) && parse_elements( d,i+1 );
}

static int parse_element( int d, int i )
{
    if (d==0) topindex = i;

    // Get a lookahead character.
    register int c = GETCH(); UNGETCH(c);

    return c=='(' ? parse_list( d+1 ) : scan_literal( d,i );
}

static int scan_literal( int d, int i )
{
    register int c, j, x, y;

    for ( j=x=y=0;; )
    {
	c = GETCH();
	if (c==EOE()) break;

	// check for enclosing doublequote characters
	if (c=='"') if (y==0) x = !x;

	// check for end-of-literal
	if (x==0)
	if (c==')' || c=='(' || isspace(c)) break;

	y = c=='\\';
	if ( j<BUFFSIZE) buff[j++] = c;
    }
    UNGETCH(c);

    if (j<BUFFSIZE) buff[j] = '\0';
    else 
    {
	fprintf( stderr,"error: buffer overflow in s-exp parse\n");
	buff[31] = '\0';
	fprintf( stderr,"buffer contents: %s...\n",buff );
    }

    if (user_callback) user_callback( buff,d,i,topindex );
    return 1;
}

////////////////////////////////////////////////////////

// This stuff helps in writing out s-exps (and everything else!)
// but is essentially brain-dead.

static const int SEXPSIZE = 64500; // HACK!
static char sexpbuff[SEXPSIZE];
static char *cursp = sexpbuff;
static char *maxsp = sexpbuff + SEXPSIZE;


void write_sexp( const char *s )
{
    register char *m = maxsp;
    register char *sp;

    for ( sp=cursp; *s && sp<m; )
    {
	*sp++ = *s++;
    }
    if ( sp==m )
    {
	fprintf(stderr,"error, sexp buffer overflow!!\n");
	sp--;
    }
    cursp = sp;
}

void reset_sexp()
{
    cursp = sexpbuff;
    *cursp = '\0';
}

void close_sexp()
{
    *cursp = '\0';
}

char *get_sexp()
{
    if (!*sexpbuff) return NULL;
    *cursp = '\0';
    return sexpbuff;
}


//////////////////////////////////////////////////////////

// Here be routines reflecting the NetworkObject grammar.
// They are not used right now, but they apparently work.

int scan( char x )
{
    int c = scan_space();
    if (c==x) { GETCH(); return 1; } else return 0;
}

int atom()
{
    int c = scan_space();
    if (c!='(' && c!=')') return scan_literal(0,0); else return 0;
}

// recursive descent grammar/parser follows ...

int atoms()	{ return (atom() && atoms()) || 1; }

int list()	{ return scan('(') && atoms() && scan(')'); }

int inlist()	{ return list(); }

int isalist()	{ return list(); }

int catlist()	{ return list(); }

int attrpair()	{ return scan('(')&& atom() && (atom()||1) && scan(')'); }

int attrpairs()	{ return (attrpair() && attrpairs()) || 1; }

int valuelist()	{ return scan('(') && attrpairs() && scan(')'); }

int attrdecl()	{ return scan('(') && catlist() && valuelist() && scan(')'); }

int attrdecls()	{ return (attrdecl() && attrdecls()) || 1; }

int attrslist()	{ return scan('(') && attrdecls() && scan(')'); }

int networkobject()
{
    return
	scan('(') &&
		atom() && atom() && inlist() && isalist() &&
		attrslist() &&
	scan(')');
}

void parse_object( const char *p )
{
    b = s = p;
    user_callback = NULL;
    int r = networkobject();
    if (r) printf("parse successful\n"); else
    {
        printf("bad parse, character %d\n", s-b+1 );
	printf("sexp is:\n%s\n",p);
    }
}

// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
// Revision 1.1  1997/12/05 01:57:18  jmartin
// Initial revision
//
//   Revision 1.8  1995/10/25  22:15:35  mwhitney
//   HUGE buffer! Should use a String Class.
//
//   Revision 1.7  1995/06/16  22:01:06  mwhitney
//   fix literal scanner, and scan past comments
//
//   Revision 1.6  1995/02/15  19:03:49  mwhitney
//   extra parm for parse callback function
//   for "topindex"
//
//   Revision 1.5  1994/10/28  15:47:03  mwhitney
//   allow for literals with escaped double-quote within double-quotes
//
//   Revision 1.4  1994/10/28  00:39:02  mwhitney
//   added NetworkObject recursize descent parser
//
//   Revision 1.3  1994/10/21  19:18:14  mwhitney
//   minor
//
//   Revision 1.2  1994/08/19  19:09:07  mwhitney
//   Many changes
//
//   Revision 1.1  1994/08/17  20:48:08  mwhitney
//   Initial revision
//
//
