#include "tml.H"

ErrorType* TMB_ERROR;
Client *CLIENT;
Server *SERVER;

RequestQueue   *REQUEST_QUEUE;
ResponseQueue  *RESPONSE_QUEUE;
MonologueQueue *MONOLOGUE_QUEUE;

tml_Method *TDS_TELL;
tml_Method *TDS_ASK;
tml_Method *TDS_COMMAND;

ToolClass* TELOS_DATA_SERVER;
ToolClass* REFINE;
ToolClass* RIGI;
ToolClass* MESSAGE_SERVER;

CompletionStatus
t_sexprToCompletionStatus(t_sexp sexp) {
    CompletionStatus status;
    char* tmp = MBprint_Cstring(sexp);

    switch(*tmp) {
    case (SUCCESS_CHAR):
	status = SUCCESS;
	break;
    case (UNKNOWN_CHAR):
	status = UNKNOWN;
	break;
    case (ERROR_CHAR):
	status = ERROR;
	break;
    case (NULLSTATUS_CHAR):
	status = NULLSTATUS;
	break;
    }
    (*MBfree_func)(tmp);
    return(status);
}	

void
TMLinit() {
    CLIENT = new Client;
    SERVER = new Server;
    TMB_ERROR = new ErrorType();

    REQUEST_QUEUE   = new RequestQueue;
    RESPONSE_QUEUE  = new ResponseQueue;
    MONOLOGUE_QUEUE = new MonologueQueue;
    
    TDS_TELL = new tml_Method("TELL");
    TDS_ASK =  new tml_Method("ASK");
    TDS_COMMAND =  new tml_Method("COMMAND");
    
    TELOS_DATA_SERVER = new ToolClass("TELOS_DATA_SERVER");
    REFINE = new ToolClass("REFINE");
    RIGI = new ToolClass("RIGI");
    MESSAGE_SERVER = new ToolClass("MESSAGE_SERVER");
}


MessageType 
TMBgetMessageType(t_sexp message) { 
    char* tmpStr = MBprint_Cstring(MBcar(
        MBnth(MBnth(message, TRANS_DATA_POSITION),MESSAGE_TYPE_POSITION)));
    MessageType tmp = (MessageType)atoi(tmpStr);
    (*MBfree_func)(tmpStr);
    return tmp;
};

NewMessageStatus 
TMBgetNewMessage() { 
    t_sexp message;
    NewMessageStatus ret;
    message = MB_getNewMessage();
    if (!(message == (t_sexp)NULL)) { 
	MessageType msgType = TMBgetMessageType(message);
	switch(msgType) {
	case REQUEST: { 
	    Request* request = new Request(message);
	    REQUEST_QUEUE->join(request);
	    ret = NEW_REQUEST;
	    break;
	}
	case RESPONSE: { 
	    Response* response = new Response(message);
	    RESPONSE_QUEUE->join(response);
	    ret = NEW_RESPONSE;
	    break;
	}
	case MONOLOGUE: { 
	    Monologue* monologue = new Monologue(message);
	    MONOLOGUE_QUEUE->join(monologue);
	    ret = NEW_MONOLOGUE;
	    break;
	}
	default:
	    ret = UNRECOGNIZABLE_MESSAGE;
	    TMB_ERROR->setError(TMB_UnrecognizableMessage);
	    char* tmpStr = MBprint_Cstring(message);
	    TMB_ERROR->setMessageBuffer(tmpStr);
	    (*MBfree_func)(tmpStr);
	    break;
	};
    }
    else
	ret = NO_NEW_MESSAGE;
    MBfree(message);
    return ret;
};

NewMessageStatus 
TMBwaitForMessage() { 
    t_sexp message;
    NewMessageStatus ret;
    message = MB_waitForMessage();
    if (!(message == (t_sexp)NULL)) { 
	MessageType msgType = TMBgetMessageType(message);
	switch(msgType) {
	case REQUEST: { 
	    Request* request = new Request(message);
	    REQUEST_QUEUE->join(request);
	    ret = NEW_REQUEST;
	    break;
	}
	case RESPONSE: { 
	    Response* response = new Response(message);
	    RESPONSE_QUEUE->join(response);
	    ret = NEW_RESPONSE;
	    break;
	}
	case MONOLOGUE: { 
	    Monologue* monologue = new Monologue(message);
	    MONOLOGUE_QUEUE->join(monologue);
	    ret = NEW_MONOLOGUE;
	    break;
	}
	default:
	    ret = UNRECOGNIZABLE_MESSAGE;
	    TMB_ERROR->setError(TMB_UnrecognizableMessage);
	    char* tmpStr = MBprint_Cstring(message);
	    TMB_ERROR->setMessageBuffer(tmpStr);
	    (*MBfree_func)(tmpStr);
	    break;
	};
    }
    else
	ret = NO_NEW_MESSAGE;
    MBfree(message);
    return ret;
};


