
#include "tml.H"

static String* sQuoteSpaceQuote = new String("\" \"");
static String* sQuoteSpaceRBracket = new String("\" )");
static String* sRBracket = new String(")");
static String* sRRBracket = new String("))");
static String* sLBracket = new String("(");
static String* sLLBracket = new String("((");
static String* sLRBracket = new String("()");
static String* sRLBracket = new String(")(");
static String* sSpace = new String(" ");
static String* sQuoteSpace = new String("\" ");
static String* sSpaceRBracket = new String(" )");
static String* sSpaceQuote = new String(" \"");

static const char* SUCCESS_STRING =    "SUCCESS";
static const char* UNKNOWN_STRING =    "UNKNOWN";
static const char* ERROR_STRING =      "ERROR";
static const char* NULLSTATUS_STRING = "NULLSTATUS";

ArgumentList::ArgumentList(t_sexp sexp): GenericList()
      { int n = MBlength(sexp);
        ArgumentString* arg;
        char* tmp;
        for (int i = 0; i < n; i++)
            { tmp = MBprint_Cstring(MBnth(sexp, i));
              arg = new ArgumentString(tmp);
              (*MBfree_func)(tmp);
              addItem(arg);
            }
      }

void ArgumentList::print (ostream& outStream)
      { outStream << "(";
        int n = getSize();
        if (n > 0)
          { for (int i = 0; i < n - 1; i++)
                outStream << *getArgument(i) << ", ";
            outStream << *getArgument(n-1);
           };
        outStream << ")";
         
      }

static const char* MESSAGE_STRING =   "MESSAGE";
static const char* MONOLOGUE_STRING = "MONOLOGUE";
static const char* REQUEST_STRING =   "REQUEST";
static const char* RESPONSE_STRING =  "RESPONSE";
static const char* COMMAND_STRING =   "COMMAND";



String *Message::makeMBDomainString(DomainList* domains)
      { String* str = new String("(");
        String* tmp;
        int n = domains->getSize();
        for (int i = 0; i < n; i++)
            { tmp = MB_makeDomainString(
                       domains->getDomain(i)->getContext(),
                       domains->getDomain(i)->getUserId(),
                       domains->getDomain(i)->getToolClass());
              *str += *sSpace;
              *str += *tmp;
              delete(tmp);
            }
        *str = (*str + " )");
        return str;
      }

String *Message::makeMBDomainSpecString(DomainList* domains)
      { String* str = new String("(");
        String* tmp;
        int n = domains->getSize();
        for (int i = 0; i < n; i++)
            { tmp = MB_makeDomainSpecString(
                       domains->getDomain(i)->getContext(),
                       domains->getDomain(i)->getUserId(),
                       domains->getDomain(i)->getToolClass());
              *str += *sSpace;
              *str += *tmp;
              delete(tmp);
            }
        *str = (*str + " )");
        return str;
      }

String *Message::makeDomainString(Domain* d)
      { String* dataString = new String("((");
        *dataString += *(d->getContext());
        *dataString += *sRLBracket;
        *dataString += *(d->getToolClass());
        *dataString += *sRLBracket;
        *dataString += *(d->getUserId());
        *dataString += *sRRBracket;
        return dataString;
      }

String *Message::makeDomainListString(DomainList* d)
      { String* dataString = new String("( ");
        String* tmp;
        int n = d->getSize();
        for (int i = 0; i < n; i++)
          { tmp = makeDomainString(d->getDomain(i));
            *dataString += *tmp;
            *dataString += *sSpace;
            delete(tmp);
          }
        *dataString += *sRBracket;
        return dataString;
      }

const char *Monologue::getTypeString()
        { const char* ret;
          switch(type) {
             case MESSAGE:
               ret = MESSAGE_STRING;
               break;
             case MONOLOGUE:
               ret = MONOLOGUE_STRING;
               break;
             case REQUEST:
               ret = REQUEST_STRING;
               break;
             case RESPONSE:
               ret = RESPONSE_STRING;
               break;
             case COMMAND:
               ret = COMMAND_STRING;
               break;
          }
           return ret;
        }

Monologue::Monologue(t_sexp sexp)
      { t_sexp data = MBnth(sexp, TRANS_DATA_POSITION);
        char* idStr = MBprint_Cstring(MBcar(MBnth(data, MESSAGE_ID_POSITION)));
        char* monologueStr =
               MBprint_Cstring(MBcar(MBnth(data, MESSAGE_MONOLOGUE_POSITION)));
        id = new MessageIdString(idStr);
        monologue = new MonologueString(monologueStr);
        (*MBfree_func)(idStr);
        (*MBfree_func)(monologueStr);

        type = MONOLOGUE;
        MBfree(sexp);
      }

void Monologue::print (ostream& outStream)
      { outStream << endl << "  Monologue" << endl << "  ---------" << endl;
        outStream << "    type:              " << *getTypeString()  << endl;
        Message::print();
        outStream << "    monologue:         " << *getMonologue()   << endl;
      }

String *Monologue::makeMonologueString()
      { char str[256];
        sprintf(str, "%d", type);
        String* dataString = new String("((");
        *dataString += *(getId());
        *dataString += *sRLBracket;
        *dataString += *(makeSenderString());
        *dataString += *sRLBracket;
        *dataString += *(makeDomainListString(getReceiver()));
        *dataString += *sRLBracket;
        *dataString += str;
        *dataString += *sRLBracket;
        *dataString += *(getMonologue());
        *dataString += *sRRBracket;
        return dataString;
      }

Boolean Monologue::sendMonologue()
      { Boolean ret;
        String* dataStr = makeMonologueString();
        String* domainStr  = makeMBDomainSpecString(getReceiver());
        MB_sendMessage(dataStr, domainStr);
        if (TMB_ERROR->found())
           ret = FALSE;
        else
           ret = TRUE;
        delete(dataStr);
        delete(domainStr);
        return ret;
      }



const char *Response::getTypeString()
        { const char* ret;
          switch(type) {
             case MESSAGE:
               ret = MESSAGE_STRING;
               break;
             case MONOLOGUE:
               ret = MONOLOGUE_STRING;
               break;
             case REQUEST:
               ret = REQUEST_STRING;
               break;
             case RESPONSE:
               ret = RESPONSE_STRING;
               break;
             case COMMAND:
               ret = COMMAND_STRING;
               break;
          }
           return ret;
        }

const char *Response::getCompletionStatusString() {
        const char *compStr;
        switch (completion) {
        case SUCCESS:
            compStr = SUCCESS_STRING;
            break;
        case UNKNOWN:
            compStr = UNKNOWN_STRING;
            break;
        case ERROR:
            compStr = ERROR_STRING;
            break;
        case NULLSTATUS:
            compStr = NULLSTATUS_STRING;
            break;
        }
        return compStr;
    }

String *Response::makeReturnValueListString()
      { String* dataString = new String("( ");
        int n = returnValues->getSize();
        for (int i = 0; i < n; i++)
          { *dataString += *(returnValues->getReturnValue(i));
            *dataString += *sSpace;
          }
        *dataString += *sRBracket;
        return dataString;
      }

String *Response::makeErrorMessageListString()
      { String* dataString = new String("( ");
        int n = errorMessages->getSize();
        for (int i = 0; i < n; i++)
          { *dataString += *sLBracket; //MJW
            *dataString += *(errorMessages->getErrorMessage(i));
            //MJW *dataString += *sSpace;
            *dataString += *sRBracket; //MJW
          }
        *dataString += *sRBracket;
        return dataString;
      }

Response::Response(t_sexp sexp)
{
        t_sexp data = MBnth(sexp, TRANS_DATA_POSITION);
        char* idStr = MBprint_Cstring(MBcar(MBnth(data,
                                                  MESSAGE_ID_POSITION)));
        char* responseToStr = MBprint_Cstring(MBcar(MBnth(data,
                                        MESSAGE_RESPONSETO_POSITION)));
        id = new MessageIdString(idStr);
        responseTo = new MessageIdString(responseToStr);
        (*MBfree_func)(idStr);
        (*MBfree_func)(responseToStr);

        sender = new Domain(MBcar(MBnth(data, MESSAGE_SENDER_POSITION)));
        receiver = new DomainList(MBcar(MBnth(data,
                                            MESSAGE_RECEIVER_POSITION)));
        type = RESPONSE;
        completion = t_sexprToCompletionStatus(MBcar(MBnth(data,
					MESSAGE_STATUS_POSITION)));
        returnValues = new ReturnValueList(MBcar(MBnth(data,
                                           MESSAGE_RETVALUES_POSITION)));
        errorMessages = new ErrorMessageList(MBcar(MBnth(data,
                                             MESSAGE_ERRMSGS_POSITION)));
        MBfree(sexp);
      }

void Response::print(ostream& outStream)
{
        outStream << "Type:              " << getTypeString() << endl;
        outStream << "ID:                " << *id << endl;
        outStream << "Sender:            ";
        getSender()->print(outStream);
        outStream << endl;
        outStream << "Receiver:          ";
        getReceiver()->print(outStream);
        outStream << endl;
        outStream << "Completion Status: " << getCompletionStatusString() << endl;
        outStream << "Response to:       " << *getResponseTo() << endl;
        ReturnValueList* returnValueList = getReturnValues();
        int retValListSize = returnValueList->getSize();
        outStream << "Return list:       ";
        for (int i = 0; i < retValListSize; i++) {
            String* str = (ReturnValueString*)(*returnValueList)[i];
            if (i>0) outStream << " ";
            outStream << *str;
        }
        outStream << endl;
        ErrorMessageList* errorMessageList = getErrorMessages();
        int errMsgListSize = errorMessageList->getSize();
        //MJW outStream << "ErrorMessage:      ";
        outStream << "ErrorMessages(" << errMsgListSize << "):\n";
        for (int j = 0; j < errMsgListSize; j++) {
            String* str = (ErrorMessageString*)(*errorMessageList)[j];
            if (j>0) outStream << " ";
            outStream << *str;
        }
        outStream << endl;
    }

String *Response::makeResponseString()
      { char typeStr[256];
        char statusStr[256];
        sprintf(typeStr, "%d", type);
        sprintf(statusStr, "%d", getCompletionStatus());

        String* senderStr = makeSenderString();
        String* domainListStr = makeDomainListString(getReceiver());
        String* retValueListStr = makeReturnValueListString();
        String* errorMessageListStr = makeErrorMessageListString();

        String* dataString = new String("((");
        *dataString += *(getId());
        *dataString += *sRLBracket;
        *dataString += *senderStr;
        *dataString += *sRLBracket;
        *dataString += *domainListStr;
        *dataString += *sRLBracket;
        *dataString += typeStr;
        *dataString += *sRLBracket;
        *dataString += *(getResponseTo());
        *dataString += *sRLBracket;
        *dataString += statusStr;
        *dataString += *sRLBracket;
        *dataString += *retValueListStr;
        *dataString += *sRLBracket;
        *dataString += *errorMessageListStr;
        *dataString += *sRRBracket;

         delete(senderStr);
         delete(domainListStr);
         delete(retValueListStr);
         delete(errorMessageListStr);

        return dataString;
      }

const char *Request::getTypeString()
        { const char* ret;
          switch(this->type) {
             case MESSAGE:
               ret = MESSAGE_STRING;
               break;
             case MONOLOGUE:
               ret = MONOLOGUE_STRING;
               break;
             case REQUEST:
               ret = REQUEST_STRING;
               break;
             case RESPONSE:
               ret = RESPONSE_STRING;
               break;
             case COMMAND:
               ret = COMMAND_STRING;
               break;
          }
           return ret;
        }

String *Request::makeArgumentListString()
      { String* dataString = new String("( ");
        int n = arguments->getSize();
        for (int i = 0; i < n; i++)
            { *dataString += *(arguments->getArgument(i));
              *dataString += *sSpace;
            }
        *dataString += *sRBracket;
        return dataString;
      }

String *Request::makeDomainListString(DomainList* d)
      { String* dataString = new String("( ");
        int n = d->getSize();
        for (int i = 0; i < n; i++)
            { *dataString += *(makeDomainString(d->getDomain(i)));
              *dataString += *sSpace;
            }
        *dataString += *sRBracket;
        return dataString;
      }

Request::Request(t_sexp sexp)
      { t_sexp data = MBnth(sexp, TRANS_DATA_POSITION);
        char* idStr =     MBprint_Cstring(MBcar(MBnth(data,
                          MESSAGE_ID_POSITION)));
        char* methodStr = MBprint_Cstring(MBcar(MBnth(data,
                          MESSAGE_METHOD_POSITION)));
        id = new MessageIdString(idStr);
        method = new tml_Method(methodStr);
        (*MBfree_func)(idStr);
        (*MBfree_func)(methodStr);

        sender = new Domain(MBcar(MBnth(data, MESSAGE_SENDER_POSITION)));
        receiver = new DomainList(MBcar(MBnth(data, MESSAGE_RECEIVER_POSITION)));
        type = REQUEST;
        arguments = new ArgumentList(MBcar(MBnth(data,
		MESSAGE_ARGUMENTS_POSITION)));
        MBfree(sexp);
      }

void Request::print(ostream& outStream)
{
        outStream << "Type:              " << getTypeString() << endl;
        outStream << "ID:                " << *id << endl;
        outStream << "Sender:            ";
        getSender()->print(outStream);
        outStream << endl;
        outStream << "Receiver:          ";
        getReceiver()->print(outStream);
        outStream << endl;
        outStream << "Method:            " << *getMethod() << endl;
        ArgumentList* argList = getArgumentList();
        int argListSize = argList->getSize();
        outStream << "Arguments          ";
        for (int i = 0; i < argList->getSize(); i++) {
            String* str = (ArgumentString*)(*argList)[i];
            if (i>0) cout << " ";
            cout << *str;
        }
        outStream << endl;
    }

String *Request::makeRequestString()
      { char str[256];
        sprintf(str, "%d", type);

        String* senderStr = makeSenderString();
        String* domainListStr = makeDomainListString(getReceiver());
        String* argumentListStr = makeArgumentListString();

        String* dataString = new String("((");
        *dataString += *(getId());
        *dataString += *sRLBracket;
        *dataString += *senderStr;
        *dataString += *sRLBracket;
        *dataString += *domainListStr;
        *dataString += *sRLBracket;
        *dataString += str;
        *dataString += *sRLBracket;
        *dataString += *(getMethod());
        *dataString += *sRLBracket;
        *dataString += *argumentListStr;
        *dataString += *sRRBracket;

        delete(senderStr);
        delete(domainListStr);
        delete(argumentListStr);

        return dataString;
      }

const char *Command::getTypeString()
        { const char* ret;
          switch(type) {
             case MESSAGE:
               ret = MESSAGE_STRING;
               break;
             case MONOLOGUE:
               ret = MONOLOGUE_STRING;
               break;
             case REQUEST:
               ret = REQUEST_STRING;
               break;
             case RESPONSE:
               ret = RESPONSE_STRING;
               break;
             case COMMAND:
               ret = COMMAND_STRING;
               break;
          }
           return ret;
        }

void Command::printCommand(int c, ostream& outStream)
        { switch (c) {
            case MS_OPEN:
                 outStream << "MS_OPEN";
                 break;
            case MS_CLOSE:
                 outStream << "MS_CLOSE";
                 break;
            case MS_CLEAR:
                 outStream << "MS_CLEAR";
                 break;
            case MS_ACCEPT_DOMAINS:
                 outStream << "MS_ACCEPT_DOMAINS";
                 break;
            case MS_GET_CLIENTS:
                 outStream << "MS_GET_CLIENTS";
                 break;
            default:
                 break;
            }
        }

void Command:: print (ostream& outStream)
        {  outStream << endl << "  Command" << endl << "  -------" << endl;
          outStream << "    type:              ";
          outStream << this->getTypeString();
          outStream << endl;
          outStream << "    command:           ";
          printCommand(getCommand());
          outStream << endl;
          Message::print();
          outStream << "    arguments:         ";
          getArgumentList()->print();
          outStream << endl;
        }

Boolean Command::sendCommand()
        { Boolean ret = TRUE;
          switch (command) {
            case MS_OPEN:
              MB_connect(SERVER->getHostId(), SERVER->getPortNumber());
              if (TMB_ERROR->found()) ret = FALSE;
              MB_id(CLIENT->getUserId());
              break;

            case MS_CLOSE:
              MB_close();
              if (TMB_ERROR->found()) ret = FALSE;
              break;

            case MS_CLEAR:
              { MB_clear();
                CLIENT->getDomainList()->clear();
                Domain* domain = new Domain(CLIENT->getUserId());
                CLIENT->getDomainList()-> addDomain(domain);
                String* domainStr= makeMBDomainString(CLIENT->getDomainList());
                MB_accept(domainStr);
                if (TMB_ERROR->found()) ret = FALSE;
                delete(domainStr);
                delete(domain);
                break;
              }
            case MS_ACCEPT_DOMAINS:
              { MB_clear();
                String* domainStr= makeMBDomainString(CLIENT->getDomainList());
                MB_accept(domainStr);
                if (TMB_ERROR->found()) ret = FALSE;
                delete(domainStr);
                break;
              }
            case MS_GET_CLIENTS:
                 // to be done
                 break;

            default:
                 break;
            }
            return ret;
        }
