// $Author: jmartin $
// $Date: 2000-02-05 02:18:36 $
// $Source: /root/cvs/Rigi/rigiedit/adt/array.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

#ifndef ARRAYOBIDH
#include "array.h"
#endif
#ifndef INDENT_H
#include "indent.h"
#endif

void Array::AllocSizeErr()
{
	cerr << "Unable to Allocate less then one Object for each Array.\n";
}

void Array::IndexRangeErr() const
{
	fprintf( stderr,"Tried to index element outside Array.\n");
	abort();
}

static int memarr=0;

Array::Array( unsigned int size )
{
	FreeFunc = NULL;
	totalsize = size;
	if (totalsize<=0) AllocSizeErr();
	else {
		series = new ObjectPtr[totalsize];
		memarr += totalsize;
		register int i = totalsize;
		register ObjectPtr* vv = &series[0];
		while (i--) *vv++ = NULL;
	}
}

Array::Array( const Array& a )
{
	FreeFunc = NULL;
	register int i = a.totalsize;
	totalsize = i;
	series = new ObjectPtr[i];
	memarr += i;
	register ObjectPtr* vv = &series[0];
	register ObjectPtr* av = &a.series[0];
	while (i--) *vv++ = *av++;
}

Array::~Array()
{
	if (series != NULL) {
		Empty();
		delete[] series;
		memarr -= totalsize;
	}
}

int GetNoObjArray() { return memarr; }

// Object routines

int Array::Printout( int indent ) const
{
	ObjectPtr ob;
	int length = 0;

	for ( unsigned i=0; i<Size(); i++ )
	{
		if (length == 0) {
			cerr << indent_line(indent);
			length = indent; 
		}
		ob = At(i);     
		if (ob != NULL) {
			length = ob->Printout(length); 
			if (length > 70) length = 0;
		}
	}
	return 0;
}

unsigned int Array::Hash() const
{
	register unsigned int h = totalsize;
	register unsigned int i = totalsize;
	register ObjectPtr* vv = &series[0];
	while (i--) h^=(*vv++)->Hash();
	return h;
}

RigiBool Array::isEqual( void* a) const
{
	return ((RigiBool)(*this==*(Array*)a));
}

void* Array::Create_class()
{
	Array*  tmp = new Array(this->Size());
	return ((void*) tmp);
}

void* Array::Duplicate_class()
{
	ArrayPtr dest = new Array(*this);

	register int i = this->totalsize;
	register ObjectPtr* vv = &series[0];
	register ObjectPtr* av = &dest->series[0];
	while (i--) {
		if (*av != NULL) 
			*vv++ = (ObjectPtr)(*av)->Duplicate_class();
		else
			*vv++ = NULL;
		av++;
	}
	return((void*) dest);
}

// Collection Routines 

ObjectPtr Array::doNext( CltnPos& pos ) const
{
	ObjectPtr ob;
	
	while ((unsigned)pos.index < Capacity()) {
		unsigned i = pos.index++;
		if ((ob = series[i]) != NULL) return ob;
	}
	return 0;
}

ObjectPtr Array::doFirst( CltnPos& pos )
{
	ObjectPtr ob;
    unsigned i;
	
	pos.index = i = 0;
	while ((unsigned)pos.index < Capacity()) {
	  i = pos.index++;
	  if ((ob = series[i]) != NULL) return ob;
	}
	return 0;
}

ObjectPtr Array::doPrev( CltnPos& pos )
{
	ObjectPtr ob;
	while (pos.index >= 0) {
		int i = pos.index--;
		if ((ob = series[i]) != NULL) return ob;
	}
	return 0;
}

ObjectPtr Array::doLast( CltnPos& pos )
{
	ObjectPtr ob;
	int i;

	pos.index = i = Capacity()-1;
	while (pos.index >= 0) {
		i = pos.index--;
		if ((ob = series[i]) != NULL) return ob;
	}
	return 0;
}

void Array::reSize( unsigned int size )
{
	if (size<=0) AllocSizeErr();

	ObjectPtr* nv = new ObjectPtr[size];
	memarr += size;

	register int i = (size<=totalsize)?size:totalsize;
	register int j = (size<=totalsize)?(0):(size - totalsize);
	register ObjectPtr* nvp = &nv[j];
	register ObjectPtr* vp = &series[0];

	while (j--) *nvp++ = NULL;
	nvp = &nv[0];
	while (i--) *nvp++ = *vp++;

	delete[] series;
	memarr -= totalsize;
	series = nv;
	totalsize = size;
}

// Routines that a local to Arrays.

void Array::operator=( const Array& a )
{
	if (series != a.series) {
		delete[] series; // we need some type of garbage collection here.
		memarr -= totalsize;
		series = new ObjectPtr[totalsize=a.totalsize];
		memarr += totalsize;
		register int i = a.totalsize;
		register ObjectPtr* vv = &series[0];
		register ObjectPtr* av = &a.series[0];
		while (i--) *vv++ = *av++; 
	}
}       

RigiBool Array::operator==( const Array& a ) const
{
	if (totalsize != a.totalsize) return RigiFalse;
	else {
		register unsigned int i = totalsize;
		register ObjectPtr* vv = &series[0];
		register ObjectPtr* av = &a.series[0];
		while (i--) { 
			if (*vv == NULL) {
				if (*av != NULL) return RigiFalse; 
			}
			else {
				if (!((*vv)->isEqual((void *)*av))) 
					return RigiFalse; 
			}
			vv++, av++; 
		}
		return RigiTrue; 
	}
}

void Array::Add( int i, ObjectPtr ptr ) 
{
	while ((unsigned)i >= totalsize) {
		reSize( totalsize * CLTN_EXPANSION_FACTOR);
	}
	series[i] = ptr;
}

ObjectPtr Array::Find( ObjectPtr ptr )
{
	if (ptr == NULL) return NULL;
	for ( unsigned i = 0; i < totalsize; i++ )
		if (ptr->isEqual((void*)series[i])) return series[i];
	return NULL;
}

ObjectPtr Array::FindAddress( ObjectPtr ptr )
{
	if (ptr == NULL) return NULL;
	for ( unsigned i = 0; i < totalsize; i++ )
		if (ptr == series[i]) return series[i];
	return NULL;
}

int Array::FindIndex( ObjectPtr ptr )
{
	if (ptr == NULL) return -1;
	for ( unsigned i=0; i<totalsize; i++ )
		if (series[i] != NULL && ptr->isEqual((void*)series[i])) return i;
	return -1;
}

ObjectPtr Array::Remove( int i )
{
	if (i < 0) return NULL;

	ObjectPtr tmp = series[i];
	series[i] = NULL;
	return tmp;

}

int Array::Occurences( ObjectPtr ptr )
{
	register unsigned i = 0;
	register int j = 0;

	if (ptr == NULL) return 0;
	for (;i < totalsize; i++)
		j += ptr->isEqual((void*)series[i]);
	return j;
}

ObjectPtr Array::At( int i ) const 
{
	if ((unsigned int)i >= totalsize)
		return NULL;
	else
		return series[i];
}

void Array::Empty()
{
	for (unsigned i = 0;i < totalsize; i++) {
		if (FreeFunc != NULL) FreeFunc(series[i]);
		series[i] = NULL;
	}
}

static int compare_ob( void** a, void** b )
{
	cerr << "Comparing in Array\n";
	if (a == NULL || *a == NULL){cerr << "NULL"; return 1;}
	if (b == NULL || *b == NULL){cerr << "NULL"; return 1;}
	return (*((ObjectPtr*)a))->isEqual(*b); 
}

void Array::Sort()
{
//      qsort((char*)&series[0],totalsize,sizeof(ObjectPtr),compare_ob);
		cerr << "Array::Sort - Not implemented yet!\n";
	return;
}


// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
// Revision 1.3  1995/11/24 18:52:33  jmcdanie
// signed/unsigned compares
//
//   Revision 1.2  1995/06/12  22:06:08  jmcdanie
//   PC port
//
//   Revision 1.1  1994/06/30  20:09:02  bcorrie
//   Initial revision
//
//

