// $Author: jmartin $
// $Date: 2000-02-05 02:18:36 $
// $Source: /root/cvs/Rigi/rigiedit/adt/hash.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

// This may look like C code, but it is really -*- C++ -*-

#ifndef HASHTBLH
#include "hash.h"
#endif
#ifndef ARRAYOBIDH
#include "array.h"
#endif

extern "C" {
	void exit(int);
}

unsigned int HashTbl::hashCapacity(unsigned int size)
/*
 * 	Establish the HashTbl capacity.  Round size up to the next highest
 * 	power of two, if necessary. This routine is called by the HashTbl
 * 	constructors
 *
 */
{

	if (size==0) {
		//fprintf(stderr,"HashTbl size must be greater then 0\n");
		abort();
	}
	count = 0;
	nbits = 0;
	for (register unsigned int s=size; s!=0; s>>=1, nbits++);
	if (size == (unsigned)(1<<(nbits-1))) --nbits;
	size = 1<<nbits;
	mask = size-1;
	return size*CLTN_EXPANSION_FACTOR;   // return HashTbl table capacity 
}

HashTbl::HashTbl(unsigned int size) : table(hashCapacity(size)) {}

HashTbl::HashTbl(CollectionPtr c) : table(hashCapacity(c->Size())) {}

HashTbl::HashTbl() : table(hashCapacity(CLTN_DEFAULT_CAPACITY)) {}

HashTbl::HashTbl(const HashTbl& s) : table(s.table.Size())
{	
	count = s.count;
	mask = s.mask;
	nbits = s.nbits;
}

void HashTbl::operator=(const HashTbl& s) 
{
	count = s.count;
	mask = s.mask;
	nbits = s.nbits;
	table = s.table;
}

unsigned int HashTbl::Size() const
{ 
	return count; 
}

unsigned int HashTbl::NumElements() const
{ 
	int mycount = 0 ;
	IterDO( table,ObjectPtr,ptr )
		mycount++ ;
	ODretI
	return mycount ;
}

	
int HashTbl::h(unsigned int long K) const {
/*
 * multiplicative HashTbl function 
 *
 * Enter:
 * 	K = key to be HashTbled
 * 	
 * Returns:
 * 	HashTbl table index
 * 	
 * Knuth Vol. 3, Section 6.4, pp. 508-512
 */
	
	const unsigned int long Aw = 2654435769UL;	
//	const unsigned int long Aw = 40503;		use for 16 bit machines? 
	return ((Aw*K)>>((8*sizeof(unsigned int))-nbits)) & mask;
}

int HashTbl::findIndexOf(ObjectPtr ob) const {
/*
 * Search this HashTbl for the specified object
 *
 * Enter:
 * 	ob = pointer to object to search for
 *
 * Returns:
 *
 * 	index of object if found or of nil slot if not found
 * 	
 * Algorithm L, Knuth Vol. 3, p. 519
 */
	if(ob == NULL)
	{
		fprintf(stderr,"object does not exist. [HashTbl::findIndexOf]\n");
		abort();
	}
	//fprintf(stderr,"[HashTbl::findIndexOf] Looking for Object.\n");
	//ob->Printout(0);

	register int i;
	for (i = h(ob->Hash()); table[i]!=NULL; i = (i-1)&mask) {
		if (table[i]->isEqual((void*) ob)) {
			return i;
		}
		
	}
	
	return i;
}

// Routines that are required to be an Object

int HashTbl::Printout(int indent) const {
	table.Printout(indent);
	return 0;
}

unsigned int HashTbl::Hash() const
{
	register ObjectPtr o;
	unsigned int h = 0;
	CltnPos pos;
	while(o = (ObjectPtr)((*this).doNext(pos))) { 
		h ^= o->Hash();
	}
	return h;
}

RigiBool HashTbl::isEqual(void* p) const {
/*
	Returns RigiTrue if this HashTbl equals the specified object.
*/

	return ((RigiBool)(*this==*(HashTbl*)p));
}


// Routines for Collection Class

ObjectPtr HashTbl::doNext(CltnPos& pos) const {
	ObjectPtr ob;
	while ((unsigned)pos.index < Capacity()) {
	  int i = pos.index++;
	  if ((ob = table[i]) != NULL) return ob;
	}
	return 0;
}

ObjectPtr HashTbl::doFirst(CltnPos& pos) const {
	ObjectPtr ob;
	int i;

	pos.index = i = 0;
	while ((unsigned)pos.index < Capacity()) {
	  i = pos.index++;
	  if ((ob = table[i]) != NULL) return ob;
	}
	return 0;
}

void HashTbl::SetElem(const CltnPos& a, ObjectPtr ob)
{
	CltnPos i = a; // shuts up warning.
	Add(ob);
}

void HashTbl::reSize(unsigned int newSize)
/*
	Change the capacity of this HashTbl to newSize.
*/
{
	if (newSize <= Size()) return;
	
	//fprintf(stderr,"Resizing the Hash to = %u\n",newSize);

	Array tmp(Size());
	tmp = this->table;
	
	*this = HashTbl(newSize);
	
	AddAll(tmp);
}

// Routines that are used by hash class

void HashTbl::AddAll(const Array& tmp)  {
	register ObjectPtr      ob;
	register int		i;
	
	//fprintf(stderr,"Adding in all the contents of an Array\n");
	for(i = 0;(unsigned)i < tmp.Capacity(); i++) {
		ob = tmp.At(i);
		if(ob != NULL) {
			//fprintf(stderr,"[HashTbl::AddAll]Adding object\n");
			//ob->Printout(0);
			Add(ob); 
		}
	}
}

ObjectPtr HashTbl::Add(ObjectPtr ob)
{
/*
 * 	Add an object to this Set. The set is made larger when  it
 * 	becomes half full.
 */
	//cerr << "[HashTbl::Add] Adding the following object.\n";
	//ob->Printout(0);

	register int i = findIndexOf(ob);

	if (table[i]==NULL) {		// add new object to HashTbl 
		table[i] = ob;
		if (++count*CLTN_EXPANSION_FACTOR > Capacity()) {
			//cerr << "[HashTbl::Add] Expanding Hash From Add.";
			//cerr << "Old capacity = " << Capacity() << '\n';
			//cerr << "[HashTbl::Add] Count = " << count;
			reSize(Capacity()*CLTN_EXPANSION_FACTOR);
		}
		return NULL;		// successful add
	}
	else {
		//cerr << "[HashTbl::Add] Object already in the table.\n";
		return table[i];	// object already in HashTbl
	}
}

ObjectPtr HashTbl::Remove(ObjectPtr ob) {
/*
 * remove object from HashTbl
 *
 * Enter:
 * 	ob = reference to object to be removed
 *
 * Returns:
 * 	pointer to removed object
 *
 * Algorithm R, Knuth Vol. 3 p. 527
 */
	register int i = findIndexOf(ob);
	ObjectPtr rob = table[i];
	if (rob==NULL)
    {
            return NULL;
	}
	else
    {
		register int j,r;
		while (RigiTrue) {
			table[j=i] = NULL;
			do {
				i = (i-1)&mask;
				if (table[i]==NULL) {
					count--;
					return rob;
				}
				r = h(table[i]->Hash());
			} while ((i<=r&&r<j) || (r<j&&j<i) || (j<i&&i<=r));
			table[j] = table[i];
		}
	}
}

ObjectPtr HashTbl::Find(ObjectPtr ob) const {
	//fprintf(stderr,"Entered HashTbl::Find()\n");
	//fprintf(stderr,"Looking for Object :\n");
	//if(ob != NULL)
		//ob->Printout(0);
	//else
		//fprintf(stderr,"NULL OBJECT\n");
	//fprintf(stderr,"\n";

	unsigned i = findIndexOf(ob);

	if(i < 0 || i > table.Size()) {
		//fprintf(stderr,"The Index Was outside the table. [HashTbl::find]\n");
		return NULL; 
	}
	else 		
		return table[i]; 
}

RigiBool HashTbl::Includes(ObjectPtr ob) const {
	
	return ((RigiBool)(occurrencesOf(ob) != 0));
}

RigiBool HashTbl::operator==(const HashTbl& s) const {
/*
 * 	Return RigiTrue if the specified HashTbl equals this HashTbl.
 */

	if (count!=s.count) return RigiFalse;
	for (register int i=0; (unsigned)i< Capacity(); i++) {
		if (table[i]!=NULL && !s.Includes(table[i]))
		    return RigiTrue;
	}
	return RigiFalse;
}

void HashTbl::operator-(const HashTbl& s)
/*
  Returns a HashTbl of all of the objects that are contained in this
  HashTbl but not in the specified HashTbl.
*/

{
	HashTbl tmp;
	tmp = *this;

	int i;
	for (i=0; (unsigned)i < tmp.Capacity(); i++) {
		if ((tmp.table[i]!=NULL) && s.Includes(tmp.table[i])) {
			Remove(tmp.table[i]); 
		}
	}

	for (i=0; (unsigned)i<s.Capacity(); i++) {
		if ((s.table[i]!=NULL) && !(tmp.Includes(s.table[i])) ) {
			Add(s.table[i]);
		}
	}

}
void HashTbl::operator^(const HashTbl& s)
/*
  Removes all elements that are in table s from this table.
*/

{

	HashTbl tmp;
	tmp = *this;

	for (register int i=0; (unsigned)i<tmp.Capacity(); i++) {
		if ((tmp.table[i]!=NULL) && s.Includes(tmp.table[i])) {
			Remove(tmp.table[i]); 
		}
	}
}

void HashTbl::operator&(const HashTbl& s)
/*
  Returns a HashTbl of all objects that are in both this HashTbl and
  the specified HashTbl.
*/
{
	Array tmp = table;

	for (register int i=0; (unsigned)i<tmp.Capacity(); i++) {
		if ((tmp[i]!=NULL) && !(s.Includes(tmp[i]))) 
			Remove(tmp[i]);		
	}
}

void HashTbl::operator|(const HashTbl& s)
/*
  Returns a HashTbl of all objects that are in either this HashTbl
  or the specified HashTbl.
*/
{
	AddAll(s.table);
}

unsigned int HashTbl::occurrencesOf(ObjectPtr ob) const {
/*
	Return the number of occurences of the specified object
	in this HashTbl (either 0 or 1).
*/
	if(ob == NULL) return 0;
	if (table[findIndexOf(ob)]!=NULL) return 1;
	else return 0;
}

ObjectPtr HashTbl::findObjectWithKey(ObjectPtr ob) { 
	return table[findIndexOf(ob)]; 
}

 

	



// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
// Revision 1.3  1995/11/24 18:53:48  jmcdanie
// signed/unsigned compares.
//
//   Revision 1.2  1995/06/12  22:06:12  jmcdanie
//   PC port
//
//   Revision 1.1  1994/06/30  20:09:09  bcorrie
//   Initial revision
//
//

