// $Author: jmartin $
// $Date: 2002-11-10 22:16:58 $
// $Source: /root/cvs/Rigi/rigiedit/adt/timer.h,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

// This may look like C code, but it is really -*- C++ -*-

#ifndef TIMER_H
#define TIMER_H
#if !defined (_MSC_VER) && !defined (__EMX__) && !defined(__WIN32__)
/*
 *	
 *	Written by
 *
 *	Bruce Eckel
 *	School of Oceanography; WB-10
 *	University of Washington
 *	Seattle, WA 98195
 *
 *	As part of the DAIMS project.  For more information, contact:
 *
 *	Tom Keffer
 *	keffer@ocean.washington.edu
 */

#include <sys/time.h>
#include <sys/resource.h>
#include <stream.h>
#include "chartype.h"
#include "array.h"

/*
 *  ++ class timer: measures use of system resources
 *
 *  history:
 *  	27 Apr 88	Bruce Eckel	Creation date
 * 	29 Apr 90	Craig Sinclar   Included rigi library
 *
 *  details:
 * 	   Whenever you want to measure a point, use the "mark"
 * 	    function.  Everything will be dumped out at the end when you're
 * 	    done.
 */


class MarkPoint : private RigiObject {
	float	 	seconds;
    	CharType	message;
  public :
	MarkPoint(float secs,char * msg)
	  { seconds = secs;
	    message.Assign(msg); }
	char* Message() const
	  { return(message.string()); }

	float Seconds() const
	  { return(seconds); }
};

class MarkPointArray : public Array {
	int 	slot;
    public:
     	MarkPointArray(unsigned int size = CLTN_DEFAULT_CAPACITY) : Array (size)
	  { slot = 0; }
        ~MarkPointArray();

   // Routines specific to this class
	void Create(float secs, char* msg)
	  { MarkPoint* new_marker = new MarkPoint(secs, msg);
	    Array::Add(slot++, (ObjectPtr)new_marker); }

	char* At(int location)
	  { return(((MarkPoint*) Array::At(location))->Message()); }

};

class Timer {
    MarkPointArray events;
    void timer_(float& secs);
  public:
	Timer() {};
	Timer(int size) : events(size) {};
	~Timer() {};

	void Mark();
    	void Mark(char * msg);
	friend ostream& operator<<(ostream&, Timer&);
};
#endif
#endif

// ------------------------------------------------------------------------
//
// $Log: not supported by cvs2svn $
// Revision 1.3  1999/06/07 17:35:04  jmartin
// changes for compilation under OS/2
//
// Revision 1.2  1994/06/30 17:41:05  bcorrie
// Fixed up RCS header stuff
//
//
