#include <stdlib.h>
#include <tk.h>

#include "rigierr.h"
#include "cl_bckstore.h"
#include "environ.h"
// For now, use a single pixmap (defined globally).
static int numFreeBS = -1;             // number of free backstores

// BckStore constructor
// If not all of the backstores have been allocated, assign one
// to this instance.
BckStore::BckStore(Display *display, Drawable d,
    int w, int h, int pixdepth, void *cdata)
{
    canw = w;
    canh = h;
    depth = pixdepth;
    dpy = display;
    drawable = d;
    clientData = cdata;
    pixmap = NIL;

    if (numFreeBS < 0) {
        char *c = EnvGet("NUMBACKSTORES");
        numFreeBS = (c == NIL)? 2 : atoi(c);
    }
}

BckStore::~BckStore() { BckStore::Unmap(); }

Pixmap BckStore::Map() {
    if (numFreeBS && pixmap == NIL) {
        numFreeBS--;
        pixmap = Tk_GetPixmap( dpy, drawable, canw, canh, depth);
        Mess("BckStore::Map %d",pixmap);
    }
    return (pixmap == NIL)? drawable : pixmap;
}

void BckStore::Unmap() {
    if (pixmap != NIL) {
        Mess("BckStore::Unmap %d",pixmap);
        Tk_FreePixmap(dpy, pixmap);
        pixmap = NIL;
        numFreeBS++;
    }
}

Pixmap BckStore::GetPixmap() const { return (pixmap == NIL)? drawable : pixmap; }

RigiBool BckStore::IsMapped() const {
    return (pixmap == NIL)? RigiFalse : RigiTrue;
}

void *BckStore::GetClientData() const { return clientData; }

int BckStore::NumFreeBackStore() const { return numFreeBS; }
