// $Author: jmcdanie $
// $Date: 1996-04-30 23:11:20 $
// $Source: /root/cvs/Rigi/rigiedit/cl_domain.h,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------


#ifndef DOMAIN_H
#define DOMAIN_H

#ifndef NUMARRAYH
#include "adt/numarray.h"
#endif
#ifndef CHARARRAYH
#include "adt/chararray.h"
#endif
#ifndef TK_COLORMAP_H
#include "tk_colormap.h"
#endif

class GraphNode;
class GraphArc;
class AttributeSet;
class AvlPos;

// So what is a domain? It describes the environment that the editor
// lives in. It consists of nodes, arcs, and attributes. Along with
// these objects, things like colors, bitmap files etc. are associated.

class RigiDomain : public RigiObject
{
private:

    CharType domain;              // The name of this domain

    AttributeSet *arcset;         // mapping, arcname -> arctype
    AttributeSet *nodeset;        // mapping, nodename -> nodetype

    // Domain information for the arcs in the domain

    int num_arctypes;             // number of arctypes in the domain
    CharArray arcnames;           // mapping arctype -> arcname
    AttributeSet *arcattrs;       // attribute set to associate with arctypes
    NumArray srctypes;            // src node type restriction for arctypes
    NumArray dsttypes;            // dst node type restriction for arctypes

    // Domain information for the nodes in the domain

    int num_nodetypes;            // number of nodetypes in the domain
    CharArray nodenames;          // mapping, nodetype -> nodename
    CharArray iconfiles;          // name of file for each icon
    AttributeSet *nodeattrs;      // attribute set to associate with nodetypes

    // Color information for the arcs and nodes in the domain.

    RigiColor *colors;            // Dynamic array of colors
    int ncolors;                  // Number of colors required

    // Private functions to this class

    int ReadArcInfo();
    int ReadNodeInfo();
    int ReadAttrInfo();
    void ReadRCLInfo();
    int ReadColorInfo();

public:

    RigiDomain(char *name = NULL);
    ~RigiDomain();

    char*       Name() const;

    int         NumNodeTypes() const;
    char*       NodeNameFromType( int type );
    int         NodeTypeFromName( char *name );
    int         NewNodeType( char *name );
    int         ChangeNodeTypeName( char *old, char* );
    char*       IconFileName( int type );
    void        SetNodeColor( int type, RigiColor* );
    int         GetNodeColor( int type, RigiColor* );

    int         NumNodeAttrs();
    const char* FirstNodeAttr(AvlPos*&);
    const char* NextNodeAttr(AvlPos*&);
    int         IsNodeAttr( int type, char* name );
    void        SetNodeAttr( GraphNode *n, char* name, char* value );
    char*       GetNodeAttr( GraphNode *n, char* name );

    int         NumArcTypes() const;
    int         ArcSrcType( int type );
    int         ArcDstType( int type );
    char*       ArcNameFromType( int type );
    int         ArcTypeFromName( char *name );
    int         NewArcType( char *name );
    int         ChangeArcTypeName( char *old, char* );
    void        SetArcColor( int type, RigiColor *col );
    int         GetArcColor( int type, RigiColor *col );

    int         NumArcAttrs();
    const char* FirstArcAttr(AvlPos*&);
    const char* NextArcAttr(AvlPos*&);
    int         IsArcAttr( int type, char* name );
    void        SetArcAttr( GraphArc *a, char *name, char *value );
    char*       GetArcAttr( GraphArc *a, char *name );

    // Equivalence and assignment operators

    RigiBool operator==(const RigiDomain& d) const ;
    void operator=(const RigiDomain& d) ;

    // Routines required by RigiObject class and all subclasses

    virtual int Printout(int) const;
    virtual unsigned int Hash() const;                 
    virtual RigiBool isEqual(void* a) const;
    virtual void* Duplicate() { return (NULL); } // not implemented
    virtual void* Create() { return (NULL); } // not implemented
    virtual void  Delete_type(ObjectPtr) {} // not implemented
};

#endif

