//
// cl_flag.h - Class for an extensible bit flag set.
//

// $Author: jmartin $
// $Date: 1999-06-11 14:36:52 $
// $Source: /root/cvs/Rigi/rigiedit/cl_flag.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef FLAG_H
#define FLAG_H

#ifndef GRAPHERR_H
#include "grapherr.h"
#endif

static const int BitsPerElement = 8 * sizeof(int);
static const int ElementAlloc = 1; // min num elements allocated

class BitSet {
private:
    int numBits;
    int numElements;
    int numBitsAlloc;
    unsigned int *elements;

public:
    BitSet();
    BitSet(const int size);
    ~BitSet();

    // Access functions that set and clear the bit fields, determine the size of
    // the bit set, and determine the state of the various fields.

    int           Size() const {return numBits;}
    void          SetSize(const int size);
    unsigned char GetBit(const int bit) const;
    void          SetBit(const int bit);
    void          ClearBit(const int bit);
    void          ClearAll();
    void          SetAll();
    void          Print(char *str) const;
    void          operator=(const BitSet&);
};


inline unsigned char BitSet::GetBit(const int bit) const {
    unsigned int rtn = 0;
    if (bit >= numBits || bit < 0) {
        mayday( "BitSet::GetBit() - bit %d not in BitSet of size %d", 
                bit, numBits);
    } else {
        int elementIdx = bit / BitsPerElement;
        int bitIdx = bit % BitsPerElement;
        rtn = (elements[elementIdx] & (1 << bitIdx))? 1 : 0;
    }
    return (unsigned char)rtn;
}


#endif

