//
// ARCS
//

// $Author: jmcdanie $
// $Date: 1996-08-23 23:12:04 $
// $Source: /root/cvs/Rigi/rigiedit/cl_grapharc.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_grapharc.C,v 1.37 1996-08-23 23:12:04 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef GRAPHARC_H
#include "cl_grapharc.h"         // Graph arc class
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Graph node class
#endif
#ifndef ATTRIBUTE_H
#include "cl_attribute.h"
#endif
#ifndef GRAPHERR_H
#include "grapherr.h"           // Want to include low level stdio error funcs
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"	// Used in computing contributor arcs
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"
#endif


#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"          // Global instance of editor's state
#endif

static int narcs=0;
int GetNoGraphArcs() { return narcs; }  // HACK, for mem debugging MJW

GraphArc::GraphArc( GraphNode* src, GraphNode* dst, ArcType type ) {
    narcs++;

    int id = editor->NextObjId();
    SetId( id );
    SetType(type);
    SetConsistent(1);

    vSrc = src;
    vDst = dst;
    active = NULL;
    attributes = NULL;
}


GraphArc::~GraphArc() {
    if (!this) mayday("GraphArc::~GraphArc() - NULL this in destructor");
    narcs--;

    // Delete attributes if present

    if (attributes) delete attributes;

    // Tell the source and destination of this arc to remove
    // the arc from their dependency lists.

    if (vSrc) vSrc->RemoveArc( this );
    if (vDst) if (vDst!=vSrc) vDst->RemoveArc( this );

    // What about the active arc?? MJW
}

void GraphArc::SetId( ArcId id ) {
    if ( id>(1<<21) || id<0 ) mayday("bad GraphArc id %d", id);

    info &= 0x03ff;	// keep lower 10 bits
    info |= id<<10;
    if ( id > editor->GetObjId() ) editor->SetObjId(id);
}

void GraphArc::SetType( ArcType t ) {
    if ( t<0 || t>255) mayday("bad GraphArc type %d", t);

    info &= ~0xff;	// clear lower 8 bits
    info |= t&0xff;
}

void GraphArc::SetConsistent( int c ) {
    if ( c!=0 && c!=1 ) mayday("bad GraphArc consistent flag %d", c);

    info &= ~(1<<9);
    info |= (c&1)<<9;
}

// Return the node at the opposite end of the arc as node n. If the node
// passed in is NULL, or it is at neither end of the arc, return NULL;

GraphNode* GraphArc::GetOtherNode(GraphNode* n) {
    if (!n) return NULL;
    if (GetSrc()->GetId() == n->GetId()) return GetDst();
    if (GetDst()->GetId() == n->GetId()) return GetSrc();
    return NULL;
}

static void GetLeaves( GraphNode *n, NodeSet *visited, NodeSet *leaves ) {
    if (visited->Find(n)) return;
    visited->Add( n );

    AvlPos *P;
    GraphArc *arc = n->FirstAdj( P, outLevArcs );
    if (arc) while (arc)
    {
        GetLeaves( arc->GetDst(), visited, leaves );
        arc = n->NextAdj( P, outLevArcs );
    }
    else leaves->Add( n );
}

void GraphArc::ComputeContributorList() {
    NodeSet src_leaves, dst_leaves, visited;
    AvlPos *P;

    EmptyContributorList();

    // Get all the leaves under the source and under the dst

    GetLeaves( GetSrc(), &visited, &src_leaves );
    visited.Empty();
    GetLeaves( GetDst(), &visited, &dst_leaves );

    // For each normal arc out of srcs, does it connect to one in dst?

    for ( GraphNode *n = src_leaves.First(P); n; n = src_leaves.Next(P) ) {
        AvlPos *Q;
        GraphArc *arc = n->FirstAdj( Q, outArcs );
        while (arc) {
            if (dst_leaves.Find( arc->GetDst() ))
                AddContributorId( arc->GetId() );
            arc = n->NextAdj( Q, outArcs );
        }
    }
}

int GraphArc::SetAttr( char*k, char*v ) {
    if (!attributes) attributes = new AttributeSet();
    return attributes->SetAttr( k,v );
}

char* GraphArc::GetAttr( char*key ) {
    if (!attributes) return NULL;
    return attributes->GetAttr( key );
}

int GraphArc::RemoveAttr( char*key ) {
    if (!attributes) return 0;
    char *attr = GetAttr( key );
    if (!attr) return 0;
    attributes->RemoveAttr( key );
    return 1;
}

char*   GraphArc::FirstAttr(AvlPos*&P) {
    if (!attributes) return NULL;
    return attributes->First(P);
}

char*   GraphArc::NextAttr(AvlPos*&P) {
    if (!attributes) return NULL;
    return attributes->Next(P);
}

int     GraphArc::NumAttrs() {
    if (!attributes) return 0;
    return attributes->Size();
}

int     GraphArc::PrintAttrs() {
    if (!attributes) return -1;
    attributes->Print();
    return 0;
}


void GraphArc::AddContributorId( ArcId id ) {
    char key[16], val[16];
    char *str = GetAttr( "NumContribs" );
    int num = str ? atoi(str) : 0;

    sprintf( key, "C%d", num );
    sprintf( val, "%d", id );
    SetAttr( key, val );
    sprintf( val, "%d", num+1 );
    SetAttr( "NumContribs", val );
}

int GraphArc::NumContributorIds() {
    char *str = GetAttr( "NumContribs" );
    return str ? atoi(str) : 1;
}

ArcId GraphArc::GetContributorId( int i ) {
    char key[16];
    sprintf( key, "C%d", i );
    char *str = GetAttr( key );
    return str ? atoi(str) : GetId();
}

void GraphArc::EmptyContributorList() {
    char *str = GetAttr( "NumContribs" );
    if (!str) return;

    int n = atoi(str);
    for ( int i=0; i<n; i++ ) {
        char key[16];
        sprintf( key,"C%d",i );
        RemoveAttr( key );
    }
    RemoveAttr( "NumContribs" );
}

