//
// File:    grapharc.h
// Author:  Brian Corrie
// Date:    January 17, 1991
// Project: Rigi V4.0
//

// $Author: jmcdanie $
// $Date: 1996-03-21 02:39:37 $
// $Source: /root/cvs/Rigi/rigiedit/cl_grapharc.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef GRAPHARC_H
#define GRAPHARC_H

#ifndef GRAPHDEF_H
#include "cl_graphdef.h"     // ArcId, ArcType
#endif

class GraphNode;
class AttributeSet;
class AvlPos;


class GraphArc
{
    unsigned int info;    // id, type, and consistent flag

    GraphNode* vSrc;      // Source node of the arc
    GraphNode* vDst;      // Destination node of the arc

    // Pointer to a class that contains all of the display specific
    // information in it for this arc.

    void* active;

    // Pointer to a set of attributes associated with this arc

    AttributeSet* attributes;

public:

    GraphArc( GraphNode*, GraphNode*, ArcType );
    ~GraphArc();

    void    SetId( ArcId );
    void    SetType( ArcType );
    void    SetConsistent( int );

    ArcId   GetId() const           { return info>>10; }
    ArcType GetType()               { return info&0x00ff; }
    int     GetConsistent()         { return (info>>9)&1; }

    void    SetSrc( GraphNode *n )  { vSrc = n; }
    void    SetDst( GraphNode *n )  { vDst = n; }
    GraphNode* GetSrc()             { return vSrc; }
    GraphNode* GetDst()             { return vDst; }
    GraphNode* GetOtherNode( GraphNode* );

    void    SetActiveArc( void* a )  { active = a; }
    void*   GetActiveArc()           { return active; }

    int     SetAttr( char*key, char*value );
    char*   GetAttr( char*key );
    int     RemoveAttr( char*key );
    char*   FirstAttr( AvlPos*& );
    char*   NextAttr( AvlPos*& );
    int     NumAttrs();
    int     PrintAttrs();

    void    AddContributorId( ArcId );
    int     NumContributorIds();
    ArcId   GetContributorId( int );
    void    EmptyContributorList();
    void    ComputeContributorList();
};

#endif
