//
// File:    riginode.c
// Author:  Brian Corrie
// Date:    January 16,1991
// Project: Rigi V4.0
//

// $Author: jmcdanie $
// $Date: 1996-09-11 19:19:06 $
// $Source: /root/cvs/Rigi/rigiedit/cl_riginode.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: cl_riginode.C,v 1.52 1996-09-11 19:19:06 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <math.h>          // for sqrt()

#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"	// for TextExtent (?!?)
#endif
#ifndef RIGIERR_H
#include "rigierr.h"		// error handling routines
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif                          // for GetConstraintType() ?!
#ifndef GLOBAL_H
#include "global.h"
#endif

#ifndef RIGINODESET_H
#include "cl_riginodeset.h"	// to help in sorting
#endif
#ifndef RIGIARCSET_H
#include "cl_rigiarcset.h"
#endif

static ActiveNode *GetActive( GraphNode *n )
{
    if (!n) mayday("NULL node in GetActive()");
    if (!n->GetActiveNode())
    {
        ActiveNode* an = new ActiveNode();
        n->SetActiveNode( an );
    }
    return (ActiveNode*)(n->GetActiveNode());
}

static int nnodes=0;

RigiNode::RigiNode( GraphNode* n, RigiWindow* w )
{
    nnodes++;
    node = n;
    window = w;
    position.x = 0; position.y = 0;
    nodesize = 64;

    flags = 0;
    SetDrawWithIcon( RigiTrue );

    // get/make an activenode for the node of this riginode

    ActiveNode* an = GetActive( node );
    an->Insert( this );
}

int GetNoRigiNodes() { return nnodes; } // MJW HACK for mem debugging

RigiNode::~RigiNode()
{
    nnodes--;

    if (node == NULL) mayday("NULL node in RigiNode destructor");

    ActiveNode* anp = (ActiveNode*)node->GetActiveNode();
    if ( anp != NULL )
    {
    RigiNode* rnp = anp->Remove(this);
    while ( rnp != NULL )
    {
        rnp = ((ActiveNode*)(node->GetActiveNode()))->Remove( this );
    }
    }
}

void RigiNode::GetExtents( Point2d& minp, Point2d& maxp ) const 
{
    int xoffset,yoffset;
    int nsize = NodeSize();
    int textoffset;

    // Get the position of the node, and calculate the extents of the node.
    // Then convert the node positions to Window space, so we can do the
    // offsets for the text, as it is non-scalable.

    // First, get the extents of the icon
    minp = GetPosition();
    maxp.x = minp.x + nsize;
    maxp.y = minp.y + nsize;

    // Change extents if necessary, based on text size that goes with node

    window->Window()->TextExtent( GetNode()->GetLabel(),
        &xoffset, NULL, &yoffset );
    textoffset = xoffset/2 + 1;
    if (textoffset > nsize/2)
    {
        minp.x = minp.x - textoffset + nsize/2;
        maxp.x = minp.x + xoffset;
    }
    maxp.y += yoffset;
}

int  RigiNode::GetState() const
{
    return ((ActiveNode*)(node->GetActiveNode()))->GetState();
}

void RigiNode::InsertArc( RigiArc* a )
{
    RigiNode *src = a->GetSrc(), *dst = a->GetDst();

    if (!src || !dst)
        mayday("attempt to insert rigiarc %d when undefined!",a);

    // Self-arcs go into the src (outgoing) list only

    if (src==this)
        srcarcs.Insert(a,dst);
        else dstarcs.Insert(a,src);
}

RigiArc* RigiNode::RemoveArc(RigiArc* a) {
    RigiArc *r;

    if (a->GetSrc()==this)
        r = srcarcs.Remove(a,a->GetDst());
    else
        r = dstarcs.Remove(a,a->GetSrc());

    if (!r)
        RigiError(ERRORFLAG, "Unable to remove arc %s -> %s",
            a->GetArc()->GetSrc()->GetLabel(),
            a->GetArc()->GetDst()->GetLabel());
    return r;
}

RigiArc* RigiNode::FindArc( RigiNode *rn )
{
    // This func used only for outgoing arcs -> search srcarcs
    return srcarcs.Find(rn);
}

RigiArc* RigiNode::FirstArc( AvlPos*& P, ArcDirection dir )
{
    if (dir==ARC_IN)  return dstarcs.First(P);
    if (dir==ARC_OUT)  return srcarcs.First(P);

    RigiArc *ra = srcarcs.First(P);
    return ra ? ra : dstarcs.First(P);
}

RigiArc* RigiNode::NextArc( AvlPos*& P, ArcDirection dir )
{
    if (dir==ARC_OUT) return srcarcs.Next(P);
    if (dir==ARC_IN)  return dstarcs.Next(P);

    // For IN/OUT, just grab next with P, set srcarcs is immaterial
    RigiArc *r1 = srcarcs.Last(P);
    RigiArc *r2 = srcarcs.Next(P);
    if (r2) return r2;

    // Must figure out which set we were actually iterating through
    if (r1->GetSrc()==this) return dstarcs.First(P);
    return NULL;
}

int RigiNode::ArcNum( ArcDirection dir )
{
    if ( dir == ARC_OUT ) return srcarcs.Size();
    if ( dir == ARC_IN  ) return dstarcs.Size();
    return srcarcs.Size() + dstarcs.Size();
}

void RigiNode::Draw()
{
    if ( this == NULL ) mayday( "NULL self in RigiNode::Draw()" );

    window->DrawNode( this );
    return;
}

RigiBool RigiNode::isVisible()
{
    // Check all of its own filters, and those of its rigiwindow
    return window->CheckNode( this );
}

void RigiNode::Move( Point2d p )
{
    if ( this == NULL ) mayday( "NULL self (this) in RigiNode::Move" );

    // If the window is an overview window, don't allow movement in y

    Point2d q = GetPosition();
    if ( window->GetType() == WIN_OVERVIEW ) p.y = q.y;

    // If movement is constrained don't allow movement in that axis
    switch ( editor->GetConstraintType() )
    {
        case CON_NONE	: break;
        case CON_HORIZ	: p.x = q.x; break;
        case CON_VERT	: p.y = q.y; break;
        default		: break;
    }

    if ( window->GetType() != WIN_OVERVIEW )
    {
        // Update the NDC position of the source node

        NDC_Point2d ndcp = window->World2NDC2d( p );
        node->SetNDCPosition( ndcp );
    }

    SetPosition( p );
}

// return node in list with minimum vertical position
RigiNode* vert_min( RigiNodeList& list )
{
    RigiNode *node, *temp;
    Point2d p1, p2;

    node = list.First();
    temp = list.Next();
    while ( temp != NULL )
    {
        p1 = node->GetPosition();
        p2 = temp->GetPosition();
        if (p1.y > p2.y) node = temp;
        temp = list.Next();
    }
    return node;
}
        
// return node in list with minimum horizontal position
RigiNode* horiz_min( RigiNodeList& list )
{
    RigiNode *node, *temp;
    Point2d p1, p2;

    node = list.First();
    temp = list.Next();
    while ( temp != NULL )
    {
        p1 = node->GetPosition();
        p2 = temp->GetPosition();
        if (p1.x > p2.x) node = temp;
        temp = list.Next();
    }
    return node;
}
        
// return node in list with minimum "diagonal" position

RigiNode* diag_min( RigiNodeList& list )
{
    return vert_min( list );
}
        
// Sorts the nodes in itself according to their
// relative vertical positionings 

void RigiNodeList::VertSort()
{
    RigiNodeList templist;
    RigiNode* tmpnode;
    RigiNode* min;

    while ( GetNoItems() != 0 ) {
        min = vert_min( *this );
        tmpnode = Remove( min );
        templist.Insert( tmpnode );
    }

    while ( templist.GetNoItems() != 0 ) {
        tmpnode = templist.First();
        templist.Remove( tmpnode );
        Insert( tmpnode );
    }
}

// Sorts the nodes in itself according to their
// relative horizontal positionings 

void RigiNodeList::HorizSort()
{
    RigiNodeList templist;
    RigiNode* tmpnode;

    while ( GetNoItems() != 0 ) {
        tmpnode = Remove( horiz_min( *this ) );
        templist.Insert( tmpnode );
    }

    while ( templist.GetNoItems() != 0 ) 
        Insert( templist.Remove( templist.First() ));
}

// Sorts the nodes in itself according to their
// relative diagonal positionings 

void RigiNodeList::DiagSort()
{
    RigiNodeList templist;
    RigiNode* tmpnode;

    while ( GetNoItems() != 0 ) {
        tmpnode = Remove( diag_min( *this ) );
        templist.Insert( tmpnode );
    }

    while ( templist.GetNoItems() != 0 ) 
        Insert( templist.Remove( templist.First() ));
}

// Sorts the nodes in itself according to their
// relative grid positionings.  If the nodes are approximately
// in the grid to start with, this will maintain their positionings. 

    static int rnf_vert( RigiNode*r1, RigiNode*r2 )
    { return (r1->GetPosition()).y - (r2->GetPosition()).y; }

    static int rnf_hori( RigiNode*r1, RigiNode*r2 )
    { return (r1->GetPosition()).x - (r2->GetPosition()).x; }

void RigiNodeList::GridSort()
{
    RigiNode *rn, *r;
    int size;

    RigiNodeSet vertset, horiset;
    AvlPos *PV, *PH;

    // First of all, sort according to vertical positioning
    while ( GetNoItems() ) vertset.Insert( RemoveFirst(), rnf_vert );

    // Now get the dimensions of the grid and go through for each row,
    // sorting according to horizontal positioning

    double ssize = sqrt((double)vertset.Size());
    size = (ssize!=(double)((int)(ssize))) ? (int)(ssize) + 1 : (int)ssize;

    for ( rn = vertset.First(PV); rn; rn = vertset.Next(PV) )
    {
        horiset.Insert( rn, rnf_hori ); 
        if ( horiset.Size()==size )
        {
            for ( r=horiset.First(PH); r; r=horiset.Next(PH) ) Insert(r);
            horiset.Empty();
        }
    }
    for ( r=horiset.First(PH); r; r=horiset.Next(PH) ) Insert(r);

    // Finally, reverse the darn list
    Reverse();
}

