// $Author: jmartin $
// $Date: 2002-09-27 00:27:42 $
// $Source: /root/cvs/Rigi/rigiedit/cl_tk_window.h,v $
// $Locker:  $
// $State: Exp $

//------------------------------------------------------------------------

#ifndef TKWINDOW_H
#define TKWINDOW_H

#include <tcl.h>
#include <tk.h>

#ifndef BCKSTORE_H
#include "cl_bckstore.h"
#endif
#include "cl_graphdef.h"
#include "cl_rigidef.h"
#include "cl_base_window.h"
#include "adt/chartype.h"

// Window class that is specific to the Tk window system. It contains the
// information that is needed to update a window.

class TKWindow : public BaseWindow
{
public:

    TKWindow( BaseWindow*, int, int, int, int, int, int );
    TKWindow( TKWindow& );
    ~TKWindow();

    char*  GetTitle();
    void   SetTitle(char*);

    // Set normal and wait cursors for the generic window
    void NormalCursor();
    void WaitCursor();

    // Message bar manipulation routines
    void SetMessage( char *text );
    void ResetMessage();

    // functions for retrieving state info of any generic window.
    void Restack(int order, BaseWindow *otherWin);
    void DestroyChildWindows();
    void Hide();
    void Show();
    void Location( int& x, int& y );
    void SetLocation( int x, int y );
    void SetGeometry( int x, int y, int w, int h );

    void SetFont( char* fontname );
    char *GetFont();

    int  Iconified();
    void SetIconified( int );

    // Scrollbar and coordinate translation

    void    SetScrollbarStartX( int scrollstart );
    void    SetScrollbarStartY( int scrollstart );
    int     ScrollbarStartX() { return winOrig.x; }
    int     ScrollbarStartY() { return winOrig.y; }
    inline  Point2d  Window2Canvas( Point2d );
    inline  Point2d  Canvas2Window( Point2d );
    inline  Point    Window2Canvas( Point );
    inline  Point    Canvas2Window( Point );

    // Dimensions

    int  Depth();
    void FrameSize( unsigned short& w, unsigned short& h );
    void SetFrameSize( int w, int h );
    void GetDrawSize( int*w, int*h );

    // basic generic window drawing operations
    void Clear();
    void DrawIcon( Point p, int size, int type, int active );
    void DrawBox( Point sp, Point dp, int thick, int type, int dpyOnly );
    void DrawLine( Point sp, Point dp, int width, int style );
    void DrawString( char* str, Point p );
    void DrawString( char* str, Point p, TextLocation loc );
    void TextExtent( char *str, int *w, int *h, int *a);

    // Drawing, batching operations
    int  Drawing();
    void Drawing(int);
    void DrawSynch( RigiBool );
    void BunchRect( RigiBool );
    void BunchLine( RigiBool );

    void Batch( RigiBool );
    void BackStore( RigiBool );
    int  BackStore();
    void Restore( int x, int y, int w, int h );
    void RestoreWindow();
    void FreeStore();

    // Get status of shift keys
    int IsShifted();

private:

    void InitExtent();
    void UpdateExtent(int x, int y, int w, int h);
    void CheckMappings();

    Tk_Window tkwin;        // TK window for this TKWindow
    Tk_Window canvaswin;    // TK window for the canvas
    CharType title;         // Title of the window
    char iconified;         // Flag for iconified window
    int winx, winy;         // pos of window on screen
    int winh, winw;         // Height and width of window
    int tkcanh, tkcanw;     // Height and width of canvas at last setting
    Point2d  winOrig;       // Loc of win within canvas in win coords
    BckStore *pBackStore;   // Unmapped backing store
    int canw, canh;         // max canvas width, height (in RigiEnv)
    RigiBool drawBatch;     // draw to backstore and then display
    RigiBool drawWindow;    // entire window to be drawn at once
    RigiBool drawRealTime;  // drawing in real time
    struct {
        int x, y, w, h;     // draw set extent (for refresh)
    } dsExtent;

};

// inline functions

Point2d TKWindow::Window2Canvas( Point2d p ) {
    p.x += winOrig.x;
    p.y += winOrig.y;
    return p;
}

Point2d TKWindow::Canvas2Window( Point2d p ) {
    p.x -= winOrig.x;
    p.y -= winOrig.y;
    return p;
}

Point TKWindow::Window2Canvas( Point p ) {
    p.x += winOrig.x;
    p.y += winOrig.y;
    return p;
}

Point TKWindow::Canvas2Window( Point p ) {
    p.x -= winOrig.x;
    p.y -= winOrig.y;
    return p;
}


#endif

