//
//

// $Author: jmcdanie $
// $Date: 1996-05-25 01:19:05 $
// $Source: /root/cvs/Rigi/rigiedit/connect.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: connect.C,v 1.21 1996-05-25 01:19:05 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif

#ifndef CONNECT_H
#include "connect.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
//#ifndef HASHNODELIST_H
//#include "cl_hashnodelist.h"
//#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif

const int STRONG = 0;
const int WEAK = 1;

void ConnectionSelect( RigiWindow* win, int coupling, float thresh )
{
    NodeSet visited;
    RigiNode *rn; AvlPos *P;

//    Mess("Coupling is %s threshold = %f",
//      (coupling? "WEAK":"STRONG"), thresh);

    // For every arc in this window determine if it is beyond the coupling
    // threshold specified by THRESH and COUPLING. If so add it to the selection
    // list.

    RigiArc *ra; AvlPos *Q;
    for ( rn=win->FirstNode(P); rn; rn=win->NextNode(P) )
    for ( ra = rn->FirstArc(Q,ARC_OUT); ra; ra = rn->NextArc(Q,ARC_OUT) )
    {
        // Get the actual arc for this rigiarc. If the coupling is WEAK, then
        // we have to do a bit of extra work. Because we are just using arcs
        // in this window, any node not connected to another will not be
        // checked for coupling strength. This is OK for high coupling, but
        // for weak coupling, a disconnected node should be selected. This is
        // handled by marking all the nodes we check for coupling via the arcs
        // as visited. After the arcs are done, we add any that haven't been
        // visited to the selection list.

        GraphArc* ap = ra->GetArc();
        if ( coupling == WEAK )
        {
            visited.Add( ap->GetSrc() );
            visited.Add( ap->GetDst() );
        }

        // If it is a COMPOSITE arc and the coupling strength is beyond the
        // threshold then add the source and destination nodes to the select
        // list.

        float num = (float)(ap->NumContributorIds());
        if( (num >= thresh && coupling == STRONG) ||
                (num <= thresh && coupling == WEAK  )     )
        {
            editor->SelectNode(ra->GetSrc());
            editor->SelectNode(ra->GetDst());
        }
    }

    // Check the nodes in this window for those that were not visited if
    // we are doing WEAK coupling.

    if ( coupling == WEAK )
    for ( rn = win->FirstNode(P); rn; rn = win->NextNode(P) )
        if ( !visited.Find(rn->GetNode()->GetId()) )
            editor->SelectNode( rn );
}

static NodeSet component;
static NodeSet winnodes;

void VisitCoupledComponent( GraphNode* np, float thresh )
{
    // Check to see if we have visited this guy already, if so just return.

    if ( component.Find(np->GetId()) ) return;

    // Not visited already, add it to the component and follow all its
    // arcs that are past the threshold.

    component.Add(np);
//    Mess("Visiting Node %s", np->GetLabel());

    // Visit the outgoing composite arcs who are coupled above the threshold
    // and whose destinations are in the WINNODES list.

    AvlPos *P;
    GraphArc* ap = np->FirstAdj( P, outArcs );
    while( ap != NULL )
    {
        int num = ap->NumContributorIds();
        if ( (float)num >= thresh && winnodes.Find(np->GetId()) )
            VisitCoupledComponent( ap->GetDst(),thresh );
        ap = np->NextAdj( P, outArcs );
    }

    // Visit the incoming composite arcs who are coupled above the threshold
    // and whose destinations are in the WINNODES list.

    ap = np->FirstAdj( P, inArcs );
    while( ap != NULL )
    {
        int num = ap->NumContributorIds();
        if ( (float)num >= thresh && winnodes.Find(np->GetId()) )
            VisitCoupledComponent( ap->GetSrc(),thresh );
        ap = np->NextAdj( P, inArcs );
    }
}

void ComponentSelect(RigiWindow* win, RigiNodeList& slist, float thresh)
{
    RigiNode *rn; AvlPos *P;

    // Want at least one node.

    if (slist.GetNoItems() < 1 )
    {
        RigiError(ERRORFLAG, "At least one node must be selected");
        return;
    }

    // Empty the hash lists. COMPONENT marks the nodes we have visited in this
    // highly coupled component, WINNODES is a copy of the nodes in the window
    // being checked for this component.

    component.Empty();
    winnodes.Empty();

    // Make a copy of this windows nodes into a hash list so we can look em up
    // quickly.

    for ( rn = win->FirstNode(P); rn; rn = win->NextNode(P) )
        winnodes.Add( rn->GetNode() );

    // Start traversing the strong component, adding to the hashlist COMPONENT

    for ( rn = slist.First(); rn; rn = slist.Next() )
    {
        VisitCoupledComponent(rn->GetNode(), thresh);

        // Add each of the nodes in COMPONENT to the selection list.

        RigiNode* winnode = win->FirstNode(P);
        while( winnode != NULL )
        {
            if ( component.Find(winnode->GetNode()->GetId()) )
                editor->SelectNode(winnode);
            winnode = win->NextNode(P);
        }

        // Empty the component list for the next search.

        component.Empty();
    }
}
