//
// $Author: jmartin $
// $Date: 2001-12-14 23:36:04 $
// $Source: /root/cvs/Rigi/rigiedit/editops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: editops.C,v 1.69 2001-12-14 23:36:04 jmartin Exp $";

//------------------------------------------------------------------------


#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef RIGISTATE_H
#include "cl_rigistate.h"               // Global instance of editor's state
#endif
#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"         // Basic graph definitions
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Class definition for graph Nodes
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef GRAPHNODELIST_H
#include "cl_graphnodelist.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"       // Class definition for RigiWindows
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"         // Class definition for RigiWindows
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"           // Class definition for RigiDomain
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef EDITOPS_H
#include "editops.h"
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"                // Basic control functions for the editor
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"           // Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"            // Utility control functions for the editor
#endif
#ifndef RIGIERR_H
#include "rigierr.h"             // Error reporting functions
#endif
#include "win_util.h"            // Window state access functions
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#include "event.h"               // Window event access for changing node names

// Operations include files
#ifndef EDIT_H
#include "edit.h"                // Edit operation function definitions
#endif

const int COMARG = 0;
const int OPARG = 1;
const int PRINTNAME = 1;
const int WIN = 2;
const int SRCID = 2;
const int NODEID = 3;
const int ARCID = 3;
const int TYPE = 2;
const int X_COORD = 3;
const int DSTID = 3;
const int Y_COORD = 4;
const int NAME = 3;
const int ARCTYPE = 4;

// -------------- utility functions ------------------------------


static RigiBool GetGraphNode(char *argv, GraphNode *&graphNode) {
    graphNode = editor->FindNode(atoi(argv));
    if (graphNode == NULL) {
        RigiError(ERRORFLAG, "Node %s not found", argv);
        return RigiFalse;
    }
    return RigiTrue;
}

static RigiBool GetRigiWindow(char *argv, RigiWindow *&rw) {
    int winNum = 0;
    if (argv != NULL && *argv) winNum = atoi(argv);
    if (winNum) {
        if ((rw = editor->GetWindowByNumber(winNum)) != NULL) return RigiTrue;
        return RigiError(ERRORFLAG, "Incorrect window (%d) specified", winNum);
    }
    return ((rw = RutilActiveWindow()) == NULL)? RigiFalse : RigiTrue;
}

static RigiBool GetGraphArc(char *argv, GraphArc *&graphArc) {
    graphArc = editor->FindArc(atoi(argv));
    if (graphArc == NULL)
        return RigiError(ERRORFLAG, "Arc %s not found", argv);
    return RigiTrue;
}

static RigiBool NumArgsOK(int argc, int argsExpected) {
    if (argc != argsExpected) {
        RigiError(ERRORFLAG, "Wrong number of arguments (%d)", argc);
        return RigiFalse;
    }
    return RigiTrue;
}

static void SetIntResult(int val, CharType &result) {
    char str[12];
    if (result.Length()) result.Append(" ");
    sprintf(str, "%d", val);
    result.Append(str);
}

static RigiBool ValidOpInWindow(char *winNum) {
    RigiWindow *rw;
    GetRigiWindow(winNum, rw);
    RigiWinType type = rw->GetType();
    if (type != WIN_GENERAL)
        return RigiError(ALERT, "This operation should not be done in this window");
    return RigiTrue;
}


// -------------- case functions ------------------------------

static int Cut(int argc, char *argv[], CharType & /*result*/) {
    // args: cmd op

    if (!NumArgsOK(argc, 2)) return 0;

    WindowSetCursor(CURSOR_WAIT);
    RigiArc* ra = editor->GetSelectedArc();
    if (ra != NULL) {
        GraphArc* graphArc = ra->GetArc();
        editor->DeselectArc();
        if (!editor->NumSelectedNodes()) {
            // if no nodes are selected, delete the selected arc.
            // first remove the arc from the canvas
            ra->Draw();

            // then delete it from the graph
            GutilDeleteArc(graphArc);
            return 1;
        }
    }
    EditCut();
    WindowSetCursor( CURSOR_NORMAL );
    return 1;
}

static int Copy(int argc, char *argv[], CharType &/*result*/) {
    // args: cmd op

    if (!NumArgsOK(argc, 2)) return 0;

    RigiWindow *rw;
    if (!GetRigiWindow("", rw)) return 0;

    WindowSetCursor(CURSOR_WAIT);
    EditCopy(rw);
    rw->Update(); 
    WindowSetCursor(CURSOR_NORMAL);
    return 1;
}

static int Paste(int argc, char *argv[], CharType &/*result*/) {
    // args: cmd op

    if (!NumArgsOK(argc, 2)) return 0;

    RigiWindow *rw;
    if (!GetRigiWindow("", rw)) return 0;

    WindowSetCursor(CURSOR_WAIT);
    EditPaste(rw);
    rw->Update(); 
    WindowSetCursor(CURSOR_NORMAL);
    return 1;
}

static int Expand(int argc, char *argv[], CharType &/*result*/) {
    // args: cmd op

    if (!NumArgsOK(argc, 2)) return 0;
    if (editor->NumSelectedNodes() != 1)
        return RigiError(ERRORFLAG, "Exactly one node must be selected");
    WindowSetCursor(CURSOR_WAIT);
    editor->DeselectArc();
    EditExpand();
    WindowSetCursor(CURSOR_NORMAL);
    return 1;
}

static int Collapse(int argc, char *argv[], CharType &result) {
    // args: cmd op type name

    if (!NumArgsOK(argc, 4)) return 0;
    if (!editor->NumSelectedNodes()) 
        return RigiError(ERRORFLAG, "At least one node must be selected");

    RigiDomain *domain = RinitGetDomain();
    RigiWindow *rw;
    if (!GetRigiWindow("", rw)) return 0;

    // determine the type of the collapsed node
    int nType;
    if (*argv[TYPE]) {
        nType = domain->NodeTypeFromName(argv[TYPE]);
    } else {
        nType = editor->CollapseNode();
    }
    if (nType < 0)
        return RigiError(ERRORFLAG, "Unknown node type");

    WindowSetCursor(CURSOR_WAIT);
    editor->DeselectArc();
    Point minp, maxp;
    rw->GetWorldExtents( &minp, &maxp, 1 );
    minp.x = (Coordinate)(minp.x + (maxp.x-minp.x)/2);
    minp.y = (Coordinate)(minp.y + (maxp.y-minp.y)/2);

    if (*argv[NAME] == 0) argv[NAME] = NULL;
    int rtn = EditCollapse(minp, argv[NAME], nType);
    WindowSetCursor(CURSOR_NORMAL);
    SetIntResult(rtn, result);
    return rtn;
}

static int NodeCreate(int argc, char *argv[], CharType &result) {
    // args: cmd op type name win x y

    if (!NumArgsOK(argc, 7)) return 0;

    RigiDomain *domain = RinitGetDomain();
    RigiWindow *rw;
    GetRigiWindow(argv[4], rw);

    // determine the type of the new node
    int nType;
    nType = (*argv[TYPE])? domain->NodeTypeFromName(argv[TYPE]) : 0;
    if (nType < 0)
        return RigiError(ERRORFLAG, "Unknown node type");

    WindowSetCursor(CURSOR_WAIT);
    if (*argv[NAME] == 0) argv[NAME] = NULL;
    RigiNode *rn = GutilNewNode(rw, argv[NAME], nType, atoi(argv[5]), atoi(argv[6]));

    // as far as I (Johannes Martin) understand it, EditUpdateAllWindows(),
    // and in particular RigiWindow::Update() which is called in there,
    // deletes all RigiNodes. Thus, rn will no longer be valid after the
    // call to EditUpdateAllWindows(). So we need to SetIntResult() 
    // _before_ calling EditUpdateAllWindows().
    SetIntResult(rn->GetNode()->GetId(), result);
    EditUpdateAllWindows();

    WindowSetCursor(CURSOR_NORMAL);
    return 1;
}

static int ArcCreate(int argc, char *argv[], CharType &result) {
    // args: cmd op type

    GraphNode *gnSrc, *gnDst;
    if (!NumArgsOK(argc, 3)) return 0;
    RigiWindow *rw;
    if (!GetRigiWindow("", rw)) return 0;
    int numNodes;
    if ((numNodes = editor->NumSelectedNodes()) != 1 && numNodes != 2)
        return RigiError(ERRORFLAG, "one or two nodes must be seleted");

    // determine the type of the new arc
    RigiDomain *domain = RinitGetDomain();
    int aType;
    aType = (*argv[TYPE])? domain->ArcTypeFromName(argv[TYPE]) : -1;
    if (aType < 0)
        return RigiError(ERRORFLAG, "Unknown arc type");

    gnSrc = gnDst = editor->LastSelectedNode();
    if (numNodes != 1) {
        AvlPos *P;
        gnSrc = editor->FirstSelectedNode(P);
        if (gnSrc == gnDst) gnSrc = editor->NextSelectedNode(P);
    }
    RigiNode *rnSrc = rw->FindRigiNode(gnSrc);
    RigiNode *rnDst = rw->FindRigiNode(gnDst);
    if (rnSrc == NULL || rnDst == NULL)
        return RigiError(ERRORFLAG, "Both arc endpoints must be in the active window");
    WindowSetCursor(CURSOR_WAIT);
    GutilNewArc(rw, aType, rnSrc, rnDst);
    WindowSetCursor(CURSOR_NORMAL);

    // single arc assumption!
    SetIntResult(gnSrc->FindArc(gnDst, outArcs)->GetId(), result);
    return 1;
}

static int Clear(int argc, char *argv[], CharType &/*result*/) {
    // args: cmd op

    if (!NumArgsOK(argc, 2)) return 0;

    WindowSetCursor(CURSOR_WAIT);
    EditClear();
    WindowSetCursor(CURSOR_NORMAL);
    return 1;
}

static int SetNodeType(int argc, char *argv[], CharType &/*result*/) {
    // args: cmd op type

    RigiDomain *domain = RinitGetDomain();

    // determine the new type
    int nType;
    nType = (*argv[TYPE])? domain->NodeTypeFromName(argv[TYPE]) : -1;
    if (nType < 0)
        return RigiError(ERRORFLAG, "Unknown node type");

    WindowSetCursor(CURSOR_WAIT);
    AvlPos *P;
    GraphNode *graphNode = editor->FirstSelectedNode(P);
    while (graphNode != NULL) {
        EventUpdateNodeType(graphNode, nType);
        graphNode = editor->NextSelectedNode(P);
    }
    WindowSetCursor(CURSOR_NORMAL);
    return 1;
}    

static int SetArcType(int argc, char *argv[], CharType &result) {
    // args: cmd op type

    RigiDomain *domain = RinitGetDomain();

    RigiArc *ra = editor->GetSelectedArc();
    if (ra == NULL)
        return RigiError(ERRORFLAG, "An arc must be selected");

    // determine the new type
    int aType;
    aType = (*argv[TYPE])? domain->ArcTypeFromName(argv[TYPE]) : -1;
    if (aType < 0)
        return RigiError(ERRORFLAG, "Unknown arc type");

    // erase previous arc and then redraw it
    WindowSetCursor(CURSOR_WAIT);
    ra->Draw();
    ra->GetArc()->SetType(aType);
    ra->Draw();
    WindowSetCursor(CURSOR_NORMAL);
    SetIntResult(ra->GetArc()->GetId(), result);
    return 1;
}    

static int ArcCreateSrcDst(int argc, char *argv[], CharType &result) {
    // args: cmd op src dst type

    GraphNode *src, *dst;
    RigiWindow *rw;
    
    if (!NumArgsOK(argc, 5)) return 0;
    if (!GetRigiWindow("", rw)) return 0;

    // convert and validate the arguments
    RigiDomain *domain = RinitGetDomain();
    int aType = domain->ArcTypeFromName(argv[ARCTYPE]);
    if (!GetGraphNode(argv[SRCID], src)) return 0;
    if (!GetGraphNode(argv[DSTID], dst)) return 0;
    if (aType < 0)
        return RigiError(ERRORFLAG, "Undefined arc type (%s)", argv[ARCTYPE]);

    // if both arc endpoints are in the active window, then we can proceed
    // as in ArcCreate, otherwise we need to call Update()
    RigiNode *rnSrc = rw->FindRigiNode(src);
    RigiNode *rnDst = rw->FindRigiNode(dst);
    if ((rnSrc == NULL) || (rnDst == NULL)) {
      // create the arc and add it to lists at src and dst
      GraphArc *graphArc = new GraphArc(src, dst, aType);
      src->InsertArc(graphArc);
      dst->InsertArc(graphArc);

      // editor initialization
      GutilAddArc(graphArc);
      rw->Update();

      SetIntResult(graphArc->GetId(), result);
    } else {
      WindowSetCursor(CURSOR_WAIT);
      GutilNewArc(rw, aType, rnSrc, rnDst);
      WindowSetCursor(CURSOR_NORMAL);

      // single arc assumption!
      SetIntResult(src->FindArc(dst, outArcs)->GetId(), result);
    }
    return 1;
}

static int ArcDelete(int argc, char *argv[], CharType &result) {
    // args: cmd op arcId

    GraphArc *graphArc;
    RigiWindow *rw;

    if (!NumArgsOK(argc, 4)) return 0;
    if (!GetRigiWindow(argv[WIN], rw)) return 0;

    // convert and validate the arcId
    if (!GetGraphArc(argv[ARCID], graphArc)) return 0;

    int arcId = graphArc->GetId();

    // Delete the arc from the graph
    GutilDeleteArc(graphArc);
    rw->RigiDisplay();
    SetIntResult(arcId, result);
    return 1;
}    
    
static int NodeDelete(int argc, char *argv[], CharType &result) {
    // args: cmd op nodeId

    GraphNode *graphNode;
    RigiWindow *rw;

    if (!NumArgsOK(argc, 4)) return 0;
    if (!GetRigiWindow(argv[WIN], rw)) return 0;

    // convert and validate the nodeId
    if (!GetGraphNode(argv[NODEID], graphNode)) return 0;

    // delete the arcs connected to the node.
    // Note that GutilDeleteArc deletes the first arc
    // from the node's arc list.
    GraphArc *graphArc;
    while ((graphArc = graphNode->FirstArc()) != NULL)
        GutilDeleteArc(graphArc);

    int nodeId = graphNode->GetId();

    // Delete the node from the graph
    GutilDeleteNode(graphNode);
    rw->RigiDisplay();
    SetIntResult(nodeId, result);
    return 1;
}

static int ArcIdByPosition(int argc, char *argv[], CharType &result) {
    // args: cmd op win x y
    Point p;
    RigiWindow *rw;
    RigiArc *ra;

    if (!NumArgsOK(argc, 5)) return 0;

    if (!GetRigiWindow(argv[WIN], rw)) return 0;
    p.x = atoi(argv[X_COORD]);
    p.y = atoi(argv[Y_COORD]);
    p.z = 0;
    p = rw->Canvas2World(rw->Window2Canvas(p));
    ra = rw->FindArc(p);

    if (ra == NULL)
        result.Assign("");
    else
        SetIntResult(ra->GetArc()->GetId(), result);
    return 1;
}

int EditOp(int argc, char *argv[], CharType &result) {
    int op;

    // must have the operation argument to start with
    if (argc < OPARG + 1) {
        op = COLLAPSE;
    } else {
        op = atoi(argv[OPARG]);
    }

    // Get the operation type and check to see if it is OK

    switch (op) {
    case CUT:
        if (!ValidOpInWindow(NULL)) return 0;
        return Cut(argc, argv, result);
    case COPY:
        if (!ValidOpInWindow(NULL)) return 0;
        return Copy(argc, argv, result);
    case PASTE:
        if (!ValidOpInWindow(NULL)) return 0;
        return Paste(argc, argv, result);
    case EXPAND:
        if (!ValidOpInWindow(NULL)) return 0;
        return Expand(argc, argv, result);
    case COLLAPSE:
        if (!ValidOpInWindow(NULL)) return 0;
        return Collapse(argc, argv, result);
    case NODE_CREATE:
        if (!ValidOpInWindow(argv[4])) return 0;
        return NodeCreate(argc, argv, result);
    case ARC_CREATE:
        if (!ValidOpInWindow(NULL)) return 0;
        return ArcCreate(argc, argv, result);
    case CLEAR:
        return Clear(argc, argv, result);
    case SET_NODE_TYPE:
        return SetNodeType(argc, argv, result);
    case SET_ARC_TYPE:
        if (!ValidOpInWindow(NULL)) return 0;
        return SetArcType(argc, argv, result);
    case ARC_CREATE_SRC_DST:
        if (!ValidOpInWindow(NULL)) return 0;
        return ArcCreateSrcDst(argc, argv, result);
    case ARC_DELETE:
        if (!ValidOpInWindow(NULL)) return 0;
        return ArcDelete(argc, argv, result);
    case NODE_DELETE:
        if (!ValidOpInWindow(NULL)) return 0;
        return NodeDelete(argc, argv, result);
    case ARC_ID_BY_POSITION:
        return ArcIdByPosition(argc, argv, result);
    case 97:
        editor->PrintArcs();
        return 1;
    case 98:
        editor->PrintNodes();
        return 1;
    case 99:
        editor->PrintNames(argv[PRINTNAME]);
        return 1;
    default:
        return RigiError(ERRORFLAG, "Unknown edit operation (%s)", argv[OPARG]);
    }
}

