// $Author: jmartin $
// $Date: 2003-03-03 14:04:54 $
// $Source: /root/cvs/Rigi/rigiedit/environ.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------
// Rigi environment configuration file handling added by G.Kacy 17 Feb 1996

#ifdef _MSC_VER
#include <direct.h>
#include <io.h>
#else
#include <unistd.h>
#endif
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef ENVIRON_H
#include "environ.h"
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef CHARTYPEH
#include "adt/chartype.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef ATTRIBUTE_H
#include "cl_attribute.h"
#endif
#ifndef CFGDATA_H
#include "cfgdata.h"
#endif


static AttributeSet vars;
static const int FileError = -1;
static const int EndOfFile = -2;
static const char *RigiStr = "$RIGI";
static const int RigiStrLen = 5;

// default configuration file name
static const char *DefaultCfgFile = "rigicfg.env";

//--------------------------------------------------------------------------
// local functions
//--------------------------------------------------------------------------


// copies fldVal to buff while substituting rigiStr with
// the value in the fldArr corresponding to CFG_RIGI.
// returns RigiFalse if buffer overflow occurs.
static RigiBool EnvSubstituteRIGI(char *fldVal, char *buff, int buffSize) {
    *buff = 0;
    char *p;
    if ((p = strstr(fldVal, RigiStr)) != NULL) {
        char *pVal = getenv("RIGI");
        if ((int)(strlen(fldVal) + strlen(pVal)) - RigiStrLen < buffSize) {
            // splice the string, replacing $RIGI with its value
            size_t prefixLen = (size_t)(p - fldVal);
            strncpy(buff, fldVal, prefixLen);           // copy prefix to $RIGI
            strcpy(&buff[prefixLen], pVal);             // copy $RIGI equivalent
            strcat(buff, &fldVal[prefixLen+RigiStrLen]);// copy suffix to $RIGI
            return RigiTrue;
        } else {
            return RigiFalse;
        }
    } else if ((int)strlen(fldVal) < buffSize) {
        strcpy(buff, fldVal);
        return RigiTrue;
    }
    return RigiFalse;
}

// Loads into buf the value associated with the id number.
// If it encounters a prefix of $RIGI, it replaces it with
// the value for RIGI.
// Returns RigiFalse if buffer is too small or id is illegal.
static RigiBool EnvGetFldVal(int id, char *buff, int buffSize)
{
    char *pVal;
    *buff = 0;
    if (id < 0 || id >= CFG_CONFIG_SIZE
        || ((pVal = EnvGet(fldArr[id].tag)) == NULL))
        return RigiFalse;
    strcpy(buff, pVal);
    return RigiTrue;
}

// Searches for a switch name in the argv array
// If no switch name is found, 0 is returned
// otherwise its position in the array is returned.
static int EnvArgExists(int argc, char *argv[], char *argName) {
  int i;
  for (i = 1; i < argc && strcmp(argName, argv[i]); i++) ;
  return (i < argc)? i : 0;
}

// Retrieves a parameter value from the argv array by its switch name.
// If no switch name is found, the name is not followed
// by a value, or the name is followed by another switch name,
// the function returns NULL.
static char *EnvGetArg(int argc, char *argv[], char *argName) {
    int i = EnvArgExists(argc, argv, argName);
    return (!i || i >= argc - 1 || *argv[i + 1] == '-')? NULL : argv[i + 1];
}

// Replaces backslashes with forward slashes
// returns a pointer to buff
static char *EnvXlateParseChar(char *buff) {
#if defined (__WINDOWS__) || defined (__OS2__) || defined(__WIN32__)
    if (buff == NULL) return NULL;
    char *pChar = buff;
    while ((pChar = strchr(pChar, PATHPARSECHAR)) != NULL)
        *pChar = UNIXPATHPARSECHAR;
#endif
    return buff;
}


// Composes the date and time string (used when writing a file)
static void EnvMakeTimeString(char *s)
{
    char timeBuff[26];
    time_t nTime;
    struct tm *ptm;

    // asctime returns the time string including '\n' - need to strip it off
    nTime = time(NULL);
    ptm = localtime(&nTime);
    strcpy(timeBuff, asctime(ptm));
    timeBuff[24] = 0;
    strcpy(s,timeBuff);
}

//
// Sets up the default configuration
//
static void EnvSetDefaultConfig(void)
{
    char *pVal;
    for (int i = 0; i < CFG_CONFIG_SIZE; i++) {
        if (i != CFG_RIGI) {
            // let the shell variables override the file variables
            if ((pVal = getenv(fldArr[i].tag)) == NULL) pVal = fldArr[i].val;
            EnvSet(fldArr[i].tag, pVal);
        }
    }
}

// read the next line in file, and pass it back it as a string in buff
// trim trailing CR's and LF's
// return the string length or FileError if error or EndOfFile if EOF
static int EnvReadLine(FILE *file, char *buff, int buffSize)
{
    int lineLen = 0;
    RigiBool eof = RigiFalse;

    // fill up buffer
    if (fgets(buff,buffSize,file) == NULL) {
        if (feof(file))
            eof = RigiTrue;
        if (ferror(file))
            return FileError;
    }

    // strip trailing line feeds, blanks, tabs and returns
    for (lineLen = strlen(buff) - 1; lineLen >= 0
        && (buff[lineLen] == '\n' || buff[lineLen] == '\r'
        || buff[lineLen] == '\t' || buff[lineLen] == ' '); lineLen--) ;

    buff[++lineLen] = 0;
    return (eof)? EndOfFile : lineLen;
}

//
// Parses the buf line and retrieves TagName and FldVal
//
static void EnvParseLine(char *buff, char *tagName, char *fldVal)
{
    int i, j;

    // extract first part before '=' (should be contiguous)
    for (i = j = 0; i < STRRESSIZE && buff[i] && buff[i] != ENVFLDPARSECHAR; i++)
        if (buff[i] != ' ' && buff[i] != '\t') tagName[j++] = buff[i];
    if (buff[i] != ENVFLDPARSECHAR) {
        *fldVal = *tagName = 0;
        return;
    }
    tagName[j] = 0;
    i++;

    // trim leading blanks from value
    for (; i < STRRESSIZE && buff[i] && (buff[i] == ' ' || buff[i] == '\t'); i++) ;
    strcpy(fldVal, &buff[i]);

}

// Check the file name specified by the scriptTag
// Return RigiTrue if the file exists and is non-empty or there is no
// file specified.
static RigiBool EnvScriptExist(char *scriptTag)
{
    struct stat statBuff;
    char *fileName = EnvGet(scriptTag);

    if (fileName == NULL) return RigiFalse;
    if (fileName[0] == 0)
        return RigiTrue;
    // get file info
    if (!stat(fileName, &statBuff)) {
        // Return true if the file exists and is non-empty
        if ((statBuff.st_mode & S_IFREG) && statBuff.st_size)
            return RigiTrue;
    }
    fprintf(stderr, "error: script file does not exist\n%s\n",fileName);
    return RigiFalse;
}


// Checks if the file exists
// Return RigiTrue if the file exists and is non-empty or RigiFalse
// if there is no file
static RigiBool EnvFileExist(char *fileName) {
    struct stat statBuff;

    // get file info
    if (!stat(fileName, &statBuff)) {
        // Return true if the file exists and is non-empty
        if ((statBuff.st_mode & S_IFREG) && statBuff.st_size)
            return RigiTrue;
    }
    return RigiFalse;
}

static RigiBool EnvCreateDir(char *dir, RigiBool writable = RigiFalse) {
  struct stat statBuff;

  if (stat(dir, &statBuff)) {
    char *d = strdup(dir);
    char *c;
    if ((c = strrchr(d, UNIXPATHPARSECHAR)) &&
	(c - d > (d[0] == UNIXPATHPARSECHAR))) {
      *c = '\0';
      if (!EnvCreateDir(d))
	return RigiFalse;
    }	
    if (
#if (defined(_MSC_VER) || defined(__WIN32__))
	mkdir(dir)
#else
	mkdir(dir, S_IFDIR|S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH)
#endif
	) {
      fprintf(stderr, "error: unable to create directory %s\n", dir);
      return RigiFalse;
    }
    stat(dir, &statBuff);
  }

  if (!S_ISDIR(statBuff.st_mode)) {
    fprintf(stderr, "error: %s is not a directory\n", dir);
    return RigiFalse;
  }
  
  if (access(dir, R_OK | X_OK | (writable ? W_OK : 0))) {
    fprintf(stderr, "error: %s permission denied for %s\n",
	    writable ? "write" : "read", dir);
    return RigiFalse;
  }

  return RigiTrue;
}


static RigiBool EnvApndFileToDir(char *dir, const char *fileTitle, char *pathFile)
{
    int n;
    // compose the full path and file name
    if (dir != pathFile) strcpy(pathFile, dir);
    n = strlen(pathFile);
    if (n && n < sizeof(DIRPATH) && pathFile[n-1] != UNIXPATHPARSECHAR)
        pathFile[n++] = UNIXPATHPARSECHAR;
    if (n + strlen(fileTitle) + 1 < sizeof(DIRPATH)) {
        strcpy(&pathFile[n], fileTitle);
        return RigiTrue;
    }
    *pathFile = 0;
    return RigiFalse;
}


// review all of configuration variables and create directories
// if required
static RigiBool EnvBldDirs()
{
    DIRPATH dir, subDir, newDir;
    char *pChar;
    int i;
    RigiBool nStatus = RigiTrue;

    for (i = 0; i < CFG_CONFIG_SIZE; i++) {
        switch (i) {
        case CFG_RIGIRCL:
        case CFG_RIGISTY:
        case CFG_RIGIURCL:
        case CFG_RIGIUSTY:
            // these are file specs with directory
            if (!EnvGetFldVal(i, dir, sizeof(DIRPATH))) {
                nStatus = RigiFalse;
                break;
            }

            // remove the file name and final parse character
            if ((pChar = strrchr(dir, UNIXPATHPARSECHAR)) == NULL) break;
            *pChar = 0;
            // check the default data against the directory structure
            if (!EnvCreateDir(dir))
	      nStatus = RigiFalse;
            break;
        case CFG_RIGIDOMAIN:
            // this is the subdirectory of RIGIINIT
            // get RIGIINIT directory path
            if (!EnvGetFldVal(CFG_RIGIINIT, dir, sizeof(DIRPATH))) {
                nStatus = RigiFalse;
                break;
            }
            // get RIGIDOMAIN subdirectory
            if (!EnvGetFldVal(i, subDir, sizeof(DIRPATH))) {
                nStatus = RigiFalse;
                break;
            }
            // concatenate domain and
	    // check the constructed path to the directory structure
            if (!EnvApndFileToDir(dir, subDir, newDir) ||
                !EnvCreateDir(newDir))
	      nStatus = RigiFalse;
            break;
        case CFG_RIGIINIT:
        case CFG_RIGILIB:
        case CFG_ICONDIR:
        case CFG_SRCDIR:
        case CFG_DBDIR:
        case CFG_DBREFDIR:
        case CFG_RIGIBIN:
        case CFG_TMPDIR:
            // check whether default directories exist
            if (!EnvGetFldVal(i, dir, sizeof(DIRPATH)) ||
		!EnvCreateDir(dir))
 	        nStatus = RigiFalse;
            break;
        default:
            // these are field values -- not directories
            // and should not build directories
            break;
        }
    }
    return nStatus;
}

static int EnvFindCfgFile(char *pathEnvFile) {
    if (EnvRead(pathEnvFile, AllParms) == RigiErr) {
        fprintf(stderr, "error: unable to access configuration file\n%s\n",
            pathEnvFile);
        return EnvError;
    }
    // check if rcl scripts are present
    if (!EnvBldDirs()
	|| !EnvScriptExist("RIGIRCL") || !EnvScriptExist("RIGISTY"))
        return EnvError;
    // save the name of the env file for editing through the
    // setup widget and for displaying in screen logo.
    EnvSet("CFGFILE",pathEnvFile);
    return EnvOK;
}

//--------------------------------------------------------------------------
// General environment variable access
//--------------------------------------------------------------------------

char *EnvGet(char *name, int asRead)
{
    char *envVal = NULL;
    static char buff[STRRESSIZE];
    if (name && (envVal = vars.GetAttr(name)) == NULL) {
        envVal = getenv(name);
    } else if (!asRead) {
        // copy buff to fldVal, replacing $RIGI if necessary
        if (!EnvSubstituteRIGI(envVal, buff, STRRESSIZE)) return NULL;
        // replace backslashes with forward slashes
        EnvXlateParseChar(buff);

        envVal = buff;
    }
    return envVal;
}

void  EnvSet(char *name)
{
    EnvSet(name, getenv(name));
}

void  EnvSet(char *name, char *val)
{
    if (name && *name && val) vars.SetAttr(name, val);
}

void EnvPrint()
{
    AvlPos *P;

    for (char *name = vars.First(P); name; name = vars.Next(P)){
        char *val = vars.GetAttr(name);
        fprintf(stderr,"%-14s %s\n",name,val?val:"<NULL>");
    }
}


char *EnvDomainDir()
{
    static CharType initdir;
    char *p;

    // Get directory path...
    // If RIGIINIT not defined use cwd.
    if ((p = EnvGet("RIGIINIT")) == NULL) p = ".";
    initdir.Assign(p);

    // Get domain subdirectory ...
    // If RIGIDOMAIN not defined use nothing.
    p = EnvGet("RIGIDOMAIN");
    if (*p != 0) {
        if (*p == '/') {
            initdir.Assign(p);
        } else {
            initdir.Append("/");
            initdir.Append(p);
        }
    }

    return initdir.string();
}

//-----------------------------------------------------------------------------
// Routines added by G.Kacy 17 February 1996
// Rigi environment configuration file handling

// Reads the environment file and sets the environment
int EnvRead(char *fileName, EnvInitMode mode) {
    char buff[STRRESSIZE], tagName[STRRESSIZE], fldVal[STRRESSIZE];
    char *pVal;
    char *rigiTag = fldArr[CFG_RIGI].tag;
    FILE *infile;

    EnvXlateParseChar(fileName);

    // save the file name , so RIGI (TK) can access the file
    EnvSet("CFGFILE", fileName);

    if ((infile = fopen(fileName,"r")) == NULL)
        return RigiErr;

    // read and parse each line
    for (;;) {
        int result = EnvReadLine(infile, buff, STRRESSIZE);
        if (result == FileError) {
            fclose(infile);
            break;
        }
        if (result == EndOfFile) {
            fclose(infile);
            return RigiOK;
        }
        if (result && *buff != ENVCOMMENTCHAR) {
            // parse the tag name and the field
            //  value from the line read from file
            EnvParseLine(buff, tagName, fldVal);

            // if this is only a reinitialization we only
            // want to restore the paths.
            if (mode == PathsOnly
                && strcmp(fldArr[CFG_RIGIINIT].tag, tagName)
                && strcmp(fldArr[CFG_RIGILIB].tag, tagName)
                && strcmp(fldArr[CFG_ICONDIR].tag, tagName)
                && strcmp(fldArr[CFG_SRCDIR].tag, tagName)
                && strcmp(fldArr[CFG_DBDIR].tag, tagName)
                && strcmp(fldArr[CFG_DBREFDIR].tag, tagName)
                && strcmp(fldArr[CFG_RIGIBIN].tag, tagName)
                && strcmp(fldArr[CFG_TMPDIR].tag, tagName))
                continue;

            // set the field value into the configuration file
            // except for the RIGI parameter itself.
            if (strcmp(tagName, rigiTag)) {
                // let the shell env variables override the file variables
                if ((pVal = getenv(tagName)) == NULL) pVal = fldVal;
                EnvSet(tagName, pVal);
            }
        }
    }
    return RigiErr;

}

// Writes the environment to file
int EnvWrite(char *fileName) {
    FILE *outfile;
    char buff[STRRESSIZE];
    char *rigiTag = fldArr[CFG_RIGI].tag;

    if ((outfile = fopen(fileName,"w")) == NULL)
        return RigiErr;

    //write file header (title + creation time)
    fprintf(outfile,"# RigiEdit configuration file\n# File:    %s\n",
        fileName);
    EnvMakeTimeString(buff);
    fprintf(outfile,"# Created: %s\n\n",buff);

    //write actual environment
    AvlPos *P;
    for (char *name = vars.First(P); name; name = vars.Next(P)) {
        char *val = vars.GetAttr(name);
        // except for the RIGI parameter, save parameter in file
        if (strcmp(name, rigiTag))
            fprintf(outfile,"%s %c %s\n", name, ENVFLDPARSECHAR, val?val:"");
    }
    fclose(outfile);
    return RigiOK;
} //EnvWrite

//
// Reads an environment file into the environment AVL set.
// Should be called before EnvGet routine.
// Searches for env file in following order:
// 1) cmd line,
// 2) RIGIUSER dir,
// 3) current working dir, and
// 4) RIGI dir
// If it fails all of these, then it constructs a default file.
// Returns: EnvOk  - initialization OK
//          EnvError - error initializing Rigi's environment
//          EnvDefault - initialization from file failed, default values assigned
//
int EnvInit(int argc, char *argv[]) {
    DIRPATH pathEnvFile, cwdEnvFile;

    // read default configuration to ensure that
    // required parameters are always found.
    EnvSetDefaultConfig();

    // get command line argument if one exists.
    char *pStr = EnvGetArg(argc, argv, "-env");
    if (pStr != NULL) {
        // read command line environment file
        strncpy(pathEnvFile, pStr, sizeof(DIRPATH));
        return EnvFindCfgFile(pathEnvFile);
    }

    // try for configuration in the current working directory
    getcwd(cwdEnvFile, sizeof(DIRPATH));
    if (!EnvApndFileToDir(cwdEnvFile, DefaultCfgFile, cwdEnvFile)) {
        fprintf (stderr,"error: invalid path specification:\n%s\n", cwdEnvFile);
        return EnvError;
    }
    // check if working environment file exists
    if (EnvFileExist(cwdEnvFile))
        return EnvFindCfgFile(cwdEnvFile);

    // try for configuration in RIGIUSER directory
    if ((pStr = getenv("RIGIUSER")) != NULL) {
        if (!EnvApndFileToDir(pStr, DefaultCfgFile, pathEnvFile)) {
            fprintf (stderr,"error: invalid path specification\n%s\n", pathEnvFile);
            return EnvError;
        }
        // check if RIGIUSER environment file exists
        if (EnvFileExist(pathEnvFile))
            return EnvFindCfgFile(pathEnvFile);
    }

    // try for configuration in RIGI root directory
    if ((pStr = getenv("RIGI")) == NULL) {
        fprintf(stderr, "error: unable to access RIGI environment variable\n");
        return EnvError;
    }
    if (!EnvApndFileToDir(pStr, DefaultCfgFile, pathEnvFile)) {
        fprintf (stderr,"error: invalid path specification\n%s\n", pathEnvFile);
        return EnvError;
    }
    // check if root environment file exists
    if (EnvFileExist(pathEnvFile))
        return EnvFindCfgFile(pathEnvFile);

    // give up and present a default configuration
    fprintf(stderr, "warning: unable to locate a configuration file;\n");
    fprintf(stderr, "         using the default configuration instead.\n");

    // save the default configuration in the RIGI directory
    if (EnvWrite (cwdEnvFile) != RigiOK)
        fprintf (stderr,"warning: unable to save the default configuration\n%s\n",
            cwdEnvFile);

    // build default directories if not present,
    // then check rcl scripts
    if (!EnvBldDirs()
	|| !EnvScriptExist("RIGIRCL") || !EnvScriptExist("RIGISTY"))
        return EnvError;
    // save the file name , so RIGI (TK) can access the file
    EnvSet("CFGFILE",cwdEnvFile);
    return EnvDefault;

}

RigiBool EnvReInit() {
    DIRPATH pathEnvFile;
    char *file = EnvGet("CFGFILE");
    if (file == NULL)
        return RigiError(ERRORFLAG, "Configuration file does not exist");

    strcpy(pathEnvFile, file);
    if (EnvFileExist(pathEnvFile)) {
        if (EnvRead(pathEnvFile, PathsOnly) == RigiErr)
            return RigiError(ERRORFLAG, "Unable to reinitialize configuration");
        return RigiTrue;
    }
    return RigiError(ERRORFLAG, "Error reading configuration file\n%s", pathEnvFile);
}

