// $Author: jmartin $
// $Date: 2002-09-27 00:27:42 $
// $Source: /root/cvs/Rigi/rigiedit/event.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------


#ifndef STREAMH
#include <stream.h>
#endif

// Include my external declarations

#ifndef EVENT_H
#include "event.h"
#endif

// Include the base classes used in this file

#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif

// Include important type and function definitions

#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

#ifndef RIGI_H
#include "rigi.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef WIN_UTIL_H
#include "win_util.h"
#endif
#ifndef EDIT_H
#include "edit.h"
#endif

static const int BOXCOLOUR = 0;		// for rubberband box
static const int BOXSIZE = 1;
static const int DELTA_NODEEDGE = 2;   // for determining if an arc is a self-arc

// Variables to keep track of the event state of the windows
static Point boxp1, boxp2, selp;
static RigiNode *seln1, *seln2;
static int origState;
static RigiBool dragged = RigiFalse;

static NodeSet selectset;

static const int NO_EVENT       = 0;
static const int PRESS          = 1;
static const int DRAG           = 2;
static const int RELEASE        = 3;

static const int NO_BUTTON      = 0;
static const int SELECT_BUTTON  = 1;
static const int ADJUST_BUTTON  = 2;

static const int MINMOVE = 4;
static const int GRANULARITY = 5;

// returns true if a new mouse event has occured when compared
// to the previous state of the mouse.
static RigiBool IsNewMouseEvent(int button, int event, Point p) {
    static int activeButton = NO_BUTTON;
    static int activeEvent = NO_EVENT;
    static Point lastp;
    int deltax, deltay;

    if (button == activeButton) {
        switch (event) {
        case RELEASE:
            activeButton = NO_BUTTON;
            activeEvent = NO_EVENT;
            return RigiTrue;
        case PRESS:
        case DRAG:
            deltax = abs(p.x - lastp.x);
            deltay = abs(p.y - lastp.y);
            if (activeEvent == PRESS &&
                (deltax >= MINMOVE || deltay >= MINMOVE)) {
                lastp = p;
                activeEvent = DRAG;
                return RigiTrue;
            }
            if (activeEvent == DRAG &&
                (deltax >= GRANULARITY || deltay >= GRANULARITY)) {
                lastp = p;
                return RigiTrue;
            }
        }
        return RigiFalse;
    }
    if (activeButton == NO_BUTTON && event == PRESS) {
        // press
        activeButton = button;
        activeEvent = event;
        lastp = p;
        return RigiTrue;
    }
    return RigiFalse;
}

static void GetSelectionSet( RigiWindow *rw ) {
    selectset.Empty();

    AvlPos *P; RigiNode *rn;
    GraphNode *n = editor->FirstSelectedNode(P);
    for ( ; n; n = editor->NextSelectedNode(P) )
        if ( rn = rw->FindRigiNode(n) ) selectset.Insert(n);
}

static void ArcMessage( RigiWindow *rw, char *m, RigiArc *ra) {
    if (!ra) {
        rw->Window()->SetMessage( m );
        return;
    }
    char msg[128];
    RigiNode *src = ra->GetSrc();
    RigiNode *dst = ra->GetDst();

    if (src==NULL||dst==NULL) mayday("NULL node in ArcSelectMessage()");
    char *s = src->GetNode()->GetLabel();
    char *d = dst->GetNode()->GetLabel();
    if (strlen(m)+strlen(s)+strlen(d)>110) return;
    sprintf(msg, "%s %d (\"%s\",\"%s\")",m,ra->GetArc()->GetId(),s,d );
    rw->Window()->SetMessage(msg);
}

static void NodeMessage( RigiWindow *rw, RigiNode *rn) {
    char buff[20];
    CharType msg("Node \"");
    msg.Append(rn->GetNode()->GetLabel());
    sprintf(buff, "\" (%d)", rn->GetNode()->GetId());
    msg.Append(buff);
    if (rn->GetState() == ACTIVE)
        msg.Append(" selected");
    else
        msg.Append(" deselected");

    rw->Window()->SetMessage(msg.string());
}

///////////////////////////////////////////////////////
// Access function to get the state of the shift key //
///////////////////////////////////////////////////////

static int shift_is_down = 0;  // Indicates if shift key is pressed
int EventShiftDown() {
    return shift_is_down;
}
void SetShift() {
    // we must go directly to keyboard since Tk does not
    // correctly report the status of the shifts when they
    // are released.
    shift_is_down = WindowIsShifted();
}

/////////////////////////////////////////////////////////
// Access functions to change the state for a RigiNode //
/////////////////////////////////////////////////////////

void EventUpdateNodeLabel( GraphNode *n, char *label ) {
    RigiNode *rn;
    if (!n) return;

    ActiveNode* an = (ActiveNode*)n->GetActiveNode();

    if (an) for ( rn = an->First(); rn; rn = an->Next() ) rn->Draw();
    editor->ChangeNodeName( n, label );
    if (an) for ( rn = an->First(); rn; rn = an->Next() ) rn->Draw();
}

void EventUpdateNodeType( GraphNode *n, int type ) {
    RigiNode *rn;
    if (!n) { mayday("NULL GraphNode in EventUpdateNodeType()!"); return; }

    ActiveNode* an = (ActiveNode*)n->GetActiveNode();

    if (an) for ( rn = an->First(); rn; rn = an->Next() ) rn->Draw();
    n->SetType( type );
    if (an) for ( rn = an->First(); rn; rn = an->Next() ) rn->Draw();
}

RigiBool EventGetSelectBox( Point& p1, Point& p2 ) {
    if (boxp1.x == IllegalPoint.x || boxp1.y == IllegalPoint.y ||
        boxp1.z == IllegalPoint.z || boxp2.x == IllegalPoint.x ||
        boxp2.y == IllegalPoint.y || boxp2.z == IllegalPoint.z )
        return RigiFalse;

    p1 = boxp1;
    p2 = boxp2;
    return RigiTrue;
}

///////////////////////
// Window Procedures //
///////////////////////

void EventDestroy( RigiWindow *rw) {
    editor->CloseWindow(rw);
}

// Process an expose event in the window over
// the region given by p,x, p.y, w, h
void EventExpose(RigiWindow* rw, Point p, int w, int h) {
     p = rw->Window2Canvas(p);
    if (!w && !h) return;
    if (!rw->BackStore()) rw->DrawStore();
    rw->Restore(p.x, p.y, w, h);
}

void EventReleaseSelect(RigiWindow *rw, Point p) {
    if (!IsNewMouseEvent(SELECT_BUTTON, RELEASE, p)) return;
    p = rw->Canvas2World(rw->Window2Canvas(p));

    if (seln1 == NULL) {
        if (dragged) {
            rw->DrawBox(boxp1, boxp2, BOXSIZE, BOXCOLOUR, 1);

            // select contained nodes
            rw->DrawSynch(RigiTrue);
            rw->BunchNode(RigiTrue);
            editor->InvertSelectionNodes(rw, boxp1, p);
            rw->BunchNode(RigiFalse);
            rw->DrawSynch(RigiFalse);

            char msg[32];
            int i = editor->NumSelectedNodes();
            sprintf( msg, "%d node%s selected", i, (i == 1)? "" : "s");
            rw->Window()->SetMessage( msg );
        }
    } else if (EventShiftDown() && !dragged) {
        // if not a move, then change selection state of node
        if (origState == ACTIVE)
            editor->DeselectNode(seln1);
        else
            editor->SelectNode(seln1);
        NodeMessage(rw, seln1);
    }
}


void EventReleaseAdjust(RigiWindow *rw, Point p) {
    if (!IsNewMouseEvent(ADJUST_BUTTON, RELEASE, p) || !dragged) return;
    p = rw->Canvas2World(rw->Window2Canvas(p));
    if (seln1 == NULL || selp.x == IllegalPoint.x || selp.y == IllegalPoint.y)
        return;

    int arctype = editor->GetArcType();
    seln2 = rw->FindNode( p );

    if (seln2 == NULL) {
        // no dst node has been chosen - erase the rubber arc line
        Point nodepoint = rw->OutArcPoint(seln1);
        rw->DrawLine(nodepoint, selp, SMALL, arctype);
        return;
    }

    Point2d pnode = seln2->GetPosition();
    boxp2 = p;

    if (seln1 != seln2 || ((p.y - pnode.y) <= DELTA_NODEEDGE
        && boxp1.y - boxp2.y >= MINMOVE)) {
        // rubberbanding an arc from a src to a chosen dst.
        // if arc already exists select it.

        // erase rubberband arc
        Point nodepoint = rw->OutArcPoint(seln1);
        rw->DrawLine(nodepoint, selp, SMALL, arctype );

        RigiArc* rap = seln1->FindArc( seln2 );
        if (rap != NULL) {
            // arc exists already
            if (editor->GetSelectedArc() == rap) {
                // arc is already selected - deselect it
                editor->DeselectArc();
            } else {
                // arc is not already selected
                editor->DeselectArc();
                editor->SelectArc(rap);
                ArcMessage(rw, "Selected arc", rap);
            }
            return;
        }

        // arc does not exist already - create it
        GutilNewArc(rw, editor->GetArcType(), seln1, seln2);
        EditUpdateAllWindows();

        ArcMessage(rw, "New arc", NULL);

    } else {

        // erase the rubber arc line if there is no legitimate destination
        Point nodepoint = rw->OutArcPoint(seln1);
        rw->DrawLine(nodepoint, selp, SMALL, arctype);
    }
}

void EventPressSelect(RigiWindow *rw, Point p) {
    SetShift();
    if (!IsNewMouseEvent(SELECT_BUTTON, PRESS, p)) return;
    RutilSetActiveWindow(rw);
    RutilRaiseWindow(rw);
    p = rw->Canvas2World(rw->Window2Canvas(p));
    selp = p;
    boxp1 = p;
    boxp2 = IllegalPoint;
    origState = NORMAL;
    dragged = RigiFalse;

    if (EventShiftDown()) {
        // moving or extending selection set
        if ((seln1 = rw->FindNode(p)) == NULL) {
            // search for arc
            RigiArc *ra = rw->FindArc(p);
            if (ra != NULL) {
                // arc is clicked on
                if (editor->GetSelectedArc() == ra) {
                    // arc clicked on was already selected
                    editor->DeselectArc();
                } else {
                    // clicked on a new arc
                    editor->DeselectArc();
                    editor->SelectArc(ra);
                    ArcMessage(rw,"Selected arc", ra);
                }
            }
        } else {
            origState = seln1->GetState();
            editor->SelectNode(seln1);
        }

    } else {
        // moving or selecting
        // deselect any selected nodes
        editor->DeselectArc();
        rw->DrawSynch(RigiTrue);
        rw->BunchNode(RigiTrue);
        editor->EmptySelection();
        rw->BunchNode(RigiFalse);
        rw->DrawSynch(RigiFalse);
        rw->Window()->ResetMessage();
        if ((seln1 = rw->FindNode(p)) == NULL) {
            RigiArc *ra = rw->FindArc(p);
            if (ra != NULL) {
                // if the cursor is on arc, select it
                editor->SelectArc(ra);
                ArcMessage(rw,"Selected arc", ra);
            }
        } else {
            // if the cursor is on a node, select it
            if ((origState = seln1->GetState()) == NORMAL)
                editor->SelectNode(seln1);
            NodeMessage(rw, seln1);
        }
    }
    GetSelectionSet(rw);
}

void EventPressAdjust(RigiWindow *rw, Point p) {
    SetShift();
    if (!IsNewMouseEvent(ADJUST_BUTTON, PRESS, p)) return;
    RutilSetActiveWindow(rw);
    RutilRaiseWindow(rw);
    p = rw->Canvas2World(rw->Window2Canvas(p));
    selp = IllegalPoint;
    boxp1 = p;
    boxp2 = IllegalPoint;
    dragged = RigiFalse;

    // save the source node to create an arc later
    seln1 = rw->FindNode( p );
}

void EventDragSelect(RigiWindow *rw, Point p ) {
    if (!IsNewMouseEvent(SELECT_BUTTON, DRAG, p)) return;
    p = rw->Canvas2World(rw->Window2Canvas(p));

    if (seln1 == NULL) {
        // rubber band box
        if (boxp2.x != IllegalPoint.x)
            rw->DrawBox(boxp1, boxp2, BOXSIZE, BOXCOLOUR,1);
        boxp2 = p;
        rw->DrawBox(boxp1, boxp2, BOXSIZE, BOXCOLOUR, 1);
        dragged = RigiTrue;
    } else if  (selp.x != p.x || selp.y != p.y) {
        // move selection set
        rw->DrawSynch(RigiTrue);
        editor->DrawNodes(&selectset);
        editor->MoveNodes(&selectset, p.x-selp.x, p.y-selp.y);
        editor->DrawNodes(&selectset);
        rw->DrawSynch(RigiFalse);
        selp = p;
        dragged = RigiTrue;
    }
}

void EventDragAdjust(RigiWindow *rw, Point p ) {
    if (seln1 != NULL) {
        if (!IsNewMouseEvent(ADJUST_BUTTON, DRAG, p)) return;
        p = rw->Canvas2World(rw->Window2Canvas(p));
        int arctype = editor->GetArcType();
        if (rw->GetType() != WIN_GENERAL) {
            RigiError(ERRORFLAG, "Arcs may not be created in this window");
            return;
        }
        if (editor->LevelArc() == arctype) {
            RigiError(ERRORFLAG, "Level arcs may not be created; use \"Collapse\"");
            return;
        }

        // rubber band arc line
        Point nodepoint = rw->OutArcPoint( seln1 );
        if (selp.x != IllegalPoint.x && selp.y != IllegalPoint.y)
            rw->DrawLine(nodepoint, selp, SMALL,arctype);
        selp = p;
        rw->DrawLine(nodepoint,selp, SMALL,arctype);
        dragged = RigiTrue;
    }
}

void EventXScroll(RigiWindow* rw, int pos) {
    RigiBool refresh = RigiFalse;

    if (pos != rw->Window()->ScrollbarStartX()) {
        rw->Window()->SetScrollbarStartX(pos);
        if (!rw->BackStore()) rw->DrawStore();
        rw->RestoreWindow();
    }
}


void EventYScroll(RigiWindow* rw, int pos) {
    if (pos != rw->Window()->ScrollbarStartY()) {
        rw->Window()->SetScrollbarStartY(pos);
        if (!rw->BackStore()) rw->DrawStore();
        rw->RestoreWindow();
    }
}

