// $Author: jmcdanie $
// $Date: 1996-09-11 19:19:08 $
// $Source: /root/cvs/Rigi/rigiedit/eventops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: eventops.C,v 1.19 1996-09-11 19:19:08 jmcdanie Exp $";

//------------------------------------------------------------------------


#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GRAPHERR_H
#include "grapherr.h"
#endif
#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"         // Basic graph definitions
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Class definition for graph Nodes
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"       // Class definition for RigiWindows
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"         // Class definition for RigiWindows
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"       // Window menu state access functions
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef MOVEOPS_H
#include "eventops.h"		 // Definitions for this file
#endif
#ifndef WIN_UTIL_H
#include "win_util.h"
#endif
#ifndef EVENT_H
#include "event.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

const int DRAG_SELECT       = 0;
const int DRAG_ADJUST       = 1;
const int PRESS_SELECT      = 2;
const int PRESS_ADJUST      = 3;
const int RELEASE_SELECT    = 4;
const int RELEASE_ADJUST    = 5;
const int EXPOSE            = 6;
const int XSCROLL           = 7;
const int YSCROLL           = 8;
const int DESTROY           = 9;
static void (*MouseFunc[])(RigiWindow *, Point) = {
    EventDragSelect,  EventDragAdjust,    EventPressSelect,
    EventPressAdjust, EventReleaseSelect, EventReleaseAdjust
};

static int GetOp(char *arg) {
    static char *ops[] = {
        "drag_select",      /* 0 */
        "drag_adjust",      /* 1 */
        "press_select",     /* 2 */
        "press_adjust",     /* 3 */
        "release_select",   /* 4 */
        "release_adjust",   /* 5 */
        "expose",           /* 6 */
        "xscroll",          /* 7 */
        "yscroll",          /* 8 */
        "destroy"           /* 9 */
    };
    for (int i = 0; i < sizeof(ops)/sizeof(char *); i++)
        if (!strcmp(arg, ops[i])) return i;
    return -1;
}

static RigiBool NumArgsOK(int argc, int argsExpected) {
    if (argc != argsExpected) {
        RigiError(ERRORFLAG, "Wrong number of arguments (%d)", argc);
        return RigiFalse;
    }
    return RigiTrue;
}

int EventOp(int argc, char **argv, CharType &/*result*/) {
    // args: 0=cmd, 1=op, 2=window #, ... 
    if (argc < 3)
        return RigiError(ERRORFLAG, "Wrong number of arguments (%d)", argc);

    RigiWindow * rw = editor->GetWindowByNumber(atoi(argv[2]));
    if (rw == NULL)
        return RigiError(ERRORFLAG, "Incorrect window (%d) specified", argv[2]);

    Point p;
    p.z = 0;
    int op;

    switch (op = GetOp(argv[1])) {
    case DRAG_SELECT:
    case DRAG_ADJUST:
    case PRESS_SELECT:
    case PRESS_ADJUST:
    case RELEASE_SELECT:
    case RELEASE_ADJUST:
        if (!NumArgsOK(argc,5)) return RigiFalse;
        p.x = atoi(argv[3]); p.y = atoi(argv[4]);

        MouseFunc[op](rw, p);
        break;
    case EXPOSE:
        if (!NumArgsOK(argc, 7)) return 0;
        p.x = atoi(argv[3]); p.y = atoi(argv[4]);
        EventExpose(rw, p, atoi(argv[5]), atoi(argv[6]));
        break;

    case XSCROLL:
        if (!NumArgsOK(argc, 4)) return 0;
        EventXScroll(rw, atoi(argv[3]));
        break;

    case YSCROLL:
        if (!NumArgsOK(argc, 4)) return 0;
        EventYScroll(rw, atoi(argv[3]));
        break;

    case DESTROY:
        if (!NumArgsOK(argc, 3)) return 0;
        EventDestroy(rw);
        break;

    default:
        return RigiError(ERRORFLAG, "Invalid operation (%s)", argv[1]);
    }    
    return RigiTrue;
}
