// fileops.C

// $Author: jmcdanie $
// $Date: 1996-09-11 19:19:08 $
// $Source: /root/cvs/Rigi/rigiedit/fileops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: fileops.C,v 1.66 1996-09-11 19:19:08 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>
#include "environ.h"

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"       // Class definition for RigiWindows
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"         // Class definition for RigiWindows
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"       // Window menu state access functions
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef FILEOPS_H
#include "fileops.h"
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"                // Basic control functions for the editor
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"                
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"           // Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"            // Utility control functions for the editor
#endif
#ifndef WINUTIL_H
#include "win_util.h"            // Window state access functions
#endif
#ifndef EVENT_H
#include "event.h"
#endif
#ifndef _ICON_H
#include "icon.h"
#endif
#ifndef TK_COLORMAP_H
#include "tk_colormap.h"
#endif
#ifndef GRAPHARCSET_H
#include "cl_grapharcset.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif

// Operations include files
#ifndef EDIT_H
#include "edit.h"                // Edit operation function definitions
#endif

#ifndef QUIT_H
#include "quit.h"		// [SRT 19Sep93]
#endif

extern int DbLoadRSF(), DbSaveRSF();

void FileOp( int op ) { FileOp( op, NULL ); }


static void node_func( GraphNode* np, void* val ) {
    CharType cmd;
    char str[80];

    cmd.Assign( (char*)val );
    sprintf( str, " %d node", np->GetId() );
    cmd.Append( str );
    Mess( "Visiting node %s, proc = %s, cmd = %s", 
        np->GetLabel(), (char*)val, cmd.string() );
    exec_command( cmd.string() );
}

typedef struct {
    char *name;
    ArcSet list;
} ArcInfo;

static void arc_func(GraphArc* ap, void* val) {
    CharType cmd;
    char str[80];
    ArcInfo *info = (ArcInfo*)val;
    int id = ap->GetId();

    // Check to see if the arc has been visited, if so just return. If not,
    // add it to the list

    if (info->list.Find( id ) != NULL) return;
    info->list.Add( ap );

    // Build the RCL command and call it

    cmd.Assign( info->name );
    sprintf( str, " %d arc", id );
    cmd.Append( str );
    Mess( "Visiting arc %d, proc = %s, cmd = %s", id, info->name, cmd.string() );
    exec_command( cmd.string() );
}

static void FileOpReset() {
    int x,y, w,h;

    Mess("Closing windows");
    editor->CloseAll();
    Mess("Clearing and deleting clipboard");
    EditClear();
    Mess("Emptying the root window");
    RigiBaseWindow()->Empty();
    RutilGetRootLocation( x,y );
    RutilGetDefFrameSize( w,h );
    RigiBaseWindow()->Window()->SetGeometry( x,y, w,h );
    RigiBaseWindow()->SetXScale(1.0F);
    RigiBaseWindow()->SetYScale(1.0F);

    Mess("Resetting the editor");
    editor->Reset();

    EnvReInit();            // reinitialize the paths
}

static void FileOpBuildBaseNode() {
    GraphNode *basenode = new GraphNode( ROOT );
    basenode->SetId(1);
    basenode->SetType(0);

    // add to editor's node sets
    GutilAddNode( basenode );

    RigiSetBaseNode(basenode);
    RigiBaseWindow()->SetNode( basenode );
    RigiBaseWindow()->Update();
}

void FileOp( int op, char *str ) {
    WindowSetCursor( CURSOR_WAIT );
    RigiWindow* activewin;
    RigiArc *ra;

    switch (op) {
        case SHOWCLIP:
            RigiClipBoard()->Show();
            RigiClipBoard()->RigiDisplay();  
            break;

        case UPDATE:
            activewin = RutilActiveWindow();
            ra = editor->GetSelectedArc();
            if (ra != NULL && ra->GetWindow() == activewin)
                editor->DeselectArc();
            editor->EmptySelection();
            editor->ResetFilter(activewin);
            activewin->Update(); 
            break;

        case LOAD:
            extern void edit_check_composite( GraphNode* );
            DbLoadRSF();
            edit_check_composite( RigiBaseNode() );
            RigiBaseWindow()->Update();
            break;

        case SAVE:
            DbSaveRSF();
            break;

        case RIGIINIT:
            FileOpReset();
            RinitInit(str);
            FileOpBuildBaseNode();
            rigi_colormap->InitColorMap();
            IconInitialize();
            break;

        case CLOSEALL:
            editor->CloseAll();
            break;

        case REFRESH:
            RutilActiveWindow()->RigiDisplay();
            break;

        case DONE:
            activewin = RutilActiveWindow();
            if (activewin == RigiBaseWindow()) {
                RigiError(ERRORFLAG, "This window may not be closed");
                break;
            }
            editor->CloseWindow(RutilActiveWindow());
            RigiBaseWindow()->RigiDisplay();
            break;

        case DUMP:
        {
            ArcInfo info;
            CharType cmd;
            ArcSet arcset;
            char tmpstr[80];
            GraphArc *arc;
            ArcDirection arcinout = ARC_INOUT;

            // For each node in the editor, call the callback function
            // associated with nodes

            GraphNode *n; AvlPos *P;
            for ( n = editor->FirstNode(P); n; n = editor->NextNode(P) )       {
                cmd.Assign( str );
                sprintf( tmpstr, " %d node", n->GetId() );
                cmd.Append( tmpstr );
                //Mess( "Visiting node %s, proc = %s, cmd = %s",
                //      n->GetLabel(), str, cmd.string() );
                exec_command( cmd.string() );

                // For each arc from this node, call the arc callback if
                // the arc has not been visited before.

                AvlPos *Q;
                arc = n->FirstArc( Q, ANY_ARC, arcinout);
                while ( arc != NULL )
                {
                    // Check to see if the arc has been visited, if so
                    // skip it. If not, add it to the list and process it

                    int id = arc->GetId();
                    if (arcset.Find( id ) == NULL) {
                        arcset.Add( arc );

                        // Build the RCL command and call it

                        cmd.Assign( str );
                        sprintf( tmpstr, " %d arc", id );
                        cmd.Append( tmpstr );
                        //Mess( "Visiting arc %d, proc = %s, cmd = %s",
                        //      id, str, cmd.string() );
                        exec_command( cmd.string() );
                    }
                    arc = n->NextArc( Q, ANY_ARC, arcinout);
                }
            }
        }
            break;

        case CURSORSET:
            op = !strcmp( str, "wait" ) ? CURSOR_WAIT : CURSOR_NORMAL;
            WindowSetCursor( op );
            return;

        case QUIT:
            quit();
            break;

        default: 
            RigiError(ERRORFLAG, "Invalid operation (%d)", op);
            break;
    }
    WindowSetCursor( CURSOR_NORMAL );
}
