//
// Move the selected group of nodes so the bounding box is at the location x,y
//

// $Author: jmcdanie $
// $Date: 1996-05-12 23:58:59 $
// $Source: /root/cvs/Rigi/rigiedit/group.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: group.C,v 1.13 1996-05-12 23:58:59 jmcdanie Exp $";

//------------------------------------------------------------------------

#include <math.h>
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif

void MoveList( RigiNodeList& list, int x, int y ) {
    Point minpt;
    NDC_Point2d ndcp;
    RigiWindow *active_window = RutilActiveWindow();

    // Scan through the list and get the minimum x and y coordinate for the
    // set of nodes.

    ndcp.x = 1.0F;
    ndcp.y = 1.0F;
    //ndcp.z = 1.0;
    minpt = active_window->NDC2World(ndcp);
//    Mess("minpt = %f, %f, %f\n", minpt.x, minpt.y, minpt.z);
        RigiNode* rnp = list.First();
        while( rnp != NULL )
    {
        GraphNode* np = rnp->GetNode();
        ndcp = np->GetNDCPosition();
        Point pt = active_window->NDC2World( ndcp );
        if ( pt.x < minpt.x ) minpt.x = pt.x;
        if ( pt.y < minpt.y ) minpt.y = pt.y;
        rnp = list.Next();
    }

    // Get the change in position of the nodes

    minpt.x = x - minpt.x;
    minpt.y = y - minpt.y;

    // Scan through the list again and set the nodes position this time

    for ( rnp = list.First(); rnp; rnp = list.Next() )
    {
        GraphNode* np = rnp->GetNode();
        ndcp = np->GetNDCPosition();
        Point2d pt = active_window->NDC2World2d( ndcp );
        pt.x += minpt.x;
        pt.y += minpt.y;
        rnp->SetPosition( pt );
    }
}

// Group the selected nodes at World location x,y with offset xo,yo

void GroupNodes( RigiNodeList& list, int x, int y, int xo, int yo) {
    RigiWindow *active_window = RutilActiveWindow();
    GraphNode *np = NULL;
    Point2d basept;
    NDC_Point2d ndcp;

    RigiNode* rnp = list.First();
    while( rnp != NULL )
    {
        if ( np == NULL )
        {
            // Get the basept from the first node, convert it to
            // the location passed to group_selected, and store it

            np = rnp->GetNode();
            ndcp = np->GetNDCPosition();
            basept = active_window->NDC2World2d( ndcp );
            basept.x = (Coordinate)x;
            basept.y = (Coordinate)y;

            rnp->SetPosition( basept );
            ndcp = active_window->World2NDC2d( basept );
            np->SetNDCPosition( ndcp );
        }
        else
        {
            // For each node, add the offset and move the
            // node to the new location.

            np = rnp->GetNode();
            basept.x += xo;
            basept.y += yo;
            rnp->SetPosition( basept );
            ndcp = active_window->World2NDC2d( basept );
            np->SetNDCPosition( ndcp );
        }
        rnp = list.Next();
    }
}

void GridNodes(RigiWindow* win, RigiNodeList& list,
int xoffset, int yoffset, int gridsize) {
    GraphNode *np;
    int xloc,yloc,x;
    int size;
    NDC_Point2d ndcp;
    Point2d worldp;

    // Don't want to do anything if empty list

    if (list.GetNoItems() < 1) return;

    // If it is a perfect square root then do not add one to the grid size.

    double ssize = sqrt((double)list.GetNoItems());
    if ( ssize != (double)((int)(ssize)) )
        size = (int)(ssize) + 1;
    else
        size = (int)ssize;


    xloc = xoffset;
    yloc = yoffset;
    x = 0;

    RigiNode* rnp = list.First();
    while( rnp != NULL ) {

        // Get the position of the rigi node and set the x  and y values

        worldp = rnp->GetPosition();
        worldp.x = (Coordinate)xloc;
        worldp.y = (Coordinate)yloc;
        rnp->SetPosition( worldp );

        // Convert the position to an NDC coordinate and set the nodes location

        ndcp = win->World2NDC2d( worldp );
        np = rnp->GetNode();
        np->SetNDCPosition( ndcp );

        x++;
        xloc += gridsize;

        if ( x >= size ) {
            xloc = xoffset;
            x = 0;
            yloc += gridsize;
        }
        rnp = list.Next();
    }
}
