// $Author: jmartin $
// $Date: 2002-11-10 22:16:58 $
// $Source: /root/cvs/Rigi/rigiedit/infoops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: infoops.C,v 1.57 2002-11-10 22:16:58 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef __CTYPE_H
#include <ctype.h>		// for isdigit()
#endif
#ifndef CHARTYPEH
#include "adt/chartype.h"
#endif

#ifndef GLOBAL_H
#include "global.h"             // For editor
#endif
#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"       // for AdjList
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif

#include "win_util.h"		// for WindowSetCursor()
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif
#include "environ.h"
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef INFOOPS_H
#include "infoops.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif
#ifndef ACTIVEARC_H
#include "cl_activearc.h"
#endif
#ifndef EVENT_H
#include "event.h"
#endif

#ifndef SIMPLESET_H
#include "cl_simpleset.h"
#endif

extern int GetAvlMem(), GetAvlCharMem(), GetAvlObjMem();
extern int GetNoEdArcs(), GetNoEdNodes();
extern int GetNoRigiNodes(), GetNoGraphNodes();
extern int GetNoRigiArcs(), GetNoGraphArcs();
extern int GetNoDupList();
extern int GetNoANodes(), GetNoAArcs();
extern int GetNoObjArray();

static RigiBool GetRigiNode(GraphNode *graphNode, RigiNode *&rn) {
    RigiWindow *rw = RutilActiveWindow();
    if (rw == NULL) return RigiFalse;
    rn = rw->FindRigiNode(graphNode);
    return (rn == NULL)?
        RigiError(ERRORFLAG, "Riginode \"%s\" does not exist", graphNode->GetLabel()) : RigiTrue;
}

const char* InfoOp( int op, int id, char *name, char *attr ) {
    RigiDomain *domain = NULL;
    RigiWindow *activewin = NULL;
    GraphNode *node = NULL; GraphArc *arc = NULL;
    RigiNode *rn = NULL;

    static SimpleSet *set = NULL;

    static CharType tmpbuf;
    static char buff[64];	// MJW: big enough for three integers
    const char *str = buff;
    *buff = '\0';

// MJW HACK - online debug info
if (op==99) {
Mess("Sizes: Node=%d, Arc=%d, RN=%d, RA=%d, AN=%d, AA=%d,",
sizeof(GraphNode),sizeof(GraphArc),sizeof(RigiNode),sizeof(RigiArc),
sizeof(ActiveNode),sizeof(ActiveArc) );
Mess("AVL sets: %d (obj) and %d (%d char)",
GetAvlObjMem(),GetAvlMem(), GetAvlCharMem() );
Mess("Editor: %d arcs %d nodes, array alloc %d objectptrs",
GetNoEdArcs(), GetNoEdNodes(), GetNoObjArray() );
Mess("nnodes=%d nrnodes=%d narcs=%d nrarcs=%d nanodes=%d naarcs=%d",
GetNoGraphNodes(), GetNoRigiNodes(), GetNoGraphArcs(), GetNoRigiArcs(),
GetNoANodes(), GetNoAArcs() );
Mess("ndup=%d", GetNoDupList() );
return NULL;
}
    if (op == GETVERSION) {
        sprintf(buff, VERSION);
        return buff;
    } else if (op == GETSHIFTED) {
        return WindowIsShifted() ? "1" : "0";
    }

    static AvlPos *NAP;
    static AvlPos *AAP;

    // find graphnode or arc matching the id passed in
    if ( !IS_ENV_OP(op) && op != GETNODEID &&
        op != GETARCCOLOR && op != SETARCCOLOR &&
        op != GETNODECOLOR && op != SETNODECOLOR &&
        op != WRITECOLORS && op != GETNODEBITMAP &&
        !IS_SET_OP(op)) {
        if (IS_NODE_OP( op )) {
            node = editor->FindNode(id);
            if (!node) {
                RigiError(ERRORFLAG, "Node %d not found",id);
                return NULL;
            }
        } else {
            arc = editor->FindArc(id);
            if (!arc) {
                RigiError(ERRORFLAG, "Arc %d not found", id);
                return NULL;
            }
        }
    }

    switch (op) {
        // Get Node Info Commands
        case GETNODENAME:
            str = node->GetLabel();
            break;
    
        case GETNODEPOSITION:
        {
            if (!GetRigiNode(node, rn)) return NULL;
            Point2d p = rn->GetPosition();
            sprintf( buff, "%d %d", p.x, p.y );
            break;
        }
        case GETNODEABSPOSITION:
        {
            if ((activewin = RutilActiveWindow()) == NULL) return NULL;
            NDC_Point2d ndcp = node->GetNDCPosition();
            Point worldp = activewin->NDC2World( ndcp );
            Point winp = activewin->World2Canvas( worldp );
            sprintf( buff, "%d %d", winp.x, winp.y );
            break;
        }
        case GETNODESIZE:
            if (!GetRigiNode(node, rn)) return NULL;
            sprintf( buff, "%d", rn->NodeSize() );
            break;

        case GETNODETYPE:
            sprintf( buff, "%d", node->GetType() );
            break;

        case GETNODEATTR:
            if (!name) {
                RigiError(ERRORFLAG, "No node attribute specified");
                return NULL;
            }
            str = node->GetAttr( name );
            break;

        case GETNODEINWINDOW:
            if ((activewin = RutilActiveWindow()) == NULL) return NULL;
            rn = activewin->FindRigiNode( node );
            str = (rn == NULL)? "0" : "1";
            tmpbuf.Assign( str );
            str = tmpbuf.string();
            break;

        case GETNODENAMEISHIDDEN:
            if (!GetRigiNode(node, rn)) return NULL;
            sprintf( buff, "%d", rn->HideName() );
            break;

        case GETNODEFIRSTATTR:
            str = node->FirstAttr(NAP);
            if (str != NULL) {
                tmpbuf.Assign( str );
                str = tmpbuf.string();
            }
            break;

        case GETNODENEXTATTR:
            str = node->NextAttr(NAP);
            if (str != NULL) {
                tmpbuf.Assign( str );
                str = tmpbuf.string();
            }
            break;

        case GETNODENUMATTRS:
            sprintf( buff, "%d", node->NumAttrs() );
            break;

        case GETARCFIRSTATTR:
            str = arc->FirstAttr(AAP);
            if (str != NULL) {
                tmpbuf.Assign( str );
                str = tmpbuf.string();
            }
            break;

        case GETARCNEXTATTR:
            str = arc->NextAttr(AAP);
            if (str != NULL) {
                tmpbuf.Assign( str );
                str = tmpbuf.string();
            }
            break;

        case GETARCNUMATTRS:
            sprintf( buff, "%d", arc->NumAttrs() );
            break;

        case GETNODEID:
            if ((node = editor->FindNode(name)) == NULL) {
                RigiError(ERRORFLAG, "Node %s not found", name);
                return NULL;
            }
            sprintf( buff, "%d", node->GetId() );
            break;

        // Set Node Info Commands
        case SETNODENAME:
            editor->ChangeNodeName( node, name );
            break;

        case SETNODEPOSITION:
        // don't do normalized coordinates yet
        {
            int x, y;
            if ( sscanf( name, "%d %d", &x, &y ) != 2 ) {
                RigiError(ERRORFLAG, "Invalid node position %s", name);
                return NULL;
            }
            Point2d p;
            if (!GetRigiNode(node, rn)) return NULL;
            p.x = (Coordinate)(x);
            p.y = (Coordinate)(y);
            rn->SetPosition( p );		
            break;
        }

        case SETNODEABSPOSITION:
        {
            int x, y;
            if ( sscanf( name, "%d %d", &x, &y ) != 2 ) {
                RigiError(ERRORFLAG, "Invalid node position %s", name);
                return NULL;
            }
            if ((activewin = RutilActiveWindow()) == NULL) return NULL;
            Point winp;
            winp.x = (Coordinate)(x);
            winp.y = (Coordinate)(y);
            Point worldp = activewin->Canvas2World( winp );
            NDC_Point2d ndcp = activewin->World2NDC2d( worldp );
            node->SetNDCPosition( ndcp );
            break;
        }

        case SETNODETYPE:
        {
            domain = RinitGetDomain();
            int type = domain->NodeTypeFromName( name );
            if (type==-1) {
                RigiError(ERRORFLAG, "Node type \"%s\" is undefined!",name);
                return NULL;
            }
            EventUpdateNodeType(node, type); 
            break;
        }
        case SETNODEATTR:
            if (!name) {
                RigiError(ERRORFLAG, "No node attribute specified");
                return NULL;
            }
            if (!attr) {
                RigiError(ERRORFLAG, "No node attribute value specified");
                return NULL;
            }
            node->SetAttr( name, attr );
            break;

        case REMOVENODEATTR:
            if (!name) {
                RigiError(ERRORFLAG, "No node attribute specified");
                return NULL;
            }
            if (node->RemoveAttr( name ) == 0)
                str = NULL;
            else
                str = name;
            break;

        case GETNODEATTRS:
            node->PrintAttrs();
            break;

        case FILTERNODE:
            if (!GetRigiNode(node, rn)) return NULL;
            editor->SetFilter(rn);
            break;

        case UNFILTERNODE:
            if (!GetRigiNode(node, rn)) return NULL;
            editor->ResetFilter(rn);
            break;

        case GETNUMARCS:
        {
            domain = RinitGetDomain();
            int arctype = domain->ArcTypeFromName( name );
            if (arctype==-1)
            {
                RigiError(ERRORFLAG, "Arc type \"%s\" is undefined!",name );
                return NULL;
            }
            ArcDirection dir = (attr && !strcmp("in",attr)) ?
                ARC_IN : ARC_OUT;

            int n = node->ArcNum( arctype, dir );
            sprintf( buff,"%d",n );
            break;
        }
        case GETNODESELECTED:
            if (!GetRigiNode(node, rn)) return NULL;
            str = rn->GetState()==ACTIVE ? "1" : NULL;
            break;

        // Get Arc Info Commands

        case GETARCNAME:
            RigiError(ERRORFLAG, "Arc names have not been implemented");
            return NULL;

        case GETARCSRC:
            sprintf( buff, "%d", arc->GetSrc()->GetId() );
            break;

        case GETARCDST:
            sprintf( buff, "%d", arc->GetDst()->GetId() );
            break;

        case GETARCTYPE:
            sprintf( buff, "%d", arc->GetType() );
            break;

        case GETARCATTR:
            str = arc->GetAttr( name );
            break;

        // Set Arc Info Commands
        case SETARCNAME:
            RigiError(ERRORFLAG, "Arc names have not been implemented");
            return NULL;

        case SETARCTYPE:
        {
            domain = RinitGetDomain();
            int type = domain->ArcTypeFromName( name );
            if (type == -1) {
                RigiError(ERRORFLAG, "Invalid arc type %s", name);
                return NULL;
            }
            GutilChangeArcType( arc, type );
            break;
        }
        case SETARCATTR:
           if (!name) {
                RigiError(ERRORFLAG, "No arc attribute specified");
                return NULL;
            }
           if (!attr) {
                RigiError(ERRORFLAG, "No arc attribute value specified");
                return NULL;
            }
            arc->SetAttr( name, attr );
            break;

        case REMOVEARCATTR:
           if (!name) {
                RigiError(ERRORFLAG, "No arc attribute specified");
                return NULL;
            }
            if (arc->RemoveAttr( name ) == 0)
                str = NULL;
            else
                str = name;
            break;

        case GETARCCOLOR:
        {
            // Get the arc type from the name given

            domain = RinitGetDomain();
            int type = domain->ArcTypeFromName( name );
            if (type == -1) {
                RigiError(ERRORFLAG, "Invalid arc type %s", name);
                return NULL;
            }

            // Get the color for that type and convert it into an 8 bit
            // RGB triple.

            RigiColor color;
            domain->GetArcColor( type, &color );
            sprintf(buff,"%d %d %d", color.red >> 8, color.green >> 8,
                                    color.blue >> 8);
            break;
        }
        case SETARCCOLOR:
        {
            // Get the type of the arc

            domain = RinitGetDomain();
            int type = domain->ArcTypeFromName( name );
            if (type == -1) {
                RigiError(ERRORFLAG, "Invalid arc type %s", name);
                return NULL;
            }

            // Perform some sanity checks on the color value passed

            RigiColor color;
            int argc;
            const char **argv;
            if (Tcl_SplitList( rcl_interp, attr, &argc, &argv ) != TCL_OK) {
                RigiError(ERRORFLAG, "Invalid color specification %s", attr);
                return NULL;
            }
            if (argc != 3) {
	        free((char*)argv);
                RigiError(ERRORFLAG, "Invalid color specification %s", attr);
                return NULL;
            }

            // Get the RGB triple and make sure the values are in the range
            // 0 - 255

            color.red = atoi(argv[0]);
            if (color.red < 0) color.red = 0;
            if (color.red > 255) color.red = 255;

            color.green = atoi(argv[1]);
            if (color.green < 0) color.green = 0;
            if (color.green > 255) color.green = 255;

            color.blue = atoi(argv[2]);
            if (color.blue < 0) color.blue = 0;
            if (color.blue > 255) color.blue = 255;

            // Make the RGB values 16 bit by shifting 8 bits. Then set the
            // arc types color

            color.red <<= 8;
            color.green <<= 8;
            color.blue <<= 8;
            rigi_colormap->SetArcColor( type, &color );
            domain->SetArcColor( type, &color );
	    free((char*)argv);
            break;
        }
        case GETNODECOLOR:
        {
            // Get the type of the node

            domain = RinitGetDomain();
            int type = domain->NodeTypeFromName( name );
            if (type == -1) {
                RigiError(ERRORFLAG, "GETNODECOLOR: Invalid node type %s", name);
                return NULL;
            }
            // Get the color and convert it into an RGB triple. The value
            // given is a 16 bit value so we must truncate the lower 8 bits
            // of the color.

            RigiColor color;
            domain->GetNodeColor( type, &color );
            sprintf(buff,"%d %d %d", color.red>>8, color.green>>8,
                                    color.blue>>8 );
            break;
        }
        case SETNODECOLOR:
        {
            // Get the type of the node

            domain = RinitGetDomain();
            int type = domain->NodeTypeFromName( name );
            if (type == -1) {
                RigiError(ERRORFLAG, "Invalid node type %s", name );
                return NULL;
            }

            // Check the consistency of the given color value

            RigiColor color;
            int argc;
            const char **argv;
            if (Tcl_SplitList( rcl_interp, attr, &argc, &argv ) != TCL_OK) {
                RigiError(ERRORFLAG, "Invalid color specification %s", attr);
                return NULL;
            }
            if (argc != 3) {
	        free((char*)argv);
                RigiError(ERRORFLAG, "Invalid color specification %s", attr );
                return NULL;
            }

            // Convert the color triple to the local color type and make sure
            // it is in the range 0 - 255.

            color.red = atoi(argv[0]);
            if (color.red < 0) color.red = 0;
            if (color.red > 255) color.red = 255;

            color.green = atoi(argv[1]);
            if (color.green < 0) color.green = 0;
            if (color.green > 255) color.green = 255;

            color.blue = atoi(argv[2]);
            if (color.blue < 0) color.blue = 0;
            if (color.blue > 255) color.blue = 255;

            // Shift the color values 8 bits to make them into 16 bit values
            // and set the node types color

            color.red <<= 8;
            color.green <<= 8;
            color.blue <<= 8;
            rigi_colormap->SetNodeColor( type, &color );
            domain->SetNodeColor( type, &color );
	    free((char*)argv);
            break;
        }
        case WRITECOLORS:
            RinitWriteColors();
            break;

        case GETNODEBITMAP:
        {
            // Get the type of the node

            domain = RinitGetDomain();
            int type = domain->NodeTypeFromName( name );
            if (type == -1) {
                RigiError(ERRORFLAG, "Invalid node type %s", name);
                return NULL;
            }
            str = domain->IconFileName( type );
            break;
        }


        case SETNEW:
            if (set) {
                RigiError(ERRORFLAG, "Tcl-accessible sets are not implemented");
                return NULL;
            }
            set = new SimpleSet();
            str = "zeb";
            break;

        case SETADD:
            if (!set) break;
            set->Insert(id);
            break;

        case SETREMOVE:
            if (!set) break;
            set->Remove(id);
            break;

        case SETFIND:
            if (!set) break;
            str = set->Find(id) ? "zeb" : NULL;
            break;

        case SETEMPTY:
            if (!set) break;
            delete set;
            set = NULL;
            break;

        case SETFIRST:
            if (!set) break;
            id = set->First();
            if (id) sprintf( buff,"%d", id ); else str = NULL;
            break;

        case SETSIZE:
            if (!set) break;
            sprintf( buff,"%d",set->Size() );
            break;

        case SETENVIRON:
            EnvSet( name, attr );
            break;

        case GETENVIRON:
            str = EnvGet( name, atoi(attr) );
            break;

        case PRINTENVIRON:
            EnvPrint();
            break;

        default:
            RigiError(ERRORFLAG, "Invalid operation (%d)",op);
            return NULL;
    }
    return str;
}

