// $Author: jmartin $
// $Date: 2002-11-19 21:41:33 $
// $Source: /root/cvs/Rigi/rigiedit/mbi.C,v $
// $Locker:  $
// $State: Exp $
//
// ------------------------------------------------------------------------

//
// mbi.C (formerly ut.C): the rigiedit side of the Rigi/mbus interface.
//

#ifndef RIGISTATE_H
#include "cl_rigistate.h"               // Global instance of editor's state
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"		// for internal Rigi graph model
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif

#ifndef DOMAIN_H
#include "cl_domain.h"			// for domain class
#endif
#include "ut_telos.h"			// for Telos-side interface
#ifndef MBI_H
#include "mbi.h"			// for this file (Rigi-side)
#endif

#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#include "environ.h"			// to get RIGIDBHOST and RIGIDBPORT
#ifndef ATTRIBUTE_H
#include "cl_attribute.h"		// for local sets
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"			// for RutilActiveWindow()
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"			// for GutilCreateNode, GutilCreateArc()
#endif
#ifndef GLOBAL_H
#include "global.h"			    // for editor
#endif

#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif

#if defined (__UT__)
static void dumpnode( GraphNode*, RigiWindow* );
static void UT_receive( char*, char*, char* );
static char *Label(GraphNode*,char*typeName);

static GraphNode *parent = NULL;
static int rsf_debug = 0;

static const int BUFFSIZE = 256;
static const int DSTSIZE = 256;
static char arcbuff[BUFFSIZE], srcbuff[BUFFSIZE], dstbuff[DSTSIZE];

static AttributeSet badnodetypes;
#endif

void UT_read( char *objname, char *s ) {
#if defined(__UT__)
    RigiWindow *rw = NULL; int i=0;

    // Permit loading of RSF info direct into the rigigraph: MJW HACK!
    if (s && strcmp(s,"db")) {
        rw = RutilActiveWindow();
        if (!rw) return;
        parent = rw->GetNode();
    }
    else parent = NULL;

    editor->EmptySelection();
    badnodetypes.Empty();

    Mess("asking \"%s\" for info from server ..",objname);

    // ASK the server, indicating callback function located here

    if (!objname || !*objname) {
        AvlPos *P;
        if (rw)
        for ( RigiNode *rn = rw->FirstNode(P); rn; rn = rw->NextNode(P) ) {
            GraphNode *n = rn->GetNode();
            if (!rn->isVisible()) continue;

            RigiDomain *domain = RinitGetDomain();
            char *typeName = domain->NodeNameFromType( n->GetType() );

            i = UT_interface( UT_ASK, Label(n,typeName), UT_receive );
        }
    } else {
        i = UT_interface( UT_ASK, objname, UT_receive );
    }
    Mess("Finished receiving info from server - %d RSF tuples read",i);

    if (rw) rw->Update();
#endif
}

void UT_write( char *objname ) {
#if defined(__UT__)
    RigiWindow *rw = RutilActiveWindow();
    GraphNode *n; AvlPos *P;

    if (!rw) return;

    if (!objname || !*objname || !strcmp( objname, "@select" ) ) {
        Mess("Writing selected set to server ..");
        for ( n=editor->FirstSelectedNode(P); n;
            n=editor->NextSelectedNode(P) ) {
            dumpnode( n, NULL );
        }
    } else if (!strcmp( objname, "@all" ) ) {
        Mess("Writing entire Rigi graph to server ..");
        for ( n = editor->FirstNode(P); n; n = editor->NextNode(P) ) {
            dumpnode( n, NULL );
        }
    } else if (!strcmp( objname, "@window" ) ) {
        Mess("Writing visible info in window to server ..");
        AvlPos *P;
        for ( RigiNode *rn = rw->FirstNode(P); rn; rn = rw->NextNode(P) ) {
            GraphNode *n = rn->GetNode();
            if (rn->isVisible()) dumpnode( n, rw );
        }
    } else {
        Mess("Writing object \"%s\" to server ..",objname);
        GraphNode *n = editor->FindNode( objname );
        if (n) dumpnode( n, rw );
            else Mess("error, can't find %s",objname);
    }

    // Must indicate to the interface that we are finished accumulating
    // data to send.

    UT_interface( UT_END );
    Mess("Finished writing info to server");
#endif
}

// This is a little confusing. Commands can be to this module,
// to the interface (TML), or to the data server via mbus.

void UT_command( char *command ) { UT_command( command, NULL ); }
void UT_command( char *command, char *name ) {
#if defined(__UT__)
    if (!command)
        mayday("no command passed to UT_command!");

    // First, the local commands
    else if (!strcmp(command,"rsf")) {
        rsf_debug = !rsf_debug;
        Mess("received RSF tuples to print %sabled",
            rsf_debug ? "en" : "dis" );
    }

    // Now, the commands which are passed on to the interface
    else if (!strcmp(command,"start"))
        UT_interface( UT_START,EnvGet("RIGIDBHOST"),EnvGet("RIGIDBPORT"),NULL);
    else if (!strcmp(command,"stop"))
        UT_interface( UT_STOP, NULL, NULL, NULL );
    else if (!strcmp(command,"debug"))
        UT_interface( UT_DEBUG );
    else if (!strcmp(command,"time"))
        UT_interface( UT_TIME, name, NULL, NULL );
    else if (!strcmp(command,"max"))
        UT_interface( UT_MAXOBJECTS, name, NULL, NULL );
    else if (!strcmp(command,"flush"))
        UT_interface( UT_FLUSH, name, NULL, NULL );
    else if (!strcmp(command,"general"))
        UT_interface( UT_TELOS, name, NULL, NULL );
    else if (!strcmp(command,"help") || !*command )
        Mess( "valid tml commands:\n\
start\nstop\n\
max <n>\ntime <s>\n\
debug\nrsf\n\
flush\n\
general <s> (passed on to the server)\n"
        );
    else Mess("unrecognized Telos command \"%s\"",command );
#endif
}

void UT_check()
{
#if defined(__UT__)
    UT_interface( UT_ASK, NULL, UT_receive );
#endif
}


/////// Local Functions ///////

#if defined(__UT__)
static void dumpnode( GraphNode* n, RigiWindow *rw ) {
    // Get the current domain being used

    RigiDomain *domain = RinitGetDomain();
    char *typeName = domain->NodeNameFromType( n->GetType() );

    // Use UT_interface to send RSF tuples of the info in this node.
    // Only doing major attributes and (visible) arcs right now.
    // (What about level arcs from window parent??)

    char srcname[64];
    char *arcname, *dstname;

    strcpy(srcname,Label(n,typeName));

    // The "type" attribute MUST be first!
    UT_interface( UT_TELL, "type", srcname, typeName );

    // HACK! If the object already has a name attribute, then
    // we'll be telling it later; do not tell now
    // (Distinguish between Refine and Rigi subclasses)

    //if (!n->GetAttr("refineName"))
        //UT_interface( UT_TELL, "rigiName", srcname, n->GetLabel() );

    // Dump out arc info

    GraphArc *a; AvlPos *P;
    ArcDirection arcout = ARC_OUT;
    for ( a=n->FirstArc(P, ANY_ARC, arcout); a;
        a = n->NextArc(P, ANY_ARC, arcout)) {
        RigiArc *ra = rw ? rw->FindRigiArc( a ) : NULL;

        if (rw)
        if (a->GetType()!=editor->LevelArc())
        if (!ra || !ra->isVisible()) continue;

        arcname = domain->ArcNameFromType( a->GetType() );
        char *typeName = domain->NodeNameFromType( a->GetDst()->GetType() );
        dstname = Label(a->GetDst(),typeName);
        UT_interface( UT_TELL, arcname, srcname, dstname );
    }

    // Dump out attributes
    for ( char *attr = n->FirstAttr(P); attr; attr = n->NextAttr(P) ) {
        char *value = n->GetAttr( attr );
        if (domain->IsNodeAttr( n->GetType(), attr ) ) {
            UT_interface( UT_TELL, attr, srcname, value );
        } else {
            // use "general" attribute category ... HACK!

            if (strlen(attr)+strlen(value)>DSTSIZE-2)
                mayday("problem in dumpnode");
            sprintf( dstbuff,"%s %s",attr,value );
            UT_interface( UT_TELL, "general", srcname, dstbuff );
        }
    }
}

// This is a callback function, used by the other side of the interface
// to communicate back info from the server. It receives RSF triples,
// and must convert this info into Rigi objects, their interrelationships,
// and their properties.

static void UT_receive( char *arcname, char *srcname, char *dstname ) {
    if (rsf_debug)
    Mess("RSF: %-15s %-15s %s",arcname,srcname,dstname);

    GraphNode *src=NULL, *dst=NULL;
    int id;

    // Get the current domain being used.

    RigiDomain *domain = RinitGetDomain();

    // First, must get the darn src object. If the srcname "label" starts
    // with a non-zero number, we assume that the object is identified
    // by Rigi id. Otherwise, it is identified by name.

    id = atoi(srcname);
    src = id ? editor->FindNode( id ) : editor->FindNode( srcname );

    if (!src) {
        // src object doesn't exist, so make one
        src = GutilCreateNode( parent, srcname, 0 );
        if (id) GutilChangeNodeId(src,id);
    } else {
        editor->SelectNode(src);
    }

    // Must strip doublequotes off of dstname, if required
    if (*dstname=='"') {
        int i = strlen(dstname);
        if ( dstname[i-1]=='"' ) {
            strcpy(dstbuff,dstname);
            dstbuff[i-1]='\0';
            dstname = dstbuff+1;
        }
    }

    // Now, figure out if this is a Rigi arc, or a Rigi attribute
    int arctype = domain->ArcTypeFromName( arcname );

    if ( arctype==-1 ) {
        // We must assume that this relation is an attribute.
        // Special case attributes: type, name/rigiName, general

        if (!strcmp(arcname,"type")) {
            int nodetype = domain->NodeTypeFromName( dstname );
            if ( nodetype==-1 ) {
                if (!badnodetypes.SetAttr( dstname, 1 ))
                    warn("node type \"%s\" not recognized",dstname);
            } else {
                src->SetType( nodetype );
            }
        } else if (!strcmp(arcname,"name") ) {
            GutilChangeNodeName( src, dstname );
        } else if (!strcmp(arcname,"rigiName") ) {
            GutilChangeNodeName( src, dstname );
        } else if (!strcmp(arcname,"general") ) {
            char buff1[512], buff2[512];
            sscanf( dstname,"%s %s",buff1,buff2 );
            src->SetAttr( buff1, buff2 );
        } else {
            src->SetAttr( arcname, dstname );
        }
    } else {
        // Valid arctype, so look for destination object. Again, we
        // use the dstname's label to determine if searching by Rigi id,
        // or by name.

        id = atoi(dstname);
        dst = id ? editor->FindNode( id ) : editor->FindNode( dstname );

        if (!dst) {
            dst = GutilCreateNode( parent, dstname, 0 );
            if (id) GutilChangeNodeId(dst,id);
        } else {
            editor->SelectNode(dst);
        }

        // Don't make another level arc, if this is one. HACK!
        if ( src==parent ) if ( arctype==editor->LevelArc() ) return;

        GraphArc *a = src->FindArc( dst, outArcs );
        if (!a) a = GutilCreateArc( src, dst, arctype );
    }
}

static char *Label( GraphNode *n, char *typeName ) {
    char *label;
    int num=0;

    // HACK! If the object is apparently a Refine object, then
    // we must use the Refine conventions for creating its label

    if (label = n->GetAttr("refineName"))
        num = atoi( label+1 );	// for refineNames of format "#35467"
    else
        num = n->GetId();
    if (!typeName) typeName = "zeb";

    static char buff[256];
/*
    if (strlen(n->GetLabel())>240) {
        fprintf(stderr,"error in Label!! too long\n");
        sprintf(buff,"badlabel");
        return buff;
    }
    sprintf(buff,"%d_%s",n->GetId(), n->GetLabel());
*/
    sprintf(buff,"%d",n->GetId());
    return buff;
}
#endif
