

// $Author: jmcdanie $
// $Date: 1996-05-20 18:20:55 $
// $Source: /root/cvs/Rigi/rigiedit/moveops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: moveops.C,v 1.39 1996-05-20 18:20:55 jmcdanie Exp $";

//------------------------------------------------------------------------


#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"         // Basic graph definitions
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Class definition for graph Nodes
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"       // Class definition for RigiWindows
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"         // Class definition for RigiWindows
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"       // Window menu state access functions
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef MOVEOPS_H
#include "moveops.h"		 // Definitions for this file
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"                // Basic control functions for the editor
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"           // Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"            // Utility control functions for the editor
#endif
#ifndef WINUTIL_H
#include "win_util.h"            // cursor stuff
#endif
#ifndef RIGIERR_H
#include "rigierr.h"             // Error reporting functions
#endif
#ifndef EVENT_H
#include "event.h"		 // for EventGetSelectBox function
#endif
#ifndef GROUP_H
#include "group.h"		 // move operations
#endif

const int NUMARGS = 3;
const int COMARG = 0;
const int OPARG = 1;
const int WINARG = 2;

// SCRIPT13
// Make global for now, when this is a class they will be in the CursorOP
// class.

static Point cursor_position;

static int GetNodeList(RigiNodeList &nlist, RigiWindow *rw) {
    AvlPos *P;
    int i = 0;
    for (RigiNode *rn = rw->FirstNode(P); rn; rn = rw->NextNode(P)) {
        i++;
        nlist.Insert(rn);
    }
    if (!i) {
         RigiError(ALERT, "There are no nodes to lay out");
         WindowSetCursor( CURSOR_NORMAL );
    }
    return i;
}

// Gets a list of the selected nodes in the active window
// returns the number of nodes selected
static int GetSelectionList(RigiNodeList &nlist, RigiWindow *rw) {
    int i;

    editor->GetSelectionList(nlist, rw);
    if ((i = editor->NumSelectedNodes()) < 1) {
         RigiError(ALERT, "At least one node must be selected");
         WindowSetCursor( CURSOR_NORMAL );
         return 0;
    }
    return i;
}

// Move operations: from notify procedure or RCL scripts

int MoveOp(int argc, char *argv[], CharType &result) {
    int op, winNum;
    RigiWindow *rw;

    // must have the operation argument to start with
    if (argc < OPARG + 1) {
        op = GRIDALL;
    } else {
        op = atoi(argv[OPARG]);
    }

    if (argc > NUMARGS) {
        RigiError(ERRORFLAG, "Wrong number of arguments (%d)", argc);
        return RigiFalse;
    }

    winNum = atoi(argv[WINARG]);
    rw = (winNum)? editor->GetWindowByNumber(winNum) : RutilActiveWindow();
    if (rw == NULL) return RigiError(ERRORFLAG, "Incorrect window (%d) specified", winNum);

    WindowSetCursor( CURSOR_WAIT );

    // Careful with this list. It is possible that an operation can result in
    // the deletion of RigiNodes. This list contains pointers to RigiNodes, so
    // the list may not point to valid objects after some operations. The ops
    // that you need to worry about are the edit operations, as they update
    // the windows. Most other ops just change the state of the nodes.

    RigiNodeList nlist;

    // Get some initial state information for the windows.
    int gridsize = editor->GetGridSize();
    Point p1, p2;
    int boxflag = EventGetSelectBox( p1, p2 );

    // Group and sort, according to posns and operation
    switch (op) {
        case GROUPDIAG:
            if (!GetSelectionList(nlist, rw)) return 0;
            nlist.DiagSort();
            GroupNodes(nlist, cursor_position.x ,cursor_position.y, gridsize, gridsize) ;
            break;
        case GROUPHORIZ:
            if (!GetSelectionList(nlist, rw)) return 0;
            nlist.HorizSort();
            GroupNodes(nlist, cursor_position.x, cursor_position.y, gridsize, 0) ;
            break;
        case GROUPVERT:
            if (!GetSelectionList(nlist, rw)) return 0;
            nlist.VertSort();
            GroupNodes(nlist, cursor_position.x, cursor_position.y, 0, gridsize);
            break;
        case GROUPGRID:
            if (!GetSelectionList(nlist, rw)) return 0;
            nlist.GridSort();
            GridNodes(rw, nlist, cursor_position.x, cursor_position.y, gridsize) ;
            break ;
        case GROUPMOVE:
            if (!GetSelectionList(nlist, rw)) return 0;
            MoveList(nlist,cursor_position.x,cursor_position.y);
            break;
        case GRIDALL:
            if (!GetNodeList(nlist, rw)) return 0;
            nlist.GridSort();
            GridNodes(rw, nlist, cursor_position.x, cursor_position.y, gridsize) ;
            break ;
        default : 
            WindowSetCursor(CURSOR_NORMAL);
            return RigiError(ERRORFLAG, "Unknown edit operation (%s)", argv[OPARG]);
    }
    rw->RigiDisplay(boxflag, p1, p2) ; 
    WindowSetCursor(CURSOR_NORMAL);
    return 1;
}

void CursorOp( int x, int y ) 
{
    RigiWindow* rw = RutilActiveWindow();

    // Since receiving request in window coordinates, convert them
    cursor_position.x = x;
    cursor_position.y = y;
    cursor_position = rw->Canvas2World(rw->Window2Canvas(cursor_position));
}

