
// $Author: jmartin $
// $Date: 2002-11-10 22:16:58 $
// $Source: /root/cvs/Rigi/rigiedit/operationops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: operationops.C,v 1.63 2002-11-10 22:16:58 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef RIGISTATE_H
#include "cl_rigistate.h"               // Global instance of editor's state
#endif
#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"         // Basic graph definitions
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Class definition for graph Nodes
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"       // Class definition for RigiWindows
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"         // Class definition for RigiNodes
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"           // Class definition for RigiDomain
#endif
#ifndef BASEWINDOW_H
#include "cl_base_window.h"       // Window menu state access functions
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef OPERATIONOPS_H
#include "operationops.h"
#endif
#ifndef NODEOPS_H
#include "nodeops.h"		// for NodeOpOpen()
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"                // Basic control functions for the editor
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"           // Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"            // Utility control functions for the editor
#endif
#ifndef RIGIERR_H
#include "rigierr.h"             // Error reporting functions
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"        // RCL interpreter stuff
#endif
#ifndef RSF_H
#include "rsf.h"
#endif

// Operations include files
#ifndef EDIT_H
#include "edit.h"                // Edit operation function definitions
#endif
#ifndef WINDOWGREP_H
#include "windowgrep.h"
#endif

#include "win_util.h"            // Window state access functions
#include "win_archierarchy.h"    // Arc hierarchy ops (depends on window system)

#ifndef MBI_H
#include "mbi.h"
#endif

// Local functions to this file

static int OperationOpenGeneral( int argc, char **argv, CharType& result );
static int OperationMakeTree( int argc, char **argv );
static int OperationGetContribs( int arcid, CharType& result );

const int WINALLOWP = 1<<0;
const int WINALLOWL = 1<<1;
const int WINALLOWC = 1<<2;
const int WINALLOWO = 1<<3;
const int WINALLOWG = 1<<4;
const int WINALLOWPG = WINALLOWP | WINALLOWG;
const int WINALLOWPO = WINALLOWP | WINALLOWO;
const int WINALLOWPOG = WINALLOWP | WINALLOWO | WINALLOWG; 
const int WINALLOWALL = WINALLOWP|WINALLOWL|WINALLOWC|WINALLOWO|WINALLOWG;
const int XFRAMEOFFSET = 20;
const int YFRAMEOFFSET = 20;

static RigiBool MenuAllowOp( RigiWindow* window, int allow ) {
    RigiWinType type = window->GetType();
    RigiBool OK = RigiFalse;
    switch( type )
    {
        case WIN_SCRATCH:     if(allow&WINALLOWC) OK = RigiTrue; break;
        case WIN_PROJECT:     if(allow&WINALLOWP) OK = RigiTrue; break;
        case WIN_OVERVIEW:    if(allow&WINALLOWO) OK = RigiTrue; break;
        case WIN_GENERAL:     if(allow&WINALLOWG) OK = RigiTrue; break;
    }
    if (!OK)
        return RigiError(ERRORFLAG, "Operation cannot be performed in this type of window");
    return RigiTrue;
}

int OperationOp( int argc, char **argv, CharType& result ) {
    int rvalue = 0;

    // Set the cursor to the WAIT cursor

    WindowSetCursor( CURSOR_WAIT );

    // Get the function name and confirm its consistency

    if ( strncmp( argv[0], "__operation", 11) != 0 )
    {
        result.Assign( "open_general: Invalid command " );
        result.Append( argv[0] );
        result.Append( ", expected " );
        result.Append( "__operation" );
        WindowSetCursor( CURSOR_NORMAL );
        return rvalue;
    }

    // Get the operation to perform

    int op = atoi( argv[1] );
    switch( op )
    {
        case OPENGENERAL:
            rvalue = OperationOpenGeneral(argc, argv, result);
            break;
        case MAKETREE:
        case MAKERTREE:
            rvalue = OperationMakeTree(argc, argv);
            break;
        case GETCONTRIBS:
            rvalue = OperationGetContribs( atoi(argv[2]), result);
            break;
        default:
            result.Assign( "Operation: Invalid operation flag " );
            result.Append( argv[1] );
            rvalue = 0;
            break;
    }

    WindowSetCursor( CURSOR_NORMAL );
    return rvalue;
}

static int OperationGetContribs( int arcid, CharType& result )
{
    char str[20]; // Buffer for arc id to string conversion

    // Get the arc that we are dealing with.

    GraphArc *arc = editor->FindArc( arcid );
    if (arc == NULL)
    {
        result.Assign( "Operation: Could not find arc ");
        sprintf( str, "%d", arcid );
        result.Append( str );
        return 0;
    }

    // For each contributor of this arc, add the arc's id to the list

    int n = arc->NumContributorIds();
    int i, contrib;
    for( i=0 ; i<n ; i++ ) {
        contrib = arc->GetContributorId( i );
        sprintf( str, " %d ", contrib );
        result.Append( str );
    }
    return 1;
}

static int OperationMakeTree( int argc, char **argv )
{
    // Get some initial state information for the windows.

    RigiWindow* activewin = RutilActiveWindow();
    if (activewin==NULL) return 0;

    // Careful with this list. It is possible that an operation can result in
    // the deletion of RigiNodes. This list contains pointers to RigiNodes, so
    // the list may not point to valid objects after some operations. The ops
    // that you need to worry about are the edit operations, as they update
    // the windows. Most other ops just change the state of the nodes.

    RigiNodeList slist;
    editor->GetSelectionList( slist, activewin );
    int i = editor->NumSelectedNodes();

    // If not one node selected, produce an error.

    if ( i != 1 )
        return RigiError(ERRORFLAG, "Exactly one node must be selected");

    // Get the current domain

    RigiDomain *domain = RinitGetDomain();

    // Get the selected node

    AvlPos *P;
    GraphNode* n = editor->FirstSelectedNode(P);
    RigiNode *rn = activewin->FindRigiNode(n);
    if (!rn) { warn("no RigiNode for selected node"); return 0; }

    // Get the type of arc to use, the angle of rotation for the graph, and
    // the hierarchy depth for the tree

    int op = atoi( argv[1] );

    int arctype;
    if ( argc < 3 || strlen( argv[2] ) == 0 ) arctype = editor->GetArcType();
    else arctype = domain->ArcTypeFromName( argv[2] );
    if (arctype == -1)
        return RigiError(ERRORFLAG, "Unknown arc type %s", ((argc>=3)? argv[2] : ""));

    double angle;
    if ( argc < 4 || strlen(argv[3]) == 0 ) angle = 0.0;
    else angle = atof( argv[3] );

    int depth = editor->GetHierarchyDepth();

    // Get the root nodes position

    Point2d p = rn->GetPosition();
    Point tmpp = activewin->World2Canvas( p );
    if (op == MAKETREE) {
        WindowArcHierarchy(arctype, ARC_OUT, activewin,
                            rn, tmpp.x, tmpp.y, depth, angle );
    } else if (op == MAKERTREE) {
        WindowArcHierarchy(arctype, ARC_IN, activewin,
                            rn, tmpp.x, tmpp.y, depth, angle );
    }
    RutilRaiseWindow(activewin);
    activewin->RigiDisplay();
    return 1;
}

static int OperationOpenGeneral(int argc, char **argv, CharType& result) {
    int tmp_argc;
    const char **tmp_argv;
    int error, i, tmp;
    int incl, level, x, y;
    char *label;
    NodeSet nodes;
    ArcFlags intypes;
    ArcFlags outtypes;
    RigiDomain *domain = RinitGetDomain();

    // Get some initial state information for the windows.
    RigiWindow* activewin = RutilActiveWindow();
    if (activewin == NULL) return 0;


    // Get the operation value, should be OPENGENERAL

    if ( atoi( argv[1] ) != OPENGENERAL ) {
        result.Assign( "open_general: Invalid operation flag " );
        result.Append( argv[1] );
        return 0;
    }

    // Get the node list

    error = Tcl_SplitList( rcl_interp, argv[2], &tmp_argc, &tmp_argv );
    if ( error == TCL_ERROR ) {
        result.Assign( "open_general: Could not extract node list from " );
        result.Append( argv[2] );
        return 0;
    }
    if ( tmp_argc == 0 ) {
        result.Assign( "open_general: Must have one node selected for Open" );
        return 0;
    }
    for( i=0 ; i<tmp_argc; i++ ) {
        tmp = atoi( tmp_argv[i] );
        Mess( "open_general: node id = %d", tmp );
        GraphNode *node = editor->FindNode( tmp );
        if ( node == NULL ) {
            result.Assign( "open_general: Could not find node " );
            result.Append( tmp_argv[i] );
            return 0;
        }
        nodes.Add( node );
    }
    free( (char*)tmp_argv );

    // Get the arc type list

    intypes.ClearAll();
    error = Tcl_SplitList( rcl_interp, argv[3], &tmp_argc, &tmp_argv );
    if ( error == TCL_ERROR ) {
        result.Assign( "open_general: Could not extract arc type list from " );
        result.Append( argv[3] );
        return 0;
    }
    for( i=0 ; i<tmp_argc; i++ ) {
        tmp = domain->ArcTypeFromName( (char*) tmp_argv[i] );
        if ( tmp == -1 ) {
            result.Assign( "open_general: Invalid arc type " );
            result.Append( tmp_argv[i] );
            return 0;
        }
        if ( tmp == ANY_ARC ) intypes.SetAll();
        else                  intypes.SetBit( tmp );
    }
    free( (char*)tmp_argv );

    // Get the arc type list

    outtypes.ClearAll();
    error = Tcl_SplitList( rcl_interp, argv[4], &tmp_argc, &tmp_argv );
    if ( error == TCL_ERROR ) {
        result.Assign( "open_general: Could not extract arc type list from " );
        result.Append( argv[4] );
        return 0;
    }
    for( i=0 ; i<tmp_argc; i++ ) {
        tmp = domain->ArcTypeFromName( (char*) tmp_argv[i] );
        if ( tmp == -1 ) {
            result.Assign( "open_general: Invalid arc type " );
            result.Append( tmp_argv[i] );
            return 0;
        }
        if ( tmp == ANY_ARC ) outtypes.SetAll();
        else                  outtypes.SetBit( tmp );
    }
    free( (char*)tmp_argv );

    incl = atoi( argv[5] );

    // Get the projection level

    level = atoi( argv[6] );
    Mess("level = %d", level);

    // Get the window position

    x = atoi(argv[7]);
    y = atoi(argv[8]);

    // Get the label
    label = argv[9];

    // Create a RigiWindow
    RigiWindow *rw = new RigiWindow( nodes, intypes, outtypes, incl, level,
        x, y, label );
    if (editor->GetInheritFilters()) rw->InheritTypeFilters(activewin);

    editor->AddWindow(rw);
    RutilSetActiveWindow(rw);
    rw->Show();
    RutilRaiseWindow(rw);
    rw->RigiDisplay();

    return 1;
}

// SCRIPTRETURN, usually number of nodes
// string used by grep and IO operations
// select_keep used by FINDNAME and GREP

int OperationOp( int op, char* string, int select_keep )
{
    WindowSetCursor( CURSOR_WAIT );

    // Get some initial state information for the windows.
    RigiWindow* activewin = RutilActiveWindow();
    if (activewin == NULL) return 0;

    // Careful with this list. It is possible that an operation can result in
    // the deletion of RigiNodes. This list contains pointers to RigiNodes, so
    // the list may not point to valid objects after some operations. The ops
    // that you need to worry about are the edit operations, as they update
    // the windows. Most other ops just change the state of the nodes.

    RigiNodeList slist;
    if (op == MAKETREE || op == MAKERTREE || op == PROJECTSYS)
        editor->GetSelectionList( slist, activewin );
    int i = editor->NumSelectedNodes();

    // Get the current domain

    RigiDomain *domain = RinitGetDomain();

    switch (op)
    {
        case MAKETREE: 
        case MAKERTREE:
        {
            if ( i != 1 ) { 
                RigiError(ERRORFLAG, "Exactly one node must be selected");
                break;
            }
                        
            if ( !MenuAllowOp( activewin,WINALLOWPG ) ) break;

            AvlPos *P;
            GraphNode* n = editor->FirstSelectedNode(P);
            RigiNode *rn = activewin->FindRigiNode(n);
            if (!rn) { warn("no RigiNode for selected node"); break; }

            Point tmpp;
            int arctype;
            if ( strlen( string ) == 0 ) arctype = editor->GetArcType();
            else arctype = domain->ArcTypeFromName( string );
            if (arctype == -1) {
                RigiError(ERRORFLAG, "Unknown arc type %s", string);
                break;
            }
            int depth = editor->GetHierarchyDepth();

            Point2d p = rn->GetPosition();
            tmpp = activewin->World2Canvas( p );
            if (op == MAKETREE) {
                WindowArcHierarchy(arctype, ARC_OUT, activewin,
                                    rn, tmpp.x, tmpp.y, depth );
            } else if (op == MAKERTREE) {
                WindowArcHierarchy(arctype, ARC_IN, activewin,
                                    rn, tmpp.x, tmpp.y, depth );
            }

            RutilRaiseWindow(activewin);
            activewin->RigiDisplay(); 
            break;
        }
        case PROJECTSYS:
        {
            if (!MenuAllowOp( activewin,WINALLOWPOG)) break;
            int num = 0;
            NodeSet nodes;

            AvlPos *P;
            GraphNode* n = editor->FirstSelectedNode(P);
            while ( n != NULL )
            {
                num++;
                nodes.Add( n );
                n = editor->NextSelectedNode(P);
            }
            if ( num == 0 )
            {
                RigiError(ERRORFLAG, "At least one node must be selected");
                break;
            }
            int x,y;
            int depth = editor->GetProjectionDepth();
            activewin->Window()->Location(x,y); 
            x += XFRAMEOFFSET; y += YFRAMEOFFSET; 

            // Set the bits for the types of arcs to traverse for the open
            // operation.

            RigiWindow *rw = new RigiWindow( nodes, WIN_PROJECT, depth,
                        domain->NumNodeTypes(),domain->NumArcTypes(),x,y);
            if (editor->GetInheritFilters()) rw->InheritTypeFilters(activewin);
            editor->AddWindow(rw);

            // Show the new window
            rw->Show();

            // make the new window active
            RutilSetActiveWindow(rw);
            RutilRaiseWindow(rw);
            rw->RigiDisplay();
            break;
        }

        case FINDNAME:
        case GREP:
        {
            int (*ifp)(char*) = (op==GREP ? window_grep : window_find);
            if (!select_keep) {
                activewin->DrawSynch(RigiTrue);
                activewin->BunchNode(RigiTrue);
                editor->EmptySelection();
                activewin->BunchNode(RigiFalse);
                activewin->DrawSynch(RigiFalse);
            }
            i = (*ifp)(string);

            char mess[32];
            sprintf( mess, "%d matching node%s found",i,i==1?"":"s");
            activewin->Window()->SetMessage( mess );

            break;
        }
        case OPENTYPED:
        {
            NodeSet nodes;
            int x, y;

            // Grab the currently active windows location.

            activewin->Window()->Location(x,y);
            x += XFRAMEOFFSET; y += YFRAMEOFFSET;

            // Add the list of nodes that are selected to the local list.

            int num = 0;
            AvlPos *P;
            GraphNode *np = editor->FirstSelectedNode(P);
            while( np != NULL ) {
                // Keep track of the number of nodes we have found and
                // add the node to the local list.

                num++;
                nodes.Add( np );
                np = editor->NextSelectedNode(P);
            }
            if (!num) {
                RigiError(ERRORFLAG, "At least one node must be selected");
                break;
            }

            // Set the bits for the types of arcs to traverse for the open
            // operation.

            ArcFlags types;
            int current_type = editor->GetArcType();
            if (current_type == domain->ArcTypeFromName("any")) {
                types.SetAll();
                types.ClearBit( domain->ArcTypeFromName("refractive") );
            }
            else types.SetBit( current_type );

            // Create the new RigiWindow. Set the inclusive flag to 1 and the
            // number of levels to traverse to 1.

            RigiWindow* rw = new RigiWindow(nodes, types, ARC_OUT, 0, 1, x, y);

            if (editor->GetInheritFilters()) rw->InheritTypeFilters(activewin);

            // Add the window to the editor and display it.

            editor->AddWindow(rw);
            RutilSetActiveWindow(rw);
            rw->Show();
            RutilRaiseWindow(rw);
            rw->RigiDisplay();

            // For ease of operations from scripts, make the newly
            // opened window the active window

            break;
        }
        case OPENPARENT:
        {
                GraphNode *n = activewin->GetNode();
                GraphArc *a = n->FirstArc( inLevArcs );
                if (!a) {
                    RigiError(ERRORFLAG, "This node has no parent");
                    break;
                }
                n = a->GetSrc();
                NodeOpOpenNode(n);
                break;
        }
        case READRSF:
                i = ReadRSF( string );
                break;
        case WRITERSF:
                i = WriteRSF( string );
                break;
        case LOADRSF:
                if ((i = LoadRSF( string, activewin )) != -1) {
                    extern void edit_check_composite( GraphNode* );
                    Mess( "Resetting composite arcs" );
                    edit_check_composite( RigiBaseNode() );
                    Mess( "Done resetting composite arcs" );
                }
                break;
        case SAVERSF:
                i = SaveRSF( string, activewin );
                break;
        default: 
            mayday( "bad op %d in OperationOp()",op );
            break;
    }
    WindowSetCursor( CURSOR_NORMAL );
    return i; 
}

