

/**********************************************************************/
/*                                                                    */
/*                                                                    */
/*                SPRINGEMBEDDER-LAYOUT ALGORITHM                     */
/*                                                                    */
/*                                                                    */
/*                 programmed by Claus-Dieter Ruscha                  */
/*                                                                    */
/*                               at                                   */
/*                                                                    */
/*                      University  of Passsau                        */
/*                                                                    */
/*                       Prof. Dr. Brandenburg                        */
/*                                                                    */
/*                             1992                                   */
/*                                                                    */
/*                                                                    */
/*                     based on a paper by                            */
/*                                                                    */
/*              Thomas Fruchterman and Edward Reingold                */
/*                                                                    */
/*                                                                    */
/*                                                                    */
/**********************************************************************/


#include <std.h>
#include <sgraph.h>
#include <slist.h>
#include <math.h>
#include "springembedder_h.h"


#define PROC void
#define MAIN void
#define FUNC /**/
#define mark(node) attr_flags(node)
#define weight(edge) *((float*) ((edge)->attrs.data))
#define has_reflexive_edge attr_flags
#define reflexive_edge(edge) (edge->snode->nr == edge->tnode->nr)
#define for_some_nodes(from_node, to_node)     do {
#define end_for_some_nodes(from_node, to_node) } while ( ((from_node)=(from_node)->suc) != (to_node)->suc )
#define for_all_nodes_exept_last_node(graph, node)     if ( ((node)=(graph->nodes)) != (Snode)nil)  do {
#define end_for_all_nodes_exept_last_node(graph, node) } while ( ((node)=(node)->suc) != graph->nodes->pre )
#define empty_string(string) (*(string) == '\0')





/**********************************************************************/
/*                                                                    */
/*             globale Variable und Datenstrukturen                   */
/*                                                                    */
/**********************************************************************/


bool    springembedder_draw_weighted_edges     = FALSE;
float	springembedder_max_force               = 3.0;
/* MAD: changed opt_distance to 258.0 from 64.0, nicer layout
   should be parametrized! */
float	springembedder_opt_distance            = 258.0;
float	springembedder_vibration               = 0.4;
int	springembedder_max_iterations          = 1000;
int	springembedder_animation_intervals     = 10;



typedef struct 
{       float x, y;
} vektor;





/**********************************************************************/
/*                                                                    */
/*                    P r o z e d u r e n                             */
/*                                                                    */
/**********************************************************************/



PROC unmark_all_nodes ( graph )
     /*****************************************************************/
     /* setzt die Markierungen aller Knoten von graph auf FALSE       */
     /*****************************************************************/
     Sgraph graph;
{    Snode  node;
     for_all_nodes( graph, node )
     {   mark( node ) = FALSE;
     }   end_for_all_nodes( graph, node ); 
}


PROC mark_connected_nodes ( graph, node )
     /*****************************************************************/
     /* markiert in graph die mit node durch Kanten verbundene Knoten */
     /*****************************************************************/
     Sgraph graph;
     Snode  node;
{    Sedge  edge;
     if ( mark( node ) == FALSE )
     {  mark( node ) = TRUE;
        for_sourcelist( node, edge )
        {   mark_connected_nodes( graph, edge->tnode );
        }   end_for_sourcelist( node, edge );
        if ( graph->directed ) 
        {  for_targetlist( node, edge )
           {   mark_connected_nodes( graph, edge->snode );
           }   end_for_targetlist( node, edge );
        }     
     }       
}


FUNC bool is_connected ( graph )
     /*****************************************************************/
     /* ueberprueft, ob graph zusammenhaengend ist (oder aus mehreren */
     /* untereinander nicht verbundenen Teilgraphen besteht)          */
     /*****************************************************************/
     Sgraph graph;
{    Snode  node;
     bool   connected;
     unmark_all_nodes( graph );
     mark_connected_nodes( graph, first_node_in_graph( graph ) );
     connected = TRUE;
     for_all_nodes( graph, node )
     {   if ( ! mark( node ) )
            connected = FALSE;
     }   end_for_all_nodes( graph, node );
     return connected;
}


FUNC bool multiple_connected_nodes ( graph )
     /*****************************************************************/
     /* ueberprueft, of graph Knoten besitzt, die durch mehr als eine */
     /* Kante verbunden sind. Eventuell mehrfach vorhandene reflexive */
     /* Kanten werden -obwohl ihre Existenz semantisch nicht sinnvoll */
     /* ist - dabei nicht reklamiert, da sie den Ablauf des Algorth.  */  
     /* nicht stoeren                                                 */
     /*****************************************************************/
     Sgraph graph;
{    Snode  node;
     Sedge  edge;
     bool multiple_connected = FALSE;
     int node_index = 0;
     unmark_all_nodes( graph );
     for_all_nodes( graph, node )
     {   node_index++;
	 for_sourcelist( node, edge )
	 {   if ( ! reflexive_edge( edge ) )
             {  if ( mark( edge->tnode ) == node_index )
	           multiple_connected = TRUE;
                else
	           mark( edge->tnode ) = node_index;
             }
         }   end_for_sourcelist( node, edge );
         if (graph->directed)
         {  for_targetlist( node, edge )
 	    {   if ( ! reflexive_edge( edge ) )
                {  if ( mark( edge->snode ) == node_index )
                      multiple_connected = TRUE;
                   else
                      mark( edge->snode ) = node_index;
                }
            }   end_for_targetlist( node, edge );
	 }
     }   end_for_all_nodes( graph, node );
     return multiple_connected;
}


FUNC bool edge_weights_succesfully_read ( graph, enough_memory )
     /*****************************************************************/
     /* liest die Gewichte der Kanten aus dem Feld edge->label in das */
     /* attrs - Feld der Kanten ein, ueberprueft dabei deren Korrekt- */
     /* heit und gibt in der Variablen enough_memory an, ob genug     */
     /* Speicherplatz dabei vorhanden war                             */
     /*****************************************************************/
     Sgraph graph;
     bool   *enough_memory;
{    Snode  node;
     Sedge  edge;
     float  *pfloat;
     bool succesfully_read = TRUE;
     for_all_nodes( graph, node )
     {   for_sourcelist( node, edge)
         {   pfloat = (float*) malloc( (unsigned) (sizeof(float)) );
             if ( pfloat != nil )
             {  if ( (edge->label == nil) || empty_string(edge->label) )
                {  *pfloat = 1.0;
                }
                else
                {  *pfloat = (float) atof( edge->label );
                   if (*pfloat <= 0.0)
                   {  *pfloat = 1.0;
                      succesfully_read = FALSE;
                   }
                } 
                edge->attrs = make_attr(ATTR_DATA, (char*) pfloat);
             } 
             else
             {  *enough_memory = FALSE;
                succesfully_read = FALSE; 
             }
         }   end_for_sourcelist( node, edge );
     }   end_for_all_nodes( graph, node );
     return succesfully_read;
}


FUNC bool neutral_edge_weights_succesfully_written( graph )
     /*****************************************************************/
     /* falls draw_weighted_layout (im Fenster) auf off gesetzt wird  */
     /* wird in das attrs-Feld der Kanten der neutrale Wert 1 geschr. */
     /*****************************************************************/
     Sgraph graph;
{    Snode  node;
     Sedge  edge;
     float  *pfloat;
     bool succesfully_written = TRUE;
     for_all_nodes( graph, node )
     {   for_sourcelist( node, edge)
         {   pfloat = (float*) malloc( (unsigned) (sizeof(float)) );
             if ( pfloat != nil )
             {  *pfloat = 1.0;
                edge->attrs = make_attr(ATTR_DATA, (char*) pfloat);
             } 
             else
                succesfully_written = FALSE; 
         }   end_for_sourcelist( node, edge );
     }   end_for_all_nodes( graph, node );
     return succesfully_written;

}


FUNC bool is_correct ( graph )
     /*****************************************************************/
     /* ueberprueft, ob graph in einer korrekten Form vorliegt, um    */   
     /* von dem Nature-Algorithmus bearbeitet zu werden               */ 
     /*****************************************************************/
     Sgraph graph;
{    bool correct = FALSE, enough_memory = TRUE;
     if ( graph == empty_graph || graph->nodes == nil )
     {       fprintf ( stderr, "No graph selected\n" );
		exit(-1);
     }
     else if ( ! is_connected( graph ) )
     {       fprintf( stderr, "Graph is not connected\n" );
		exit(-1);
     }  
     else if ( multiple_connected_nodes( graph ) ) {       
       fprintf( stderr, "Graph contains multiple edge(s).\n" );
       exit(-1);
     }
     else if ( springembedder_draw_weighted_edges && (! edge_weights_succesfully_read( graph, &enough_memory)) )
     {       if ( ! enough_memory )
                fprintf( stderr, "Springembedder: not enough memory avaiable !\n" );
             else
                fprintf( stderr, "on or several edge-weights are incorrect\n" );
		exit(-1);
     }
     else if (! springembedder_draw_weighted_edges && (! neutral_edge_weights_succesfully_written( graph )) )
     {       fprintf(stderr,  "Springembedder: not enough memory avaiable !\n" );
		exit(-1);
     }
     else {
       correct = TRUE;
     }

     return correct;
}	 


PROC straighten_edges ( sgraph )
     /*****************************************************************/
     /* "biegt" eventuell vorhandene nicht geradlienig verlaufende    */
     /* Kanten im Graphen gerade, da der Nature-Algorithmus gerad-    */
     /* lienig verlaufende Kanten vorraussetzt. Reflexive Kanten wer- */
     /* den dabei nicht begradigt. Knoten, die reflexive Kanten ent-  */
     /* halten, werden als solche markiert, um spaeter schneller die  */
     /* reflexiven Kanten wiederzufinden.                             */
     /*****************************************************************/
     Sgraph sgraph;
{    Snode  snode;
     Sedge  sedge;
     int x1, y1, x2, y2;
     unmark_all_nodes( sgraph );
     for_all_nodes( sgraph, snode )
     {   for_sourcelist( snode, sedge )
         {   if ( ! reflexive_edge ( sedge ) )
             {  
/*
		gedge = graphed_edge( sedge );
                edgeline = ( Edgeline ) edge_get( gedge, EDGE_LINE );
                if ( ! is_single_edgeline( edgeline ) )
                {  free_edgeline( edgeline );
                   x1 = snode_x( snode ); x2 = sedge->tnode->x;
                   y1 = snode_y( snode ); y2 = sedge->tnode->y; 
                   straight_edgeline = add_to_edgeline( new_edgeline( x1, y1 ), x2, y2 );
                   edge_set( graphed_edge( sedge ), EDGE_LINE, straight_edgeline, 0 ); 

                }
*/
             }
             else
                has_reflexive_edge( snode ) = TRUE;
          }  end_for_sourcelist( snode, sedge );
          if (sgraph->directed) 
	  {  for_targetlist( snode, sedge )
	     {   if ( ! reflexive_edge ( sedge ) )
                 {
/*
		    gedge = graphed_edge( sedge );
	            edgeline = ( Edgeline ) edge_get( gedge, EDGE_LINE );
                    if ( ! is_single_edgeline( edgeline ) )
                    {  free_edgeline( edgeline );
 		       x1 = snode_x( snode ); x2 = sedge->snode->x;
		       y1 = snode_y( snode ); y2 = sedge->snode->y; 
		       straight_edgeline = add_to_edgeline( new_edgeline( x1, y1 ), x2, y2 );
		       edge_set( graphed_edge( sedge ), EDGE_LINE, straight_edgeline, 0 ); 
                    }
*/
                 }
	      }  end_for_targetlist( snode, sedge );
	   }
     }     end_for_all_nodes( sgraph, snode );
}


PROC number_nodes ( graph )
     /*****************************************************************/
     /* nummeriert die Knoten von graph durck                         */
     /*****************************************************************/
     Sgraph graph;
{    Snode  node;
     int    number=0;
     for_all_nodes( graph, node )
     {   node->nr = number++;
     }   end_for_all_nodes ( graph, node );
}


FUNC int number_of_nodes ( graph )
     /*****************************************************************/
     /* stellt fest, wieviel Knoten graph bestitzt                    */
     /*****************************************************************/
     Sgraph graph;
{    Snode  node;
     int n;
     n = 0;
     for_all_nodes( graph, node )
     {   n++;
     }   end_for_all_nodes( graph, node );
     return n;
}


FUNC vektor *init_node_positions ( graph, N, enough_memory )
     /*****************************************************************/
     /* gibt ein Pointer auf ein Array der Laenge N mit Eintraegen    */
     /* vom Typ vektor zurueck, indem die Koordinaten der Knoten von  */
     /* graph stehen und ordnet jedem 'node' von 'graph' einen Index  */
     /* i zu, so dass gilt : node_positions[i] gehoert zum i-ten      */
     /* 'node' der Knotenliste (wird fuer compute_forces benoetigt)   */
     /*****************************************************************/    
     Sgraph graph;
     int    N;
     bool   *enough_memory;
{
	Snode  node;
	vektor *node_positions;
	int    i;
	
	node_positions = (vektor*) malloc( (unsigned) ( N * sizeof( vektor ) ) );
	
	if ( node_positions == nil )
        {  *enough_memory = FALSE;
	}
        else
        {  i = 0;
           for_all_nodes( graph, node )
           {   node_positions[i].x = snode_x( node );
               node_positions[i].y = snode_y( node );
               mark( node ) = i;
               i++; 
           }   end_for_all_nodes( graph, node );
	};
	return node_positions;
}


PROC init_null ( forces, N )
     /*****************************************************************/
     /* initialisiert das Array forces der Laenge N vom Typ vektor    */
     /* mit null - Werten                                             */
     /*****************************************************************/
     vektor *forces;
     int N;
{    int i;
     for ( i = 0; i < N; i++ )
     {   forces[i].x = 0.0;
         forces[i].y = 0.0;
     }
}


FUNC vektor *init_forces ( N, enough_memory )
     /*****************************************************************/
     /* gibt ein Pointer auf ein Array der Laenge N mit Eintraegen    */
     /* vom Typ vektor zurueck, die alle Koordinaten den Wert 0       */
     /* haben                                                         */
     /*****************************************************************/
     int N;
     bool *enough_memory;
{    vektor *forces;
     forces = (vektor*) malloc( (unsigned) ( N * sizeof( vektor ) ) );
     if ( forces == nil )
     {  *enough_memory = FALSE;
     }
     else /* else-Zweig nicht unbedingt erforderlich */
        init_null( forces, N );
     return forces;
}


FUNC  float vek_abs ( v )
     /*****************************************************************/
     /* berechnet die euklidische Norm fuer ein Element von Typ       */
     /* vektor, die ja der Laenge des vektors entspricht              */
     /*****************************************************************/
      vektor v;
{     return   hypot( v.x, v.y ); /* hypot( x, y ) = sqrt( x*x + y*y ) */
}


FUNC float compute_max_force ( forces, N )
     /*****************************************************************/
     /* bestimmt die groesste Kraft aus dem Array forces der Laenge N */
     /*****************************************************************/
     vektor *forces;
     int N;
{    float abs_force, max_abs_force;
     int i;
     max_abs_force = 0;
     for ( i = 0; i < N; i++ )
     {   abs_force = vek_abs ( forces[i] );
         if ( abs_force > max_abs_force )
            max_abs_force = abs_force;
     }
     return max_abs_force;
}
 
   
FUNC vektor add ( v1, v2 )
     /*****************************************************************/
     /* addiert zwei Elemente von Type vektor komponentenweise        */
     /*****************************************************************/
     vektor v1, v2;
{    vektor v;
     v.x = ( v1.x + v2.x );
     v.y = ( v1.y + v2.y );
     return v;
}


FUNC vektor sub ( v1, v2 )
     /*****************************************************************/
     /* substrahiert zwei Elemente von Type vektor komponentenweise   */
     /*****************************************************************/
     vektor v1, v2;
{    vektor v;
     v.x = ( v1.x - v2.x );
     v.y = ( v1.y - v2.y );
     return v;
}


FUNC vektor sca_diff ( v, scalar )
     /*****************************************************************/
     /* diffidiert ein Element vom Typ vektor mit einem Element vom   */
     /* Typ float durch uebliche komponentenweise Division            */
     /*****************************************************************/
     vektor v;
     float scalar;
{    v.x = v.x / scalar;
     v.y = v.y / scalar;
     return v;
}


FUNC vektor center ( node_positions, N )
     /*****************************************************************/
     /* berechnet den Mittelpunkt der Knotenpositionen aus dem Array  */
     /* node_positions der Laenge N                                   */
     /*****************************************************************/
     vektor *node_positions;
     int N;
{    vektor center; 
     int i;
     center.x = 0.0;
     center.y = 0.0;
     for ( i = 0; i < N; i++ )
         center = add( center, node_positions[i] );
     return sca_diff( center, ( (float) N ) ); 
}


PROC compute_forces ( optimal_distance, forces, node_positions, graph )
     /*****************************************************************/
     /* berechnet die Kraefte, die an den Knoten des Graphen graph    */
     /* wirken. Dabei werden die Koordinaten der Knoten im Array      */
     /* node_position uebergeben. Das Ergebnis wird im Array forces   */
     /* zurueckgegeben. Die Berechnung der Kraefte ist nach Meinung   */
     /* des Programmierers sehr gut optimiert.                        */
     /*****************************************************************/
     float  optimal_distance;
     vektor *forces, *node_positions;
     Sgraph graph;
{    Snode  node1,node2;
     Sedge  edge;
     vektor attractive_force,  repulsive_force, force, delta;
     float  attractive_factor, repulsive_factor;
     float  sum_sqr_delta_xy, sqr_optimal_distance = optimal_distance * optimal_distance;
     int    node_index_1, node_index_2;
     for_all_nodes_exept_last_node( graph, node1 )
     {   
         /* compute edge forces */
         node_index_1 = node1->nr;
         for_sourcelist( node1, edge )
         {   if ( ! reflexive_edge (edge) && (node1->nr < edge->tnode->nr) )
             {  node_index_2 = edge->tnode->nr;
                delta.x = node_positions[node_index_2].x - node_positions[node_index_1].x;
                delta.y = node_positions[node_index_2].y - node_positions[node_index_1].y;
                if ((delta.x == 0) && (delta.y == 0))  /* Prevent 0 distance, which is considered unnatural */
                {  delta.x = 1;
                   delta.y = 1;
                } 
                sum_sqr_delta_xy = delta.x * delta.x + delta.y * delta.y;

                attractive_factor  = sqrt(sum_sqr_delta_xy) / ( weight(edge) * optimal_distance );
                attractive_force.x = attractive_factor * delta.x;
                attractive_force.y = attractive_factor * delta.y;

                repulsive_factor  = weight(edge) * weight(edge) * sqr_optimal_distance / sum_sqr_delta_xy;
                repulsive_force.y = repulsive_factor * delta.y;
                repulsive_force.x = repulsive_factor * delta.x;

                force.x = attractive_force.x - repulsive_force.x;
                force.y = attractive_force.y - repulsive_force.y;

                forces[node_index_1].x += force.x;
                forces[node_index_1].y += force.y;
                forces[node_index_2].x -= force.x;
                forces[node_index_2].y -= force.y;

                mark (edge->tnode) = node_index_1;
             }
         }   end_for_sourcelist( node1, edge );
         if (graph->directed)
         {  for_targetlist( node1, edge )
            {   if ( ! reflexive_edge (edge) && ( node1->nr < edge->snode->nr ) )
                {  node_index_2 = edge->snode->nr;
                   delta.x = node_positions[node_index_2].x - node_positions[node_index_1].x;
                   delta.y = node_positions[node_index_2].y - node_positions[node_index_1].y;
                   if ((delta.x == 0) && (delta.y == 0))  /* Prevent 0 distance, which is considered unnatural */
                   { delta.x = 1;
                     delta.y = 1;
                   } 
                   attractive_factor  = hypot (delta.x, delta.y) / ( weight(edge) * optimal_distance );
                   attractive_force.x = attractive_factor * delta.x;
                   attractive_force.y = attractive_factor * delta.y;

                   repulsive_factor  = weight(edge) * weight(edge) * sqr_optimal_distance / (delta.x * delta.x + delta.y * delta.y);
                   repulsive_force.x = repulsive_factor * delta.x;
                   repulsive_force.y = repulsive_factor * delta.y;

                   force.x = attractive_force.x - repulsive_force.x;
                   force.y = attractive_force.y - repulsive_force.y;

                   forces[node_index_1].x += force.x;
                   forces[node_index_1].y += force.y;
                   forces[node_index_2].x -= force.x;
                   forces[node_index_2].y -= force.y;

                   mark (edge->snode) = node_index_1;
                }
            }   end_for_targetlist( node1, edge );
         }

        /* compute_remaining_forces */
        node2 = node1->suc;
        for_some_nodes ( node2, last_node_in_graph(graph) )
        {   if ( mark (node2) != node_index_1 )
            {  node_index_2 = node2->nr;
               delta.x = node_positions[node_index_2].x - node_positions[node_index_1].x;
               delta.y = node_positions[node_index_2].y - node_positions[node_index_1].y;
               if ((delta.x == 0) && (delta.y == 0))  /* Prevent 0 distance, which is considered unnatural */
               {  delta.x = 1;
                  delta.y = 1;
               } 
               repulsive_factor  = sqr_optimal_distance / (delta.x * delta.x + delta.y * delta.y);
               repulsive_force.x = repulsive_factor * delta.x;
               repulsive_force.y = repulsive_factor * delta.y;

               forces[node_index_1].x -= repulsive_force.x;
               forces[node_index_1].y -= repulsive_force.y;
               forces[node_index_2].x += repulsive_force.x;
               forces[node_index_2].y += repulsive_force.y;
            }
        }   end_for_some_nodes (node2, last_node_in_graph(graph) );

     }  end_for_all_nodes_exept_last_node( graph, node1 );

}

    
PROC compute_new_node_positions ( node_positions, forces, N, temperature )
     /*****************************************************************/
     /* berechnet entsrechend der an den Knoten wirkenden Kraefte     */
     /* (forces) und der maximal zulaessigen Verschiebung der Knoten  */
     /* (temperature) die neuen Knotenpositionen                      */
     /*****************************************************************/
     vektor *node_positions, *forces;
     int N;
     float temperature;
{    vektor shift;
     float abs_force;
     int i;
     for ( i = 0; i < N; i++ )
     {   abs_force = vek_abs (forces[i]);
         if ( abs_force < temperature)
         {  shift.x = forces[i].x;
            shift.y = forces[i].y;
         }
         else
         {  shift.x = forces[i].x / abs_force * temperature;
            shift.y = forces[i].y / abs_force * temperature;
         }
         node_positions[i].x += shift.x;
         node_positions[i].y += shift.y;
     }
}

static int   cool_phase;
static float cool_max_force_1,
             cool_max_force_2,
             cool_max_force_3,
             cool_max_force_4;

FUNC float cool ( max_force )
     /*****************************************************************/
     /* berechnet abhaengig von max_force die fuer diesen Schleichen- */
     /* durchlauf maximal zulaessige Verschiebung (temperature) der   */
     /* Knoten. Nachdem der Graph angefangen hat zu schwingen, wird   */
     /* in Phase 2 die maximale zulaessige Verschiebung erheblich     */
     /* niedriger gesetzt, so das eine Schwingung nicht mehr auftritt.*/
     /*****************************************************************/
     float max_force;
{
     static float temperature;
     cool_max_force_4 = cool_max_force_3;
     cool_max_force_3 = cool_max_force_2;
     cool_max_force_2 = cool_max_force_1;
     cool_max_force_1 = max_force;

     /* Test ob Graph schwingt */
     if ( (cool_max_force_4 != 0.0) &&
          (fabs((cool_max_force_1 - cool_max_force_3) / cool_max_force_1) < springembedder_vibration) &&
          (fabs((cool_max_force_2 - cool_max_force_4) / cool_max_force_2) < springembedder_vibration) )
     {  cool_phase = 2;
     }
     if ( cool_phase == 1 )
     {  temperature = sqrt( max_force );
     }
     else 
     {  /* cool_phase 2 */
        temperature = sqrt( max_force ) / 15;
     }
     return temperature;
}

static	springembedder_animation = FALSE;




PROC write_node_positions_back_to_graph ( node_positions, old_center, N, graph )
     /*****************************************************************/
     /* schreibt die Koordinaten der Knoten des S-Graphen graph zu-   */
     /* rueckt in den entsprechenden Graphed-Graphen. Dabei werden    */
     /* die Knoten so verschoben, dass der Mittelpunkt des Graphen    */
     /* wieder auf dem urspruenglichen Mittelpunkt (old_center) zu    */
     /* liegen kommt. Besitzt ein Knoten reflexive Kanten so werden   */
     /* diese mit dem Knoten mitverschoben.                           */
     /*****************************************************************/

     vektor *node_positions, old_center;
     int N;
     Sgraph graph;
{
     Snode  node;
     Sedge  sedge;
     int i = 0;
     
     for_all_nodes( graph, node )
     {
         node->x = (int)(node_positions[i].x);
	 node->y = (int)(node_positions[i].y);
	 i++;
     }   end_for_all_nodes( graph, node );
}



MAIN springembedder ( graph )
     /*****************************************************************/
     /*               Hauptporzedur des Algorithmus                   */
     /* Der Algorithmus wurde im wesentlichen nach einer Arbeit von   */
     /* T.Fruchterman und E. Reingold ( University of Illionis )      */
     /* implementiert. Das Abruchkriterium der Schleife und die max.  */
     /* Verschiebung der Knoten wurde so geaendert, dass eine be-     */
     /* liebig genaue Berechnung der Knotenkoordinaten moeglich ist   */   
     /*****************************************************************/
Sgraph graph;
{
     int	N; /* Anzahl der Knoten des von GraphEd uebergebenen Graphen */
     float	temperature; /* max. Verschiebung eines Knotens in einem Schritt */
     vektor	*node_positions, *forces;
     vektor	old_center;
     bool	enough_memory = TRUE;
     int	iteration_count = 0;
     int	animation_iteration_count = 0;
     float	optimal_distance, max_force;
     cool_phase       = 1;
     cool_max_force_1 = 0.0;
     cool_max_force_2 = 0.0,
     cool_max_force_3 = 0.0,
     cool_max_force_4 = 0.0;
	
     optimal_distance = (float)springembedder_opt_distance;

     if ( is_correct (graph ) )
     {  straighten_edges ( graph );
        number_nodes ( graph ); 
        N              = number_of_nodes( graph );
        node_positions = init_node_positions( graph, N, &enough_memory );
        old_center     = center( node_positions, N );
        forces         = init_forces( N, &enough_memory );
        if ( ! enough_memory )
        {  fprintf (stderr, "Springembedder : not enough memory avaiable !\n"); 
		exit(-1);
        } 
        else 
        {  init_null( forces, N ); 
           compute_forces( optimal_distance, forces, node_positions, graph );
           /* forces ist ein VAR-Parameter */
           max_force = compute_max_force( forces, N );
		   
           while (max_force > springembedder_max_force  &&
                 iteration_count < springembedder_max_iterations)
           {     init_null( forces, N );
                 compute_forces( optimal_distance, forces, node_positions, graph );
                 /*  forces ist ein VAR-Parameter  */
                 max_force   = compute_max_force( forces, N );
                 temperature = cool( max_force );
                 compute_new_node_positions( node_positions, forces, N, temperature );
                 /*  node_positions ist ein VAR-Parameter  */
	
                 iteration_count ++;

           }
           write_node_positions_back_to_graph ( node_positions, old_center, N, graph );
           free( (char*) node_positions ); free( (char*) forces );
        } 
     }
}


MAIN do_nothing_springembedder ()
     /*****************************************************************/
     /* Nur notwendig um im file springembedder_sf in der Prozedur    */
     /* char *springembedder_menu_callback_proc ( menu, menu_item )   */
     /* einen Wert zurueckzugeben.                                    */
     /*****************************************************************/
{    ;
}


MAIN fast_springembedder ( sgraph )
     /*****************************************************************/
     /* Normalversion des Algorithmus. Zwischenergebnisse werden      */
     /* nicht angezeigt.                                              */
     /*****************************************************************/
     Sgraph sgraph;
{    springembedder_animation = FALSE;
     springembedder (sgraph);
}





