//
//

// $Author: jmcdanie $
// $Date: 1996-10-12 00:06:11 $
// $Source: /root/cvs/Rigi/rigiedit/quality.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: quality.C,v 1.34 1996-10-12 00:06:11 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif

#ifndef CHARARRAYH
#include "adt/chararray.h"
#endif
#ifndef GRAPHARCSET_H
#include "cl_grapharcset.h"
#endif

#ifndef EXACT_H
#include "exact.h"
#endif
#ifndef QUALITY_H
#include "quality.h"
#endif
#ifndef RIGI_H
#include "rigi.h"
#endif

#ifndef GLOBAL_H
#include "global.h"
#endif
#ifndef CL_RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef TREELISTS_H
#include "treelists.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif
#ifndef TK_TEXTSW_H
#include "tk_textsw.h"
#endif

static HashNodeList marklist;
static char buffer[800];
static RigiReport report_type;

static void VisitComponent( HashNodeList &nlist, GraphNode* np )
{
    // Visit the component of the graph that this node is a part of.
    // First, mark this node as visited. Then for each of its outgoing
    // arcs, visit it if necessary. It is not necessary to visit the node
    // if the node is not in the subgraph we care about (NODELIST) or the
    // node has already been visited (it is in the list MARKLIST).

    marklist.Add( *np );

    AvlPos *P;
    GraphArc* ap = np->FirstAdj( P, outArcs );
    while( ap != NULL )
    {
        GraphNode* dst = ap->GetDst();
        if ( nlist.Find( dst->GetId() ) != NULL  &&
            marklist.Find( dst->GetId() ) == NULL    )
        {
            VisitComponent( nlist,dst );
        }
        ap = np->NextAdj( P, outArcs );
    }

    // Now repeat the process with the incoming arcs as well.

    ap = np->FirstAdj( P, inArcs );
    while( ap != NULL )
    {
        GraphNode* src = ap->GetSrc();
        if ( nlist.Find( src->GetId() ) != NULL  &&
            marklist.Find( src->GetId() ) == NULL    )
        {
            VisitComponent( nlist,src );
        }
        ap = np->NextAdj( P, inArcs );
    } 
}

static int GetNumComponents( HashNodeList &list )
{
    // Get the number of disjoint graph components in the subgraph given
    // by LIST. This is done by traversing LIST and if it has not already
    // been marked as visited, we visit the graph component that the node
    // is a part of, marking each node we visit. For each call to
    // VisitComponent() we have found a new component, since we will never
    // call VisitComponent() on a component of the graph that has already
    // been visited in a component traversal.

    int num = 0;
    marklist.Empty();

    IterDO( list,GraphNode*,np )
    {
        if ( marklist.Find( np->GetId() ) == NULL )
        {
            VisitComponent( list,np );
            num++;
        }
    }
    ODretI;
    return num;
}

void GetCouplingSets( HashNodeList &list,
                    ArcSet &hlist, ArcSet &mlist, ArcSet &llist,
                    float thigh, float tlow )
{
    int count = 0;
    AvlPos *P;

    // For each node, check its outgoing composite arcs and if it is
    // outside either of the high or low thresholds, add it to the
    // appropriate list. The LIST should not be empty.

    IterDO( list,GraphNode*,np )
    {
        GraphArc* ap = np->FirstAdj( P, outArcs );
        while( ap != NULL )
        {
            GraphNode *dst = ap->GetDst();

            // We only care about arcs that are in the subgraph, and we only
            // care about composite arcs. We only care about composite arcs
            // because at the lowest level there is a composite arc for every
            // normal arc within a subsystem. As we go up the hierarchy, no
            // normal arcs exist, so we only have composie arcs to work with.

            if ( list.Find( dst->GetId() ) )
            {
                float contribs = (float)(ap->NumContributorIds());
                if (contribs > thigh) hlist.Add(ap);
                if (contribs < tlow)  llist.Add(ap);
                if (contribs >= tlow && contribs <= thigh) mlist.Add(ap);
            }
            ap = np->NextAdj( P, outArcs );
        }
    }
    ODretI;
}

float PartitionQuality( RigiNodeList& list, float thigh, float tlow )
{

    // Put each of the nodes in the list in a hash list for fast look up.

    HashNodeList hashlist;
    RigiNode* rnp = list.First();
    while( rnp != NULL )
    {
        hashlist.Add( *(rnp->GetNode()) );
        rnp = list.Next();
    }

    //  Compute the high and low coupling sets for the subgraph.

    ArcSet highset,medset,lowset;
    GetCouplingSets( hashlist,highset,medset,lowset,thigh,tlow );

    // Compute the number of components in the subgraph

    int numcomp = GetNumComponents( hashlist );

    // Get the sets sizes for the computation of the partition quality.

    float highsize = (float)highset.Size();
    float medsize = (float)medset.Size();
    float lowsize = (float)lowset.Size();
    float num = (float)list.GetNoItems();

    // Compute the partition quality and return the value.

    float pq;
    float complete = num*(num-1);
    float maxpq = complete;
    float minpq = -complete;
    float factor = num;

    // If only one node then pq = 1

    pq = 1.0F / (1.0F + (lowsize + num*medsize + 2*num*highsize)/(complete+1));

    // Write the report to the window if desired

    if ( report_type == REPORT_VERBOSE )
    {
        sprintf(buffer,"  Partition quality = %g\n",pq);
        if (!TextWriteStr(buffer)) return 0.0F;
        sprintf(buffer,"    High strength interfaces   = %g\n",highsize);
        if (!TextWriteStr(buffer)) return 0.0F;
        sprintf(buffer,"    Medium strength interfaces = %g\n",medsize);
        if (!TextWriteStr(buffer)) return 0.0F;
        sprintf(buffer,"    Low strength interfaces    = %g\n\n",lowsize);
        if (!TextWriteStr(buffer)) return 0.0F;
    }

    return pq;
}

    static int funcexternal;
    static int dataexternal;
    static int funcinternal;
    static int datainternal;

        static GraphArc *f1( GraphArc *ap )
        {
            GraphNode* src = ap->GetSrc();
            GraphNode* dst = ap->GetDst();
            if (children.Find(src->GetId()) == NULL)
                mayday("Could not find src of requirement (arc id %d, src %d)",
                    ap->GetId(), src->GetId());
            if (ap->GetType() == CALL_ARC)
                funcexternal++;
            if (ap->GetType() == DATA_ARC) {
                if (src->GetType() != DAT_NODE || dst->GetType() != DAT_NODE)
                    dataexternal++;
            }
            return NULL;
        }
        static GraphArc *f2( GraphArc* ap )
        {
            GraphNode* src = ap->GetSrc();
            GraphNode* dst = ap->GetDst();
            if (children.Find(dst->GetId()) == NULL)
                mayday("Could not find dst of provision (arc id %d, dst %d)", 
                    ap->GetId(), dst->GetId());
            if (ap->GetType() == CALL_ARC)
                funcexternal++;
            if (ap->GetType() == DATA_ARC)
            {
                if (src->GetType() != DAT_NODE || dst->GetType() != DAT_NODE)
                    dataexternal++;
            }
            return NULL;
        }
        static GraphArc *f3( GraphArc *ap )
        {
            GraphNode* s = ap->GetSrc();
            GraphNode* d = ap->GetDst();
            if (children.Find(s->GetId())==NULL)
                mayday("Could not find src of internalization (arc id %d, src %d)",
                    ap->GetId(), s->GetId());
            else if (children.Find(d->GetId())==NULL)
                mayday("Could not find dst of the internalization (arc id %d, dst %d)",
                    ap->GetId(), d->GetId());
            if (ap->GetType() == CALL_ARC)
                funcinternal++;
            if ((ap->GetType() == DATA_ARC ) && 
                (s->GetType() == MODULE_NODE || d->GetType() == MODULE_NODE)) {
                datainternal++;
            }
            return NULL;
        }

void CheckQuality( RigiNodeList& list )
{
    ArcSet require,provide,internal;

    // Check to see if the list is empty. If so report an error message. This
    // rotuine assumes that the nodes in the list are valid nodes to apply this
    // operation to.
    
    if ( list.GetNoItems() == 0 )
    {
        RigiError(ERRORFLAG, "Graph quality requires one or more nodes selected.");
        return;
    }

    // Get some info from the control panel.

    report_type = editor->GetReportGeneration();
    float thigh = editor->GetHighThreshold();
    float tlow = editor->GetLowThreshold();

    // Turn on batching of text writes to window to speed up the process
    TextStartReport();

    if (!TextWriteStr("Graph Quality (C Language Domain)\n"))
        return;
    sprintf(buffer,"  High Threshold        = %g\n",thigh);
    if (!TextWriteStr(buffer)) return;
    sprintf(buffer,"  Low Threshold         = %g\n",tlow);
    if (!TextWriteStr(buffer)) return;
    sprintf(buffer,"  Set Size              = %g\n\n",(float)(list.GetNoItems()));
    if (!TextWriteStr(buffer)) return;

    // Compute the partition quality of the subgraph defined by LIST

    float pq = PartitionQuality( list,thigh,tlow );

    // For each node, compute the Encapsulation quality for that node and keep
    // track of the sum.

    float eq = 0.0F;
    float hq = 0.0F;

    RigiNode* rnp = list.First();
    while( rnp != NULL )
    {
        // Compute the interface between this node and the rest of the graph.
        // This is done by traversing the subtree rooted at this node, and
        // collecting all of the arcs (Normal arcs, excluding Composite and
        // Refractive arcs) that the subtree requires, provides, and
        // internalizes. This information is used to compute the data and
        // control encapsualtion qualities.

        GraphNode* np = rnp->GetNode();
        NodeInterface(np,0,RigiTrue,require,provide,internal);

        // Process each of the requirement arcs. If it is a requirement of
        // a function, then add one to the function interface count. If it
        // is a requirement of a datatype, then we ignore it, as the
        // requirement between two datatypes is not considered an interface.

        funcinternal = 0;
        datainternal = 0;
        funcexternal = 0;
        dataexternal = 0;

        require.Iterator(f1);

        // Process each of the provide arcs. If it is a provision of
        // a function, then add one to the function interface count. If it
        // is a provision of a datatype, then add one to the datatype interface
        // count.

        provide.Iterator(f2);

        // Process each of the internal arcs. If it is an internalization of
        // a function call, then add one to the function internal count. If it
        // is an internalization of a datatype, then add one to the datatype
        // internalization count.

        internal.Iterator(f3);

        // Calculate the control encapsulation quailty.

        float neweq = 1.0F - (float)funcexternal/(funcinternal+funcexternal+1);
        sprintf(buffer,"  Control encapsulation quality for node %s = %g\n",
                        rnp->GetNode()->GetLabel(),neweq);
        if (!TextWriteStr(buffer)) return;
        if (report_type == REPORT_VERBOSE)
        {
            sprintf(buffer,"    Internalizations = %d\n",funcinternal);
            if (!TextWriteStr(buffer)) return;
            sprintf(buffer,"    Interfaces       = %d\n",funcexternal);
            if (!TextWriteStr(buffer)) return;
        }

        // Compute the data encapsulation quality.

        float newhq = 1.0F - (float)dataexternal/(datainternal+dataexternal+1);
        sprintf(buffer,"  Data encapsulation quality for node %s = %g\n",
                        rnp->GetNode()->GetLabel(),newhq);
        if (!TextWriteStr(buffer)) return;
        if (report_type == REPORT_VERBOSE)
        {
            sprintf(buffer,"    Internalizations = %d\n",datainternal);
            if (!TextWriteStr(buffer)) return;
            sprintf(buffer,"    Interfaces       = %d\n",dataexternal);
            if (!TextWriteStr(buffer)) return;
        }
        if (!TextWriteStr("\n")) return;

        hq += newhq;
        eq += neweq;
        rnp = list.Next();
    }

    // Normalize the encapsulation quality.

    eq = eq/(float)(list.GetNoItems());
    hq = hq/(float)(list.GetNoItems());

    // Compute the compostion quality

    float cq = (hq + eq + pq)/3.0F;

    // Output the composition quality report to the output window.

    if (!TextWriteStr("\n  Summary:\n")) return;
    sprintf(buffer,"    Partition Quality             = %g\n",pq);
    if (!TextWriteStr(buffer)) return;
    sprintf(buffer,"    Control Encapsulation Quality = %g\n",eq);
    if (!TextWriteStr(buffer)) return;
    sprintf(buffer,"    Data Encapsulation Quality    = %g\n",hq);
    if (!TextWriteStr(buffer)) return;
    sprintf(buffer,"    Composition Quality           = %g\n",cq);
    if (!TextWriteStr(buffer)) return;

    // Turn off batching of the text window
    TextEndReport();
    CharType cmd("w_texteditor ");
    cmd.Append(TextGetFileName());
    exec_command(cmd.string());
}

#undef OLDQUALITY
#ifdef OLDQUALITY

#endif

