
// $Author: jmartin $
// $Date: 2002-11-10 22:16:58 $
// $Source: /root/cvs/Rigi/rigiedit/rigiparamops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: rigiparamops.C,v 1.45 2002-11-10 22:16:58 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

// Definition include files

#ifndef _CCHEADER_H_
#include "CCheader.h"		// C++ base definitions
#endif
#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"		// Error messages
#endif
#ifndef RIGIPARAMOPS_H
#include "rigiparamops.h"	// Includes the definitions for this file
#endif

#ifndef GLOBAL_H
#include "global.h"             // Access to the global editor
#endif
#include "environ.h"            // Access to the environment variables
#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Include the class def. for editor state
#endif

#ifndef GRAPHNODE_H
#include "cl_graphnode.h"       // for GetId()
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"          // Include the class def. for domain info
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"		// Basic control functions for the editor
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"		// Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"		// Utility control functions for the editor
#endif
#include "win_util.h"		// Window state access functions
#ifndef RIGIINIT_H
#include "rigi_init.h"		// Rigi domain model
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"	// For the Rcl interpreter.
#endif

// Which parameters to be set
const int OVDEPTHPARAM		= 0;
const int PROJDEPTHPARAM	= 1;
const int HIERDEPTHPARAM	= 2;
const int SCALEFACTORPARAM	= 3;
const int SCALEFUNCTIONPARAM	= 4;
const int ANNOPARAM		= 5;
const int MODEPARAM		= 6;
const int COMMANDPARAM		= 7;
const int GRIDPARAM		= 9;
const int THIGHPARAM		= 10;
const int TLOWPARAM		= 11;
const int EISIMPLEPARAM		= 12;
const int CURRENTARCTYPEPARAM	= 13;
const int INHERITFILTERPARAM	= 14;
const int CONSTRAINTFUNCPARAM	= 15;
const int MBARPARAM		= 16;

// MJW: Get parameter values
const int GETARCTYPEPARAM	= 17;
const int GETPARAMETER		= 18;

const int NODETYPEPARAM		= 19;
const int NODENAMEPARAM		= 20;
const int ARCTYPEPARAM		= 21;
const int ARCNAMEPARAM		= 22;
const int NUMNODETYPEPARAM	= 23;
const int NUMARCTYPEPARAM	= 24;
const int NEWNODETYPEPARAM	= 25;
const int NEWARCTYPEPARAM	= 26;

const int FIRSTNODEATTRPARAM	= 27;
const int NEXTNODEATTRPARAM	= 28;
const int NUMNODEATTRPARAM	= 29;
const int FIRSTARCATTRPARAM	= 30;
const int NEXTARCATTRPARAM	= 31;
const int NUMARCATTRPARAM	= 32;

const int GETNODEORDNAME	= 33;	// HACK!
const int GETDOMAINNAME		= 34;

const int CHANGENODETYPENAME	= 35;
const int CHANGEARCTYPENAME	= 36;


const int SETLEVELARC		= 40;
const int SETCOMPOSITEARC	= 41;
const int SETCOLLAPSENODE	= 42;

const int GETLEVELARC		= 45;
const int GETCOMPOSITEARC	= 46;
const int GETCOLLAPSENODE	= 47;

const int MOVEUNISONPARAM	= 48;
const int GETROOTIDPARAM	= 49;

// Arc set operations

const int ARCSETGET		= 50;
const int ARCSETCLEAR		= 51;
const int ARCSETSET		= 52;
const int ARCSETADD		= 53;
const int ARCSETDEL		= 54;

int RigiParamOp( int argc, char **argv, CharType &result ) {
    int op;
    // Check to see if the parameters are correct.

    if (argc != 2 && argc != 3)
        return RigiError(ERRORFLAG, "Invalid parameter for operation %s", argv[0]);

    // Set up some stuff that is required by the operations

    RigiDomain* domain = RinitGetDomain();
    op = atoi( argv[1] );

    // Perform the operations

    switch( op ) {
        case ARCSETGET:
        {
            int i, numarcs = domain->NumArcTypes();
            result.Assign("");
            for(i=0 ; i<numarcs; i++) {
                if ( editor->InArcSet(i) ) {
                    result.Append(domain->ArcNameFromType(i));
                    result.Append(" ");
                }
            }
            break;
        }
        case ARCSETCLEAR:
            editor->ClearArcSet();
            break;
        case ARCSETSET:
        {
            editor->ClearArcSet();
            int i, targc;
            const char **targv;
            if (Tcl_SplitList( rcl_interp, argv[2], &targc, &targv ) != TCL_OK)
                return RigiError(ERRORFLAG, "Invalid arc type set specification");
            for( i=0 ; i<targc ; i++ )
                editor->SetArcSet((char*)targv[i], 1);
	    free((char*)targv);
            break;
        }
        case ARCSETADD:
            editor->SetArcSet( argv[2], 1 );
            break;
        case ARCSETDEL:
            editor->SetArcSet( argv[2], 0 );
            break;
        default:
            break;
    }

    return 1;
}

int RigiParamOp( int whichparam, int newvalue ) {
    int r = newvalue;
    RigiBool boolvar;
    RigiReport report;
    RigiConstraint constraint;
    WindowSetCursor( CURSOR_WAIT );

    switch (whichparam) {
    case OVDEPTHPARAM:
        editor->SetOverviewDepth( newvalue );
        break ;
    case PROJDEPTHPARAM:
        editor->SetProjectionDepth( newvalue );
        break ;
    case HIERDEPTHPARAM:
        editor->SetHierarchyDepth( newvalue );
        break;
    case SCALEFACTORPARAM:
        editor->SetScaleFactor( newvalue );
        break;
    case SCALEFUNCTIONPARAM:
        editor->SetScaleType( (RigiScale)newvalue );
        break;
    case ANNOPARAM:
        warn("Operations: Annotations are no longer supported");
        break;
    case MODEPARAM:
        warn("Operations: Latex mode is no longer supported");
        break;
    case GRIDPARAM:
        editor->SetGridSize( newvalue );
        break;
    case THIGHPARAM:
        editor->SetHighThreshold( (float)newvalue );
        break;
    case TLOWPARAM:
        editor->SetLowThreshold( (float)newvalue );
        break;
    case EISIMPLEPARAM:
        report = (newvalue)? REPORT_SIMPLE : REPORT_VERBOSE;
        editor->SetReportGeneration( report );
        break;
    case CURRENTARCTYPEPARAM:
        editor->SetArcType( newvalue );
        break;
    case INHERITFILTERPARAM:
        boolvar = (!newvalue)? RigiTrue : RigiFalse;
        editor->SetInheritFilters( boolvar );
        break;
    case MOVEUNISONPARAM:
        boolvar = (newvalue)? RigiTrue : RigiFalse;
        editor->SetMoveUnison( boolvar );
        break;
    case CONSTRAINTFUNCPARAM:
        if (newvalue == 0)      constraint = CON_NONE;
        else if (newvalue == 1) constraint = CON_HORIZ;
        else                    constraint = CON_VERT;
        editor->SetConstraintType( constraint );
        break;
    case GETARCTYPEPARAM:
        r = editor->GetArcType();
        break;
    case GETPARAMETER:
        r = 0;
        break;
    default: 
        break ;
    }
    WindowSetCursor( CURSOR_NORMAL );
    return r;
}

int RigiGetParam( int whichparam, char *result, int buffsize, char *name ) {
    // result is an in/out parameter
    // name is used infrequently, when extra parameter is required

    static const int INVALIDPARAM = -9999;
    static AvlPos *NP, *AP;

    RigiDomain *domain = RinitGetDomain();
    int r = INVALIDPARAM;
    const char *s = NULL;

    switch (whichparam) {
    case OVDEPTHPARAM:
        r = editor->GetOverviewDepth();
        break;
    case PROJDEPTHPARAM:
        r = editor->GetProjectionDepth();
        break;
    case HIERDEPTHPARAM:
        r = editor->GetHierarchyDepth();
        break;
    case SCALEFACTORPARAM:
        r = editor->GetScaleFactor();
        break;
    case SCALEFUNCTIONPARAM:
        r = editor->GetScaleType();
        break;
    case ANNOPARAM:
        r = editor->GetDisplayAnnotation();
        break;
    case MODEPARAM:
        warn( "Operations: Latex mode is no longer supported" );
        r = 0;
        break;
    case COMMANDPARAM:
        r = editor->GetDisplayCommand();
        break;
    case GRIDPARAM:
        r = editor->GetGridSize();
        break;
    case THIGHPARAM:
        r = (int)editor->GetHighThreshold();
        break;
    case TLOWPARAM:
        r = (int)editor->GetLowThreshold();
        break;
    case EISIMPLEPARAM:
        r = editor->GetReportGeneration();
        break;
    case CURRENTARCTYPEPARAM:
        r = editor->GetArcType();
        break;
    case INHERITFILTERPARAM:
        r = editor->GetInheritFilters();
        break;
    case MOVEUNISONPARAM:
        r = editor->GetMoveUnison();
        break;
    case CONSTRAINTFUNCPARAM:
        r = editor->GetConstraintType();
        break;
    case MBARPARAM:
        r = editor->GetDisplayMessage();
        break;
    case NODETYPEPARAM:
        r = domain->NodeTypeFromName( result );
        break;
    case NODENAMEPARAM:
        s = domain->NodeNameFromType( atoi(result) ); 
        break;
    case ARCTYPEPARAM:
        r = domain->ArcTypeFromName( result );
        break;
    case ARCNAMEPARAM:
        s = domain->ArcNameFromType( atoi(result) ); 
        break;
    case NUMNODETYPEPARAM:
        r = domain->NumNodeTypes();
        break;
    case NUMARCTYPEPARAM:
        r = domain->NumArcTypes();
        break;
    case NEWNODETYPEPARAM:
        r = domain->NewNodeType( result );
        break;
    case NEWARCTYPEPARAM:
        r = domain->NewArcType( result );
        break;
    case FIRSTNODEATTRPARAM:
        s = domain->FirstNodeAttr(NP);
        break;
    case NEXTNODEATTRPARAM:
        s = domain->NextNodeAttr(NP);
        break;
    case NUMNODEATTRPARAM:
        r = domain->NumNodeAttrs();
        break;
    case FIRSTARCATTRPARAM:
        s = domain->FirstArcAttr(AP);
        break;
    case NEXTARCATTRPARAM:
        s = domain->NextArcAttr(AP);
        break;
    case NUMARCATTRPARAM:
        r = domain->NumArcAttrs();
        break;
    case GETNODEORDNAME:
    {
        int type = atoi(result);
        s = domain->NodeNameFromType( type );
        break;
    }
    case GETDOMAINNAME:
        s = EnvGet( "RIGIDOMAIN" );
        break;
    case CHANGENODETYPENAME:
        if ( name && result ) {
            domain->ChangeNodeTypeName( result, name );
        }
        break;
    case CHANGEARCTYPENAME:
        if ( name && result ) {
            r = domain->ChangeArcTypeName( result, name );
        }
        break;
    case SETLEVELARC:
        editor->SetLevelArc( domain->ArcTypeFromName( result ) );
        break;
    case SETCOMPOSITEARC:
        editor->SetCompositeArc( domain->ArcTypeFromName( result ) );
        break;
    case SETCOLLAPSENODE:
        editor->SetCollapseNode( domain->NodeTypeFromName( result ) );
        break;
    case GETLEVELARC:
        r = editor->LevelArc();
        break;
    case GETCOMPOSITEARC:
        r = editor->CompositeArc();
        break;
    case GETCOLLAPSENODE:
        r = editor->CollapseNode();
        break;
    case GETROOTIDPARAM:
        r = RigiBaseNode()->GetId();
        break;
    default:
        RigiError(ERRORFLAG, "Invalid operation (%d)", whichparam);
        break;
    }
    
    if (r == INVALIDPARAM) {
        result[buffsize-1] = '\0';
        if (s)
            strncpy( result,s,buffsize-1 );
        else
            result[0] = '\0';
    } else
        sprintf( result, "%d", r );

    return 1;
}


