// $Author: jmcdanie $
// $Date: 1996-05-22 22:20:51 $
// $Source: /root/cvs/Rigi/rigiedit/selectdst.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: selectdst.C,v 1.18 1996-05-22 22:20:51 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif

#ifndef SELECTDST_H
#include "selectdst.h"
#endif

#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"
#endif
#ifndef RIGIARC_H
#include "cl_rigiarc.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef GRAPHARC_H
#include "cl_grapharc.h"
#endif
#ifndef ACTIVENODE_H
#include "cl_activenode.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif

void SelectDst(RigiNodeList& slist, RigiWindow* win, ArcType type)
{
    SelectNeighbour(slist,win,type,outArcs);
}

void SelectSrc(RigiNodeList& slist, RigiWindow* win, ArcType type)
{
    SelectNeighbour(slist,win,type,inArcs);
}

void SelectNeighbour(RigiNodeList& nlist, RigiWindow* win, ArcType type, 
AdjList adj) {
    // For each node in the list, get all of its destination/source nodes.

    RigiNode* rn = nlist.First();
    while( rn != NULL ) {
        GraphNode* gn = rn->GetNode();
        AvlPos *P;
        GraphArc* ga = gn->FirstAdj(P, adj);

        while( ga != NULL ) {

            // Only want to follow arcs of type TYPE

            if (ga->GetType() == type || ANY_ARC == type)  {
                // This is ugly because no hash lists. All we want to do
                // is figure out which RigiNode goes with this node/window
                // pair. We have to get the active node and search all of
                // its RigiNodes to find the one that is in the window we
                // are interested in

                GraphNode *neighbour_gn = (adj == inArcs || adj == inLevArcs )?
                    ga->GetSrc() : ga->GetDst();

                ActiveNode* an = (ActiveNode*)(neighbour_gn->GetActiveNode());
                if (an != NULL){
                    RigiNode* neighbour_rn = an->First();
                    while( neighbour_rn != NULL) {
                        if (neighbour_rn->GetWindow() == win)
                            break;
                         neighbour_rn = an->Next();
                    }
                    if (neighbour_rn != NULL)
                        editor->SelectNode(neighbour_rn);
                }
            }
            ga = gn->NextAdj( P, adj );
        }
        rn = nlist.Next();
    }
}

void SelectRecursive( RigiWindow* win, ArcType type ) {
    // For each node in the list, get all of its destination nodes.

    // Look for recursive dependencies amongst all nodes
    AvlPos *P, *Q;
    RigiNode* rn = win->FirstNode(P);
    while( rn != NULL ) {
        GraphNode* gn = rn->GetNode();
        // Since it is recursive - only need to look at one set of arcs
        GraphArc* ga = gn->FirstAdj( Q, inArcs );
        while( ga != NULL ) {

            // Only want to follow arcs of type TYPE

            if ( ga->GetType()==type || ANY_ARC==type ) {
                // This is ugly because no hash lists. All we want to do
                // is figure out which RigiNode goes with this node/window
                // pair. We have to get the active node and search all of
                // its RigiNodes to find the one that is in the window we
                // are interested in

                GraphNode* neighbour_gn;
                neighbour_gn = ga->GetSrc();

                ActiveNode* an = (ActiveNode*)(neighbour_gn->GetActiveNode());
                if (an != NULL) {
                    RigiNode* neighbour_rn = an->First();
                    while( neighbour_rn != NULL ) {
                        if ( neighbour_rn->GetWindow() == win )
                            break;
                        else neighbour_rn = an->Next();
                    }
                    // Check if it is a directly recursive
                    if ( neighbour_rn != NULL && neighbour_rn == rn ) 
                        editor->SelectNode( neighbour_rn );
                }
            }
            ga = gn->NextAdj( Q, inArcs );
        }
        rn = win->NextNode(P);
    }
}

void SelectType( RigiWindow *rw, NodeType type )
{
    AvlPos *P;
    for ( RigiNode *rn = rw->FirstNode(P); rn; rn = rw->NextNode(P) )
        if (rn->GetNode()->GetType()==type) editor->SelectNode(rn);
}

