
// $Author: jmartin $
// $Date: 2000-02-05 02:18:34 $
// $Source: /root/cvs/Rigi/rigiedit/statisticsops.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: statisticsops.C,v 1.41 2000-02-05 02:18:34 jmartin Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif
#include <sys/types.h>

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif

#ifndef CSTUFF_H
#include "cstuff.h"
#endif
#ifndef GRAPHDEF_H
#include "cl_graphdef.h"         // Basic graph definitions
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"        // Class definition for graph Nodes
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"       // Class definition for RigiWindows
#endif
#ifndef RIGINODE_H
#include "cl_riginode.h"         // Class definition for RigiNodes
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"           // Class definition for the domain
#endif
#ifndef GLOBAL_H
#include "global.h"              // Global definitions for the editor
#endif
#ifndef STATISTICSOPS_H
#include "statisticsops.h"
#endif
#ifndef RCL_COMMANDS_H
#include "rcl_commands.h"
#endif

// Utility include files
#ifndef RIGI_H
#include "rigi.h"                // Basic control functions for the editor
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"           // Data base & object creation utilities
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"            // Utility control functions for the editor
#endif
#ifndef RIGIERR_H
#include "rigierr.h"             // Error reporting functions
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"           // for node and arc info
#endif
#include "win_util.h"

// Operations include files
#ifndef EXACT_H
#include "exact.h"
#endif
#ifndef TK_TEXTSW_H
#include "tk_textsw.h"
#endif
#ifndef QUALITY_H
#include "quality.h"             // Composition quality operations
#endif

#define MAX_INT_SIZE  (12)	// max number of bytes for printed num


void StatisticsOp( int op )
{
    WindowSetCursor( CURSOR_WAIT );

    // Get some initial state information for the windows.

    RigiWindow *activewin = RutilActiveWindow();
    RigiArc *selectarc = editor->GetSelectedArc();

    // Careful with this list. It is possible that an operation can result in
    // the deletion of RigiNodes. This list contains pointers to RigiNodes, so
    // the list may not point to valid objects after some operations. The ops
    // that you need to worry about are the edit operations, as they update
    // the windows. Most other ops just change the state of the nodes.

    RigiNodeList slist;
    editor->GetSelectionList( slist, activewin );

    // Now do the appropriate statistics report
    switch ( op )
    {
        case EINTERFACE:
            if ( selectarc != NULL )
            {
                ArcExactInterface(selectarc->GetArc());
            }
            else {
                if (editor->NumSelectedNodes() < 1)
                {
                    RigiError(ERRORFLAG, "Exact interface requires a selected node or arc");
                    break ;
                }
                NodeListExactInterface(slist);
            }
            break ;
        case EINTERFACETYPED:
            if (editor->NumSelectedNodes() < 1)
            {
                RigiError(ERRORFLAG, "Exact interface requires a selected node or arc");
                break ;
            }
            NodeListExactInterface(slist,editor->GetArcType()); 
            break ;
        case GQUALITY:
            CheckQuality( slist ) ; 
            break ;
        case WINSTATS:
        {
            char printStr[128];
            RigiDomain *domain = RinitGetDomain();
            GraphNode* np = activewin->GetNode() ;
            int total, showed;
            int gtotal = 0, gshowed = 0;

            // Turn on batching of text writes to speed up the writing
            TextStartReport();

            sprintf(printStr, "Statistics for Window %-60.60s\n\n", np->GetLabel());
        
            if (!TextWriteStr(printStr)) {
                TextEndReport();
                break;
            }

            if (!TextWriteStr(" Node Type           Visible  Filtered     Total\n") ||
                !TextWriteStr(" -----------------  --------  --------  --------\n")) {
                TextEndReport();
                break;
            }
            int numtypes = domain->NumNodeTypes();
	    int i;
            for ( i=0; i<numtypes; i++ ) {
                total = activewin->NodeNumber(i);
                // do not show the ones that are not here at all
                if (total) {
                    gtotal += total;
                    showed = activewin->ShowedNodeNumber(i);
                    gshowed += showed;
                    // check what kind of the report we are generating
                    if ( editor->GetReportGeneration() == REPORT_VERBOSE ) {
                        sprintf(printStr," %-17.17s%10d%10d%10d\n", domain->NodeNameFromType(i),
                            showed, (total - showed), total);
                        if (!TextWriteStr(printStr)) {
                            TextEndReport();
                            break ;
                        }
                    }
                }
            }
            sprintf(printStr,"*Nodes - All Types%10d%10d%10d\n\n", gshowed,
                (gtotal - gshowed), gtotal);
            if (!TextWriteStr(printStr)) {
                TextEndReport();
                break ;
            }

            // Arcs per type
            if (!TextWriteStr(" Arc Type            Visible  Filtered     Total\n") ||
                !TextWriteStr(" -----------------  --------  --------  --------\n")) {
                TextEndReport();
                break;
            }
            numtypes = domain->NumArcTypes();
            for(i=0; i<numtypes; i++) {
                total = activewin->ArcNumber(i);
                // do not show the ones that are not here at all
                if(total) {
                    gtotal += total;
                    showed = activewin->ShowedArcNumber(i);
                    gshowed += showed;
                    // check what kind of the report we are generating
                    if ( editor->GetReportGeneration() == REPORT_VERBOSE ) {
                        sprintf(printStr," %-17.17s%10d%10d%10d\n", domain->ArcNameFromType(i),
                            showed, (total - showed), total);
                        if (!TextWriteStr(printStr)) {
                            TextEndReport();
                            break ;
                        }
                    }
                }
            }
            sprintf(printStr,"*Arcs - All Types %10d%10d%10d\n\n",  gshowed,
                (gtotal - gshowed), gtotal);
            if (!TextWriteStr(printStr)) {
                TextEndReport();
                break ;
            }

            TextEndReport();
            CharType cmd("w_texteditor ");
            cmd.Append(TextGetFileName());
            exec_command(cmd.string());
            break ;
        }
        default : 
            break ;
    }
    WindowSetCursor( CURSOR_NORMAL );
}

