
// $Author: jmcdanie $
// $Date: 1996-05-18 00:37:33 $
// $Source: /root/cvs/Rigi/rigiedit/tidy.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: tidy.C,v 1.16 1996-05-18 00:37:33 jmcdanie Exp $";

//------------------------------------------------------------------------

/*
    MJW [951106]
    I changed this completely. The last revision used ancient
    code involving many now deleted members of GraphNodes. Using
    recursive approach now.
*/
#include <stdio.h>
#include "cl_riginodeset.h"
#include "cl_grapharc.h"

static const int MAX  = 64;
static const int SIZE = 100;

static RigiNodeSet *visited;
static int maxdepth;
static int *extrema;

static void Revisit( RigiNode *rn, int depth, int offset );

int Tidy( RigiNode *rn, int depth ) {

    // initialize at root
    if (!depth) {
        visited = new RigiNodeSet();
        extrema = new int[MAX];
        memset(extrema, 0, sizeof(extrema));
        maxdepth = 0;
    }

    int posn;        // where this node thinks it should go
    int offset;      // diff between where really should go and posn
    int left=0,right=0,i=0;
    int max = maxdepth;
    if (depth==MAX) depth--; // HACK to prevent crashing

    if (visited->Find(rn)) return 0;
    visited->Insert(rn);

    // Find x-axis positions of all subtrees. Keep leftmost and rightmost.
    AvlPos *P;
    for ( RigiArc *ra=rn->FirstArc(P,ARC_OUT); ra; ra=rn->NextArc(P,ARC_OUT) )
    {
        right = Tidy( ra->GetDst(), depth+1 );
        if (i==0) left = right;
        i = 1;
    }

    // Calculate where I think I should go
    posn = i ? (left+right)/2 : SIZE;

    // Calculate the offset from where I really should go
    offset = depth>max ? 0 : extrema[depth];
    offset = offset + SIZE - posn;

    // Don't move left if kids were constrained
    if (offset<0 && max>depth) offset = 0;

    // The offset applies to the entire subtree's right contour
    if (offset && i)
        for ( i=depth+1; i<=maxdepth; i++ ) extrema[i]+=offset;

//Mess("posn %d, depth %d, offset %d, max=%d e=%d",
//posn,depth,offset,max,extrema[depth]);

    Point2d p;
    p.x = posn; p.y = offset;
    rn->SetPosition(p);
    extrema[depth] = posn + offset;
    if (depth>maxdepth) maxdepth = depth;

    // At end, we must reassign node positions based on the offsets
    if (!depth) {
        visited->Empty();
        Revisit( rn, 0, 0 );

        // clean up
        maxdepth = 0;
        delete[] extrema;
        delete visited;
        visited = NULL;
    } 
    return posn + offset;
}

void Revisit( RigiNode *rn, int depth, int offset )
{
    if (visited->Find(rn)) return;
    visited->Insert(rn);

    Point2d p = rn->GetPosition();
    offset += p.y;
    p.x += offset;
    p.y = SIZE * depth;
    rn->SetPosition(p);

    AvlPos *P;
    for ( RigiArc *ra=rn->FirstArc(P,ARC_OUT); ra; ra=rn->NextArc(P,ARC_OUT) )
        Revisit( ra->GetDst(), depth+1, offset );
}
