
// $Author: jmcdanie $
// $Date: 1996-05-26 18:13:04 $
// $Source: /root/cvs/Rigi/rigiedit/view.C,v $
// $Locker:  $
// $State: Exp $

const static char rcsid[] = "$Id: view.C,v 1.49 1996-05-26 18:13:04 jmcdanie Exp $";

//------------------------------------------------------------------------

#ifndef STREAMH
#include <stream.h>
#endif

#ifndef VIEW_H
#include "view.h"
#endif
#ifndef GRAPHNODESET_H
#include "cl_graphnodeset.h"
#endif
#ifndef RIGIWINDOW_H
#include "cl_rigiwindow.h"
#endif
#ifndef GRAPHNODE_H
#include "cl_graphnode.h"
#endif
#ifndef DOMAIN_H
#include "cl_domain.h"
#endif
#ifndef RIGIERR_H
#include "rigierr.h"
#endif
#ifndef GRAPHUTIL_H
#include "graphutil.h"
#endif
#ifndef RIGIUTIL_H
#include "rigiutil.h"
#endif
#ifndef RIGIINIT_H
#include "rigi_init.h"
#endif
#include "environ.h"
#include "win_util.h"
#ifndef RIGI_H
#include "rigi.h"
#endif

#ifndef RIGISTATE_H
#include "cl_rigistate.h"       // Global instance of editor's state
#endif
#ifndef GLOBAL_H
#include "global.h"
#endif

// Error message buffer for this file
static char tmpstr[MAXPATHLEN+MAXNAMLEN+80+80];
static const char WINDOWTAG[] = "RIGIWINDOW";

static int CorruptViewFile(FILE *fp, char *fileName) {
    RigiError(ERRORFLAG, "Corrupt view file %s", fileName);
    if (fp != NULL) fclose(fp);
    return 0;
}


// Peggy's Changes (1/6/92), Load View Fix
// Previously when a view was loaded the windows were loaded in reverse order. 
// This also mucked up the location of windows.  
// 
// I think the reason for this problem is the AddWindow routine, which
// uses a lower level routine Insert from cl_list.c.  Insert inserts to
// to the front of the list, therefore the window most recently added
// is at the front of the list.  However, when a view is saved and
// then loaded the order was changed, perhaps because the previous programmer
// did not notice how AddWindow works.  
//
// This recursive procedure solves this problem by writing the windows
// to the file (fp) in reverse order. 

void ReverseAddWindow( RigiWindow* basewin, RigiWindow* rw, FILE* fp ) {
    RigiWindow* nextwin = editor->NextWindow();
    if ( nextwin && nextwin != basewin && nextwin->GetType() != WIN_SCRATCH)
        ReverseAddWindow(basewin,nextwin,fp);

    if (rw != basewin && rw->GetType() != WIN_SCRATCH)
        rw->WriteArchive(fp);
}

// Write a view file given the name of the view. This function prepends
// the view directory and appends the .view prefix to the view name given

int DoViewWrite(char* selection)
{
    char* viewdir;
    CharType viewname;

    // Check to see if a selection was made
    if (selection == NULL || strlen(selection) == 0)
        return RigiError(ERRORFLAG, "Error, no selection made for opening a view!");

    // Check to make sure the directory for the views has been set
    viewdir = EnvGet("DBDIR");
    if (viewdir == NULL) return 0;
    viewname.Assign( viewdir );

    // Construct the view file from the directory and the view names
    viewname.Append("/");
    viewname.Append(selection);
    viewname.Append(".vw");
    Mess( "DoViewWrite: view = %s", viewname.string() );
    ViewWrite(viewname.string());
    return 1;
}

// Read a view file given the name of the view. This function prepends
// the view directory and appends the .view prefix to the view name given

int DoViewRead(char* selection)
{
    char* viewdir;
    CharType viewname;

    // Check to see if a selection was made
    if (selection == NULL || strlen(selection) == 0) 
        return RigiError(ERRORFLAG, "No selection made for opening a view!");

    // Check to make sure the directory for the views has been set
    viewdir = EnvGet("DBDIR");
    if (viewdir == NULL) return 0;
    viewname.Assign( viewdir );

    // Get rid of trailing special characters at the end of the selection
    // If these special characters exist in the middle of the selection,
    // then they are not removed. We remove anything that is less than
    // character ! and greater than ~, which are the ascii characters that
    // bound the printable characters, not including the space character.
    char* sptr = selection + strlen(selection) - 1;
    while(sptr > selection && (*sptr < '!' || *sptr > '~'))
    {
        *sptr = '\0';
        sptr--;
    }
    viewname.Append("/");
    viewname.Append(selection);
    viewname.Append(".vw");
    Mess( "DoViewRead: view = %s", viewname.string() );
    if (!ViewRead(viewname.string())) return 0;
    return 1;
}

int ViewWrite(char* filename ) {
    FILE* fp;

    // Open the file for writing
    fp = fopen(filename,"w");
    if (fp == NULL)
        return RigiError(ERRORFLAG, "Unable to open view file %s", filename);

    // Write the root window to the view file
    RigiWindow* basewin = RigiBaseWindow();
    if (basewin == NULL) {
        RigiError(ERRORFLAG, "NULL base window - unable to write view");
        fclose(fp);
        return 0;
    }
    if (!basewin->WriteArchive(fp)) return 0;

    // For each window, write it to the view file. If their are no windows
    // in the list, print a warning message.
    // See comments at ReverseAddWindow procedure above. 
    RigiWindow* curwin = editor->FirstWindow();
    if (curwin == NULL) {
        RigiError(ERRORFLAG, "Window states were not written");
        fclose(fp);
        return 0;
    } else {
        ReverseAddWindow(basewin, curwin, fp);    
    }

    fprintf(fp,"END_RIGIWINDOWS\n");

    // Write the global control information that is necessary to restore the
    // state of the editor
    editor->WriteArchive(fp);

    fprintf(fp,"0 0    10 10\n"); 
    
    // Close the file
    fclose(fp);

    return 1;
}

int ViewRead( char* rootname )
{
    FILE* fp;
    char tag[80];
    char filename[MAXPATHLEN+MAXNAMLEN+80];
    NodeSet nodes;
    ArcFlags inarctypes, outarctypes;
    RigiDomain *domain = RinitGetDomain();

    int i,result,viewcount;
    int wintype,windepth,wininc,numnodes,grasid;
    int nnode,narc;
    int x,y,w,h;
    GraphNode* node;

    // passed in the complete view name, with the .view extension already
    sprintf(filename,"%s",rootname);

    fp = fopen(filename,"r");
    if (fp == NULL)
        return RigiError(ERRORFLAG, "Unable to open file %s", filename);

    // Read each window in the view file
    viewcount = 0;
    result = fscanf(fp,"%s",tag);
    if (result != 1 || strcmp(WINDOWTAG,tag) != 0)
        return CorruptViewFile(fp, filename);

    while (result == 1 && strcmp(WINDOWTAG,tag) == 0) {
        // Read in the window type
        result = fscanf(fp,"%d",&wintype);
        if (result!=1 || 
            (wintype != WIN_GENERAL && wintype != WIN_OVERVIEW
             &&  wintype != WIN_SCRATCH  && wintype != WIN_PROJECT ))
            return CorruptViewFile(fp, filename);

        // Read the window depth (ignored for normal windows)
        result = fscanf(fp,"%d",&windepth);
        if (result != 1) return CorruptViewFile(fp, filename);

        // Read the number of nodes and arcs. This should be the same
        // number as the system is using. If it is not, give a warning.

        result = fscanf(fp,"%d %d",&nnode,&narc);
        if (result != 2) return CorruptViewFile(fp, filename);

        int dn = domain->NumNodeTypes();
        int da = domain->NumArcTypes();
        if (nnode != dn
            && !RigiError(ALERT, "View contains %d node types;\n%d were expected", nnode, dn)) {
            fclose(fp);
            return 0;
        }

        if (narc  != da 
            && !RigiError(ALERT, "View contains %d arc types;\n%d were expected", narc, da)) {
            fclose(fp);
            return 0;
        }
        // Read in the number of parent nodes
        result = fscanf(fp,"%d",&numnodes);
        if (result != 1 || numnodes <= 0) return CorruptViewFile(fp, filename);

        // Read in each of the parents node ids. Do not forget to empty
        // the list 8-)

        nodes.Empty();
        for ( i=0; i<numnodes; i++ ) {
            result = fscanf(fp,"%d",&grasid);
            if (result != 1) return CorruptViewFile(fp, filename);
            node = GutilNodeExists(grasid);
            if (node == NULL)
                return CorruptViewFile(fp, filename);
            else
                nodes.Add(node);
        }

        // If the window is a GENERAL window, read in its appropriate info

        if (wintype == WIN_GENERAL)
        {
            char *buff = new char[narc+1];
            char format[32];
            inarctypes.SetSize(narc);
            outarctypes.SetSize(narc);

            // Build a format so fscanf reads in the correct string length

            sprintf(format, "%%%ds", narc );

            // Read in and process the incoming arc set

            if (fscanf( fp, format, buff ) != 1) {
                delete[] buff;
                return CorruptViewFile(fp, filename);
            }

            if ((int)strlen(buff) != narc 
                && !RigiError(ALERT, "Inconsistent number of IN arcs (%d vs %d)",
                    narc, strlen(buff))) {
                    fclose(fp);
                    delete[] buff;
                    return 0;
            }

            for ( i=0; i<narc; i++ ) {
                if (buff[i] == '1')
                    inarctypes.SetBit(i);
                else if (buff[i] == '0')
                    inarctypes.ClearBit(i);
                else {
                    CorruptViewFile(fp, filename);
                    delete[] buff;
                    return 0;
                }
            }

            // Read in and process the outgoing arc set

            if (fscanf( fp, format, buff ) != 1) {
                CorruptViewFile(fp, filename);
                delete[] buff;
                return 0;
            }
            if ((int)strlen(buff) != narc
                && !RigiError(ALERT, "Inconsistent number of OUT arcs (%d vs %d)",
                    narc, strlen(buff))) {
                    fclose(fp);
                    delete[] buff;
                    return 0;
            }

            for ( i=0; i<narc; i++ ) {
                if (buff[i] == '1')
                    outarctypes.SetBit(i);
                else if (buff[i] == '0')
                    outarctypes.ClearBit(i);
                else {
                    CorruptViewFile(fp, filename);
                    delete[] buff;
                    return 0;
                }
            }

            // Read in the rest of the relevant GENERAL window information

            fscanf(fp, "%d", &wininc);
            delete[] buff;
        }

        // Create the RigiWindow. We only create a new window if the window is
        // not the root window. If it is the root window, we only want one, so 
        // we use the one already in existence.

        RigiWindow* newwin;
        if (RigiBaseNode()->GetId() == node->GetId()) {
            newwin = RigiBaseWindow();
        // B0 - remove the following line for root window
        // This allows the root window to be iconified in the view also
        /*    newwin->Hide(); */
        } else {
            int x = 0, y = 0;
            RutilGetRootLocation(x, y);
            if (wintype == WIN_GENERAL)
                newwin = new RigiWindow( nodes, inarctypes, outarctypes, 
                                        wininc, windepth, x, y, NULL);
            else
                newwin = new RigiWindow( nodes,(RigiWinType)wintype,windepth,
                            domain->NumNodeTypes(),domain->NumArcTypes(),x,y);
            editor->AddWindow(newwin) ;
        }
        // MJW: make this window active, for drawing consistency
        RutilSetActiveWindow( newwin );

        // Read in the windows state (plus filters!)

        newwin->ReadArchive(fp);

        // Show the new window

        newwin->Show(); 

        // P11 Update filter message in case any filters were set in the view

        newwin->UpdateFilterMessage();
        RutilRaiseWindow(newwin);
        newwin->RigiDisplay();

        // Increment the count of the number of windows, and read in the
        // next TAG.
        viewcount++;
        result = fscanf(fp,"%s",tag);
    }

    // Read the global stuff
    editor->ReadArchive(fp);

    // Display some diagnostic warnings if the view file may have been corrupted
    if (viewcount == 0) return CorruptViewFile(fp, filename);

    // Read the location and size of the editor
    result = fscanf(fp,"%d %d %d %d",&x,&y,&w,&h);
    if (result != 4) return CorruptViewFile(fp, filename);
    
    // Close the view file
    fclose(fp);

    // Put the .txt extension on the root name
    // But I need to remove the .view before I do this
    char stem[MAXPATHLEN+MAXNAMLEN+80];
    int len, stemlen;
    len = strlen(rootname);
    stemlen = len - 5;  // take off 5 for .view
    strncpy(stem, rootname, stemlen);  // just left with * with .view part
    stem[stemlen] = '\0';  // Null terminate
    sprintf(filename,"%s.txt",stem);

    // Open the file. If open fails don't worry about it, as it is not
    // necessary to have a text file accompany a view
    FILE* tmpfp = fopen(filename,"r");
    if (tmpfp == NULL) return 1;
    fclose(tmpfp);

    return 1;
}

